/* buffer.c --
   
   Part of: tclmore
   Contents: buffer implementation
   Date: Sat Jan 24, 2004
   
   Abstract
   
	This module implements buffers: a  linked list of byte arrays to
	be  used  to store  data  coming from  an  entity  and going  to
	another.
   
   Copyright (c) 2004 Marco Maggi
   
   The author  hereby grant permission to use,  copy, modify, distribute,
   and  license this  software  and its  documentation  for any  purpose,
   provided that  existing copyright notices  are retained in  all copies
   and that  this notice  is included verbatim  in any  distributions. No
   written agreement, license, or royalty  fee is required for any of the
   authorized uses.  Modifications to this software may be copyrighted by
   their authors and need not  follow the licensing terms described here,
   provided that the new terms are clearly indicated on the first page of
   each file where they apply.
   
   IN NO  EVENT SHALL THE AUTHOR  OR DISTRIBUTORS BE LIABLE  TO ANY PARTY
   FOR  DIRECT, INDIRECT, SPECIAL,  INCIDENTAL, OR  CONSEQUENTIAL DAMAGES
   ARISING OUT  OF THE  USE OF THIS  SOFTWARE, ITS DOCUMENTATION,  OR ANY
   DERIVATIVES  THEREOF, EVEN  IF THE  AUTHOR  HAVE BEEN  ADVISED OF  THE
   POSSIBILITY OF SUCH DAMAGE.
   
   THE  AUTHOR  AND DISTRIBUTORS  SPECIFICALLY  DISCLAIM ANY  WARRANTIES,
   INCLUDING,   BUT   NOT  LIMITED   TO,   THE   IMPLIED  WARRANTIES   OF
   MERCHANTABILITY,    FITNESS   FOR    A    PARTICULAR   PURPOSE,    AND
   NON-INFRINGEMENT.  THIS  SOFTWARE IS PROVIDED  ON AN "AS  IS" BASIS,
   AND  THE  AUTHOR  AND  DISTRIBUTORS  HAVE  NO  OBLIGATION  TO  PROVIDE
   MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
*/


/** ------------------------------------------------------------
 ** Header files.
 ** ----------------------------------------------------------*/

#if 0
#define TCLMORE_DEBUG_MESSAGES
#define TCLMORE_DEBUGGING	1
#endif

#include "tclmoreInt.h"
#include <assert.h>
#include <stddef.h>
#include <string.h>


/** ------------------------------------------------------------
 ** Module's type declarations.
 ** ----------------------------------------------------------*/

typedef More_Byte	Byte;

/* Buffer --

   Type of buffer.  Instances of this type are chained in a linked list.


   dataAreaSize -	Number of bytes in the buffer data area.
   
   nextBuffer -		Pointer to the next buffer in the chain.

   firstByte -		Pointer to the first byte in the data area.

   firstFreeByte -	Pointer  to the  first unused  byte in  the data
			area.   NULL if  the  buffer is  full; equal  to
			firstByte if the buffer is empty. To compute the
			number   of  bytes  used   in  the   data  area:
			(firstFreeByte-firstByte).
*/

typedef struct Buffer {
  size_t		dataAreaSize;
  struct Buffer *	nextBuffer;
  Byte *		firstByte;
  Byte *		firstFreeByte;
  Byte			dataArea[1];
} Buffer;

/* SharedBuffer --

  Type used as base for shared  buffers. When a buffer is shared between
  two instances of some type,  possibly owned by different threads, this
  type allows exclusive access and reference counting.


  mutex -		Mutex  used  for  exclusive  acces. It  is  used
			automatically the shared buffer functions.

  readerAttached -	Boolean: true if an entity has registered itself
			to use the buffer as its input.

  writerAttached -	Boolean: true if an entity has registered itself
			to use the buffer as its output.

  buffer -		Pointer to the first buffer in a chain.

  bufferSize -		Dimension of  the data area for  the next buffer
			allocated.

  readerCallback -	Callback  function  used  to notify  the  entity
			attached to  the output side that the  buffer is
			readable.

  writerCallback -	Callback  function  used  to notify  the  entity
			attached to  the input  side that the  buffer is
			writable.


    A controller is  responsible to create an instance  of this type and
  send references to it to a couple of entities: one using the buffer as
  input and one using the buffer  as output. When the source or the sink
  both detach themself from the buffer, the buffer is freed.

*/

typedef struct SharedBuffer {
  Tcl_Mutex		mutex;
  Buffer *		buffer;
  size_t		bufferSize;
  Byte			readerAttached;
  Byte			writerAttached;
  More_Callback		readerCallback;
  More_Callback		writerCallback;
} SharedBuffer;



/** ------------------------------------------------------------
 ** Preprocessor symbols.
 ** ----------------------------------------------------------*/

#define debug			More_Debug
#define question		More_Question

#define DEFAULT_BUFFER_SIZE	4096

#define MovePointerToLastBuffer(buffer)		\
	while ((buffer)->nextBuffer) { (buffer) = (buffer)->nextBuffer; }

#define GetMutex(buffer)	(&((buffer)->mutex))


/** ------------------------------------------------------------
 ** Assertions and conditions.
 ** ----------------------------------------------------------*/

/*
  An empty buffer can be:

  * the  only  buffer  in  the   chain,  for  an  indefinite  number  of
    operations;

  * temporarily the  first buffer in  the chain: in the  read operation,
    when all the  data is consumed, the first buffer is  empty but it is
    released immediately if it is not the only one in the chain;

  * temporarily the last buffer in the chain: in a write operation, when
    a buffer  is filled with data and  more data must be  written, a new
    buffer is allocated and appended to  the chain, but it is written to
    immediately.

    With the single exception  of "More_BufferAlloc()", at the beginning
  and end of a function: if a buffer  is empty it is the only one in the
  chain.  For "More_BufferAlloc()":  the condition  is true  when  a new
  chain is initialised.
*/
#define Condition_EmptyBufferIsAlone(buffer)	\
  assert(((buffer)->firstByte == (buffer)->firstFreeByte)? \
	 (NULL == (buffer)->nextBuffer) : 1)

/*
  "firstByte" always references the first byte in the data area.
*/
#define Invariant_FirstByte(buffer)	\
  assert((buffer)->firstByte == &((buffer)->dataArea[0]))

/*
  If  the buffer is  not full  (firstFreeByte !=  NULL), "firstFreeByte"
  references a byte in the data area.
*/
#define Invariant_FirstFreeByte(buffer)	\
  assert(((buffer)->firstFreeByte)? \
	  ((buffer)->firstByte + (buffer)->dataAreaSize) > (buffer)->firstFreeByte : 1)


/** ------------------------------------------------------------
 ** Module's function prototypes.
 ** ----------------------------------------------------------*/

static Buffer *	BufferAlloc	_ANSI_ARGS_((Buffer *buffer, size_t size));
static void	BufferFree	_ANSI_ARGS_((Buffer *buffer));
static Buffer *	BufferPopMaybe	_ANSI_ARGS_((Buffer *buffer));
static int	BufferRead	_ANSI_ARGS_((Buffer **bufferVar,
					     More_Block block));
static void	BufferWrite	_ANSI_ARGS_((Buffer *buffer,
					     More_Block block,
					     size_t newSize));
static int	BufferEmpty	_ANSI_ARGS_((Buffer *buffer));
static size_t	BufferUsed	_ANSI_ARGS_((Buffer *buffer));
static void	BufferReadAll	_ANSI_ARGS_((Buffer **bufferPtr,
					     Byte * outPtr));
static void	BufferReadChunk _ANSI_ARGS_((Buffer *buffer,
					     int outLen,
					     Byte * outPtr));

static void	SharedBufferFinal _ANSI_ARGS_((SharedBuffer *base));

static void	SharedBufferNotify _ANSI_ARGS_((SharedBuffer *base));


/* More_BufferAlloc --

	Allocates a new buffer.

   Arguments:

	None.

   Results:

        Returns the pointer to the base structure.

*/

More_Buffer
More_BufferAlloc (void)
{
  SharedBuffer *	base;


  base = (SharedBuffer *) ckalloc(sizeof(SharedBuffer));
  memset(base, '\0', sizeof(SharedBuffer));

  base->buffer		= NULL;
  base->readerAttached	= 1;
  base->writerAttached	= 1;
  base->bufferSize	= DEFAULT_BUFFER_SIZE;
  return ((More_Buffer) base);
}

/* More_BufferFree --

	Releases a shared buffer.

   Arguments:

	buffer -	The buffer token.

   Results:

        Detaches both the reader and the writer.

*/

void
More_BufferFree (More_Buffer buffer)
{
  More_BufferDetachReader(buffer);
  More_BufferDetachWriter(buffer);
}

/* More_BufferReaderCallback --

	Registers the callback used to notify the entity attached to the
	reader side that the buffer is readable.

   Arguments:

	buffer -	Buffer token.
	callback -	Callback token.

   Results:

	Stores values  in the instance function fields;  if the function
	pointer is NULL: the callback is reset.

	  The  notification  function  is  invoked.  If  the  reader  is
	attached  and  data  is   available:  the  function  is  invoked
	immediately.

*/

void
More_BufferReaderCallback (More_Buffer buffer, More_Callback callback)
{
  SharedBuffer *	base = (SharedBuffer *) buffer;


  debug("%s reader callback for %p",
	((callback.function)? "registering" : "clearing"), base);
  Tcl_MutexLock(GetMutex(base));
  base->readerCallback = callback;
  SharedBufferNotify(base);
  Tcl_MutexUnlock(GetMutex(base));
  debug("leaving reader registration");
}

/* More_BufferWriterCallback --

	Registers the callback used to notify the entity attached to the
	input side that the buffer is writable.

   Arguments:

	buffer -	Buffer token.
	callback -	Callback token.

   Results:

	Stores values  in the instance function fields:  if the function
	pointer is NULL: the callback is reset.

	  The  notification  function  is  invoked.  If  the  writer  is
	attached: the  function is invoked immediately  because a buffer
	is always writable.

*/

void
More_BufferWriterCallback (More_Buffer buffer, More_Callback callback)
{
  SharedBuffer *	base = (SharedBuffer *) buffer;


  debug("%s writer callback for %p",
	((callback.function)? "registering" : "clearing"), base);
  Tcl_MutexLock(GetMutex(base));
  base->writerCallback = callback;
  SharedBufferNotify(base);
  Tcl_MutexUnlock(GetMutex(base));
  debug("leaving writer registration");
}

/* More_BufferDetachReader --

	Detaches the reader from the buffer.

   Arguments:

	buffer -	Buffer token.

   Results:

        Resets  the reader  flag.  If  the writer  has  already detached
        itself: the  buffer is released; else  the notification function
        is invoked.

*/

void
More_BufferDetachReader (More_Buffer buffer)
{
  SharedBuffer *	base = (SharedBuffer *) buffer;


  Tcl_MutexLock(GetMutex(base));
  debug("detaching reader from %p", base);
  base->readerAttached = 0;
  if (! base->writerAttached)
    {
      debug("no writer, finalising %p", base);
      Tcl_MutexUnlock(GetMutex(base));
      SharedBufferFinal(base);
    }
  else
    {
      debug("writer present");
      if (More_CallbackPresent(base->writerCallback))
	{
	  debug("invoke callback %p", base);
	  More_CallbackInvoke(base->writerCallback);
	}
      Tcl_MutexUnlock(GetMutex(base));
    }
}

/* More_BufferDetachWriter --

	Detaches the writer from the buffer.

   Arguments:

	buffer -	Buffer token.

   Results:

        Resets  the writer  flag.  If  the reader  has  already detached
        itself: the  buffer is released; else  the notification function
        is invoked.

*/

void
More_BufferDetachWriter (More_Buffer buffer)
{
  SharedBuffer *	base = (SharedBuffer *) buffer;


  Tcl_MutexLock(GetMutex(base));
  debug("detaching writer from %p", base);
  base->writerAttached = 0;
  if (! base->readerAttached)
    {
      debug("no reader, finalising buffer %p", base);
      Tcl_MutexUnlock(GetMutex(base));
      SharedBufferFinal(base);
    }
  else
    {
      debug("reader present, callback present? %s",
	    question(More_CallbackPresent(base->readerCallback)));
      if (More_CallbackPresent(base->readerCallback))
	{
	  debug("invoke callback %p", base);
	  More_CallbackInvoke(base->readerCallback);
	}
      Tcl_MutexUnlock(GetMutex(base));
    }
}

/* More_BufferRead --

	Reads data from a buffer.

   Arguments:

	buffer -	Buffer token.
	block -		Output block token.

   Results:

        Returns the number of bytes read.

*/

int
More_BufferRead (More_Buffer buffer, More_Block block)
{
  SharedBuffer *	base = (SharedBuffer *) buffer;
  int	e;


  Tcl_MutexLock(GetMutex(base));
  e = (base->buffer)? BufferRead(&(base->buffer), block) : 0;
  SharedBufferNotify(base);
  Tcl_MutexUnlock(GetMutex(base));
  return e;
}

/* More_BufferReadAllBlock --

	Reads all the bytes in  the buffer storing them in a dynamically
	allocated block. The memory is allocated with "ckalloc()".

   Arguments:

	buffer -	Buffer token.

   Results:

        Returns a block referencing the bytes.

*/

More_Block
More_BufferReadAllBlock (More_Buffer token)
{
#undef  CHUNK_SIZE
#define CHUNK_SIZE	DEFAULT_BUFFER_SIZE
  SharedBuffer *	base = (SharedBuffer *) token;
  More_Block		block = { 0, NULL };


  Tcl_MutexLock(GetMutex(base));
  if (base->buffer && (! BufferEmpty(base->buffer)))
    {
      More_Block	chunk;
      int		readNum = 0, readTotal = 0;

      do
	{
	  chunk.len = CHUNK_SIZE;
	  block.len += chunk.len;
	  block.ptr = ckrealloc(block.ptr, (unsigned)block.len);
	  chunk.ptr = block.ptr + readTotal;

	  readTotal += readNum = BufferRead(&(base->buffer), chunk);
	}
      while (! BufferEmpty(base->buffer));

      block.ptr = ckrealloc(block.ptr, (unsigned)readTotal);
      block.len = readTotal;
    }
  SharedBufferNotify(base);
  Tcl_MutexUnlock(GetMutex(base));
  return block;
}

/* More_BufferReadAllObj --

	Reads all the  bytes in the buffer storing them  in a byte array
	object.

   Arguments:

	buffer -	Buffer token.

   Results:

        Returns a pointer to the new object.

*/

Tcl_Obj *
More_BufferReadAllObj (More_Buffer token)
{
#undef  CHUNK_SIZE
#define CHUNK_SIZE	DEFAULT_BUFFER_SIZE
  SharedBuffer *	base = (SharedBuffer *) token;
  Tcl_Obj *		objPtr = Tcl_NewByteArrayObj(NULL, 0);


  Tcl_MutexLock(GetMutex(base));
  if (base->buffer && (! BufferEmpty(base->buffer)))
    {
      More_Block	chunk, block;
      int		readNum = 0, readTotal = 0;

      do
	{
	  chunk.len = CHUNK_SIZE;
	  block.len += chunk.len;
	  block.ptr = Tcl_SetByteArrayLength(objPtr, block.len);
	  chunk.ptr = block.ptr + readTotal;

	  readTotal += readNum = BufferRead(&(base->buffer), chunk);
	}
      while (! BufferEmpty(base->buffer));

      Tcl_SetByteArrayLength(objPtr, readTotal);
    }
  SharedBufferNotify(base);
  Tcl_MutexUnlock(GetMutex(base));
  return objPtr;
}

/* More_BufferWrite --

	Writes data to the buffer.

   Arguments:

	buffer -	Buffer token.
	block -		Input block token.

   Results:

        Fills the buffer with data.

*/

void
More_BufferWrite (More_Buffer buffer, More_Block block)
{
  SharedBuffer *	base = (SharedBuffer *) buffer;


  debug("enter write function");
  Tcl_MutexLock(GetMutex(base));
  if (! base->buffer)
    {
      base->buffer = BufferAlloc(NULL, base->bufferSize);
    }
  BufferWrite(base->buffer, block, base->bufferSize);
  SharedBufferNotify(base);
  Tcl_MutexUnlock(GetMutex(base));
}

/* More_BufferEmpty --

	Tests if data is available in the buffer.

   Arguments:

	buffer -	Buffer token.

   Results:

        Returns true if the buffer is empty, otherwise returns false.

*/

int
More_BufferEmpty (More_Buffer buffer)
{
  SharedBuffer *	base = (SharedBuffer *) buffer;
  int			e;


  Tcl_MutexLock(GetMutex(base));
  e = (base->buffer)? BufferEmpty(base->buffer) : 1;
  Tcl_MutexUnlock(GetMutex(base));
  return e;
}

/* More_BufferSetSize --

	Selects a new value for the next buffer allocated.

   Arguments:

	buffer -	Buffer token.
	bufferSize -	New buffer size.

   Results:

        Stores the value.

*/

void
More_BufferSetSize (More_Buffer buffer, size_t bufferSize)
{
  SharedBuffer *	base = (SharedBuffer *) buffer;

  Tcl_MutexLock(GetMutex(base));
  base->bufferSize = bufferSize;
  Tcl_MutexUnlock(GetMutex(base));
}

/* More_BufferGetSize --

	Acquires the current buffer size.

   Arguments:

	buffer -	Buffer token.

   Results:

        Returns the buffer size.

*/

size_t
More_BufferGetSize (More_Buffer buffer)
{
  SharedBuffer *	base = (SharedBuffer *) buffer;
  size_t		bufferSize;

  Tcl_MutexLock(GetMutex(base));
  bufferSize = base->bufferSize;
  Tcl_MutexUnlock(GetMutex(base));
  return bufferSize;
}

/* More_BufferEof --

	Tests  the "end  of file"  condition  on a  shared buffer.   The
	condition  is true  if the  writer  has detached  itself and  no
	unread data is in the buffer.

   Arguments:

	buffer -	Buffer token.

   Results:

        Returns true if the writer has detached itself.

*/

int
More_BufferEof (More_Buffer buffer)
{
  SharedBuffer *	base = (SharedBuffer *) buffer;
  int			bool;

  Tcl_MutexLock(GetMutex(base));
  bool = (! base->writerAttached) && \
    ((base->buffer)? BufferEmpty(base->buffer) : 1);
  Tcl_MutexUnlock(GetMutex(base));
  return bool;
}

/* More_BufferAlive --

	Tests the  effect of writing data  to the buffer.  If the reader
	has detached itself the operation is useless.

   Arguments:

	buffer -	Buffer token.

   Results:

        Returns true if the reader has not detached itself.

*/

int
More_BufferAlive (More_Buffer buffer)
{
  SharedBuffer *	base = (SharedBuffer *) buffer;
  int			bool;

  Tcl_MutexLock(GetMutex(base));
  bool = base->readerAttached;
  Tcl_MutexUnlock(GetMutex(base));
  return bool;
}

/* SharedBufferNotify --

	Notifies the reader and writer about the state of the buffer.

   Arguments:

	base -		Pointer to the buffer.

   Results:

        For the  reader, if:  it is attached,  no notification  was done
        after the  last read  operation, there is  data, a  callback has
        been  registered: the  callback is  invoked;  a flag  is set  to
        signal the  notification: it will  be cleared by the  first read
        operation on the channel.

          For the writer,  if: it is attached, no  notification was done
        after the last write  operation, a callback has been registered:
        the  callback  is   invoked;  a  flag  is  set   to  signal  the
        notification: it will be cleared by the first write operation.

	  Note that  notifications to  the reader are  sent even  if the
	writer has detached itself; the same for inverted roles.

*/

static void
SharedBufferNotify (SharedBuffer *base)
{
  int	notEmptyOrEof;

  debug("enter notification function");


  notEmptyOrEof = \
    (((base->buffer)? (! BufferEmpty(base->buffer)) : 0) ||
     (! base->writerAttached));

  if (base->readerAttached && More_CallbackPresent(base->readerCallback) &&
      notEmptyOrEof)
    {
      debug("invoking reader callback");
      More_CallbackInvoke(base->readerCallback);
    }

  if (base->writerAttached && More_CallbackPresent(base->writerCallback))
    {
      debug("invoking writer callback");
      More_CallbackInvoke(base->writerCallback);
    }
  debug("leaving notification function");
}

/* SharedBufferFinal --

	Finalises a buffer releasing all the resources.

   Arguments:

	base -		Pointer to the buffer.

   Results:

        Finalises the mutex and frees memory.

*/

static void
SharedBufferFinal (SharedBuffer *base)
{
  debug("finalising %p", base);
  Tcl_MutexFinalize(GetMutex(base));
  BufferFree(base->buffer);
  ckfree((char *) base);
  debug("finalised %p", base);  
}

/* BufferAlloc --

	Allocates a new buffer and appends  it at the end of an existing
	chain; or  alloctes a  new buffer,  first of a  new chain.  If a
	request to append a new buffer is issued, the last buffer in the
	chain must be full.

   Arguments:

	buffer -	Pointer to a buffer in a chain or NULL
			if a new chain is to be initialised.
	size -		The size of the data area in bytes. Must be
			greater than zero, if it is not: defaults
			to 4096.

   Results:

        Returns the pointer to the new buffer: the last in the chain.

*/

static Buffer *
BufferAlloc (Buffer *buffer, size_t size)
{
  Buffer *	last;


  if (size <= 0)
    {
      size = DEFAULT_BUFFER_SIZE;
    }

  debug("request to allocate a buffer of %d bytes, is it for a new chain? %s",
	size, (NULL == buffer)? "yes" : "no");

  last = (Buffer *) ckalloc(sizeof(Buffer)+size-1);
  memset(last, '\0', sizeof(Buffer));

  last->firstByte	= &(last->dataArea[0]);
  last->firstFreeByte	= last->firstByte;
  last->dataAreaSize	= size;

  Invariant_FirstByte(last);
  Invariant_FirstFreeByte(last);

  if (! buffer)
    {
      debug("allocated a new chain, first buffer %p", last);
      Condition_EmptyBufferIsAlone(last);
      return last;
    }
  else
    {
      Buffer *	prev = buffer;

      debug("allocated a tail buffer for %p, new buffer %p",
	    buffer, last);
      MovePointerToLastBuffer(prev);
      assert(NULL == prev->nextBuffer);
      prev->nextBuffer = last;
      return last;
    }
}

/* BufferFree --

	Frees the  resources allocated to  a buffer and all  the buffers
	next  to the selected  one. If  the selected  buffer is  not the
	first in a chain: it must  be detached from the chain before the
	invocation to this function.

   Arguments:

	buffer -	Pointer to the buffer. If NULL: nothing happens.

   Results:

	The chain is traversed until the end, releasing all the buffers.

*/

static void
BufferFree (Buffer *buffer)
{
  Buffer *	prevBuffer;


  if (! buffer)
    {
      return;
    }

  Condition_EmptyBufferIsAlone(buffer);
  Invariant_FirstByte(buffer);
  Invariant_FirstFreeByte(buffer);

  debug("releasing buffers after %p, is it the last? %s",
	buffer, (buffer->nextBuffer)? "no" : "yes");

  while (buffer->nextBuffer)
    {
      prevBuffer = buffer;
      buffer = buffer->nextBuffer;
      ckfree((char *) prevBuffer);
    }
  ckfree((char *)buffer);
}

/* BufferPopMaybe --

	Frees the first buffer in a chain. This operation is meant to be
	used to free the first buffer  in a chain, when all its data has
	been consumed.  If the  selected buffer is  the only one  in the
	chain, it is not freed and this function does nothing.

   Arguments:

	buffer -	Pointer to the buffer to free. Must be the first
			in a chain.

   Results:

        Frees  the selected  buffer if  it is  not the  only one  in the
        chain. Returns the pointer to  the new first buffer in the chain
        or the buffer itself.

*/

static Buffer *
BufferPopMaybe (Buffer *buffer)
{
  Buffer *	prevBuffer;


  Condition_EmptyBufferIsAlone(buffer);
  Invariant_FirstByte(buffer);
  Invariant_FirstFreeByte(buffer);

  if (buffer->nextBuffer)
    {
      prevBuffer = buffer;
      buffer = buffer->nextBuffer;
      ckfree((char *) prevBuffer);
    }

  Condition_EmptyBufferIsAlone(buffer);
  Invariant_FirstByte(buffer);
  Invariant_FirstFreeByte(buffer);
  return buffer;
}

/* BufferRead --

	Reads   the  requested   number   of  bytes   in  the   provided
	buffer. Updates the state of the buffer.

   Arguments:

	bufferVar -	Pointer to a variable holding a pointer
			to the buffer. Should be the first one
			in a chain.
	block -		The output block token.

   Results:

	Returns the number of bytes read; zero if the source is empty.

	  If the  source chain is  not made of  a single buffer  and the
	read operation has  consumed all the data in  the first element:
	the first  element is freed  with a call  to "BufferPopMaybe()";
	the buffer variable is updated with the pointer to the new first
	buffer in the chain.

*/

static int
BufferRead (Buffer **bufferVar, More_Block block)
{
  Buffer *	buffer = *bufferVar;
  int		copied = 0;
  size_t	used;


  Condition_EmptyBufferIsAlone(buffer);
  Invariant_FirstByte(buffer);
  Invariant_FirstFreeByte(buffer);

  if ((! block.len) && (! block.ptr))
    {
      return 0;
    }

  debug("entering read for buffer %p, request to read %d bytes",
	buffer, block.len);

  while (block.len)
    {
      used = BufferUsed(buffer);
      debug("%d bytes left to be read, %d used bytes in buffer",
	    block.len, used);
      if (! used)
	{
	  break;
	}
      if (used > block.len)
	{
	  BufferReadChunk(buffer, block.len, block.ptr);
	  copied += block.len;
	  break;
	}
      else /* used <= block.len */
	{
	  BufferReadAll(&buffer, block.ptr);
	  copied += used;
	  block.len -= used;
	  block.ptr += used;
	}
      Condition_EmptyBufferIsAlone(buffer);
      Invariant_FirstByte(buffer);
      Invariant_FirstFreeByte(buffer);
    }

  debug("returning with %d bytes read", copied);
  *bufferVar = buffer;
  Condition_EmptyBufferIsAlone(buffer);
  Invariant_FirstByte(buffer);
  Invariant_FirstFreeByte(buffer);
  return copied;
}

/* BufferReadAll --

	Reads all the bytes in the first buffer.

   Arguments:

	bufferPtr -	Pointer to the variable holding a pointer
			to the first buffer.
	outPtr -	Pointer to the output buffer.

   Results:

        Copies all the bytes to  the output buffer.  If the first buffer
        in the  chain is  the only one:  updates its state  and returns;
        else: pops  it from  the chain and  updates the  buffer variable
        with a pointer to the new first buffer.

*/

static void
BufferReadAll (Buffer **bufferPtr, Byte * outPtr)
{
  Buffer *	buffer = *bufferPtr;
  size_t	used;


  Condition_EmptyBufferIsAlone(buffer);
  Invariant_FirstByte(buffer);
  Invariant_FirstFreeByte(buffer);

  used = BufferUsed(buffer);

  debug("reading a chunk of %d bytes from %p", used, buffer);

  memcpy(outPtr, buffer->firstByte, (size_t) used);
  if (buffer->nextBuffer)
    {
      debug("consumed data in first buffer, popping a buffer");
      *bufferPtr = BufferPopMaybe(buffer);
      Condition_EmptyBufferIsAlone(*bufferPtr);
      Invariant_FirstByte(*bufferPtr);
      Invariant_FirstFreeByte(*bufferPtr);
    }
  else
    {
      debug("consumed data in tail buffer");
      buffer->firstFreeByte = buffer->firstByte;
      Condition_EmptyBufferIsAlone(buffer);
      Invariant_FirstByte(buffer);
      Invariant_FirstFreeByte(buffer);
    }
}

/* BufferReadChunk --

	Reads a chunk  of data from the first buffer  in the chain. This
	function is used  when the number of bytes to  read is less than
	the number of bytes in the buffer.

   Arguments:

	buffer -	Pointer to the first buffer.
	outLen -	Number of bytes to read.
	outPtr -	Pointer to the output block.

   Results:

        Copies the chunk to the  output block; then shifts the data left
        in the buffer to the beginning  of the data area: if this is the
        only buffer in  the chain, this operation allows  the next write
        request to fill the buffer.

   Note about the shift operation:

	Delaying  the shift operation  until a  real write  operation is
	requested  would improve performance,  at the  cost of  adding a
	"firstUsedByte"  pointer  in the  structure  and  a little  more
	logic. This  is not  done because it  is believed by  the author
	that the  performance improvement is  little.

	  Choosing  carefully  the  size  of  the buffer  is  much  more
	important.   Usually there are  two situations:  one in  which a
	stream of data  is transferred, mostly in one  direction; one in
	wich chunks of data are transferred bidirectionally to implement
	a  communication protocol.   In the  first case  the performance
	improvement would  be significant  if read and  write operations
	are mixed. In the second case the size of the buffer required to
	hold the whole chunk should be known at run time.

*/

static void
BufferReadChunk (Buffer *buffer, int outLen, Byte * outPtr)
{
  assert(BufferUsed(buffer) > outLen);

  debug("reading last chunk of %d bytes", outLen);
  memcpy(outPtr, buffer->firstByte, (size_t)outLen);
  memmove(buffer->firstByte, (buffer->firstByte + outLen),
	  (size_t)outLen);
  if (buffer->firstFreeByte)
    {
      buffer->firstFreeByte -= outLen;
    }
  else
    {
      buffer->firstFreeByte = buffer->firstByte + \
	(buffer->dataAreaSize - outLen);
    }
  debug("%d bytes left in buffer", BufferUsed(buffer));
}

/* BufferWrite --

	Appends  the bytes  to  the end  of  the buffer  chain. If  some
	buffers in the chain are half filled: they are ignored; only the
	tail buffer is filled before the allocation of a new element.

   Arguments:

	bufferPtr -	Pointer to the first buffer in the
			chain.
	block -		The input block token.
	size -		Size of new buffers to be allocated
			if required. If zero defaults to
			DEFAULT_BUFFER_SIZE.

   Results:

        Traverses the chain  until the last buffer is  found; then fills
        the  last  buffer. If  more  bytes  are  present in  the  input:
        allocates a new  buffer, appends it to the  end and writes data.
        Loops appending  new buffers until  all the input data  has been
        consumed.

*/

static void
BufferWrite (buffer, block, size)
     Buffer  *		buffer;
     More_Block		block;
     size_t		size;
{
  Byte *	firstFreeByte;
  size_t	free;


  Invariant_FirstByte(buffer);
  Invariant_FirstFreeByte(buffer);
  Condition_EmptyBufferIsAlone(buffer);

  if ((! block.len) || (! block.ptr)) { return; }
  if (! size)  { size = DEFAULT_BUFFER_SIZE; }

  debug("request to write %d bytes to buffer %p", block.len, buffer);

  MovePointerToLastBuffer(buffer);
  if (! buffer->firstFreeByte)
    {
      debug("tail buffer is full, allocating new one");
      buffer = BufferAlloc(buffer, size);
    }

  while (block.len)
    {
      firstFreeByte = buffer->firstFreeByte;
      free = buffer->dataAreaSize - (firstFreeByte - buffer->firstByte);
      debug("%d bytes free in the current buffer", free);
      if (free > block.len)
	{
	  debug("writing last chunk of %d bytes", block.len);
	  memcpy(firstFreeByte, block.ptr, (size_t)block.len);
	  firstFreeByte += (ptrdiff_t)block.len;
	  buffer->firstFreeByte = \
	    ((buffer->firstByte + buffer->dataAreaSize) < firstFreeByte)? \
	    NULL : firstFreeByte;
	  break;
	}
      else
	{
	  debug("writing a chunk of %d bytes, current buffer is full", free);
	  memcpy(firstFreeByte, block.ptr, free);
	  buffer->firstFreeByte = NULL;
	  buffer = BufferAlloc(buffer, size);
	  block.len -= free;
	  block.ptr += free;
	}
      Invariant_FirstByte(buffer);
      Invariant_FirstFreeByte(buffer);
      /* Here the last buffer may be empty and not alone. */
    }

  Invariant_FirstByte(buffer);
  Invariant_FirstFreeByte(buffer);
  Condition_EmptyBufferIsAlone(buffer);
  debug("returning with %d bytes left free in the tail buffer",
	(buffer->firstFreeByte)? \
	(buffer->dataAreaSize - \
	 (buffer->firstFreeByte - buffer->firstByte)) : 0);
}

/* BufferEmpty --

	Tests if the chain is empty. If the first buffer in the chain is
	empty, then  the first buffer is  the only one in  the chain and
	the whole chain is empty.

   Arguments:

	buffer -	Pointer to the buffer.

   Results:

        Returns true if the buffer is empty, otherwise returns false.

*/

static int
BufferEmpty (Buffer *buffer)
{
  Condition_EmptyBufferIsAlone(buffer);
  Invariant_FirstByte(buffer);
  Invariant_FirstFreeByte(buffer);
  return (buffer->firstByte == buffer->firstFreeByte);
}

/* BufferUsed --

	Determines the number of bytes used in the selected buffer.

   Arguments:

	buffer -	Pointer to the buffer.

   Results:

        Returns the number of bytes.

*/

static size_t
BufferUsed (Buffer *buffer)
{
  if (buffer->firstFreeByte)
    {
      return ((buffer->firstFreeByte)? \
	      (buffer->firstFreeByte - buffer->firstByte) : buffer->dataAreaSize);
    }
  else
    {
      return buffer->dataAreaSize;
    }
}


/* end of file */
/*
Local Variables:
mode: c
page-delimiter: "^$"
End:
*/
