/* bufferchannel.c --
   
   Part of: tclmore
   Contents: channel driver attached to buffers.
   Date: Mon Feb  2, 2004
   
   Abstract
   
	This module  defines the implementation of  a Tcl_Channel driver
	linked to  a couple of  More_Buffer instances used as  input and
	output.

	  This driver is  used as base for the  [varchan] and [pipechan]
	commands.
   
   Copyright (c) 2004 Marco Maggi
   
   The author  hereby grant permission to use,  copy, modify, distribute,
   and  license this  software  and its  documentation  for any  purpose,
   provided that  existing copyright notices  are retained in  all copies
   and that  this notice  is included verbatim  in any  distributions. No
   written agreement, license, or royalty  fee is required for any of the
   authorized uses.  Modifications to this software may be copyrighted by
   their authors and need not  follow the licensing terms described here,
   provided that the new terms are clearly indicated on the first page of
   each file where they apply.
   
   IN NO  EVENT SHALL THE AUTHOR  OR DISTRIBUTORS BE LIABLE  TO ANY PARTY
   FOR  DIRECT, INDIRECT, SPECIAL,  INCIDENTAL, OR  CONSEQUENTIAL DAMAGES
   ARISING OUT  OF THE  USE OF THIS  SOFTWARE, ITS DOCUMENTATION,  OR ANY
   DERIVATIVES  THEREOF, EVEN  IF THE  AUTHOR  HAVE BEEN  ADVISED OF  THE
   POSSIBILITY OF SUCH DAMAGE.
   
   THE  AUTHOR  AND DISTRIBUTORS  SPECIFICALLY  DISCLAIM ANY  WARRANTIES,
   INCLUDING,   BUT   NOT  LIMITED   TO,   THE   IMPLIED  WARRANTIES   OF
   MERCHANTABILITY,    FITNESS   FOR    A    PARTICULAR   PURPOSE,    AND
   NON-INFRINGEMENT.  THIS  SOFTWARE IS PROVIDED  ON AN "AS  IS" BASIS,
   AND  THE  AUTHOR  AND  DISTRIBUTORS  HAVE  NO  OBLIGATION  TO  PROVIDE
   MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
   
*/

/*

            reader     --------------    writer
                ---<--| input buffer |---<--
               |       --------------       |
      -----------                          -----------
     | channel A |                        | channel B |
      -----------                          -----------
               |       ---------------      |
                --->--| output buffer |-->--
            writer     ---------------   reader

*/

/** ------------------------------------------------------------
 ** Header files.
 ** ----------------------------------------------------------*/

#if 0
#  define TCLMORE_DEBUG_MESSAGES
#  define TCLMORE_DEBUGGING	1
#endif

#include <string.h>
#include <errno.h>
#include <assert.h>
#include <stddef.h>
#include <limits.h>

#include "tclmoreInt.h"


/** ------------------------------------------------------------
 ** Preprocessor symbols.
 ** ----------------------------------------------------------*/

#define debug			More_Debug
#define question		More_Question

#define CHANNEL_NAME_LENGTH	(16 + TCL_INTEGER_SPACE)
#define CHANNEL_PATTERN		"bufchan%u"


/** ------------------------------------------------------------
 ** Module's type declarations.
 ** ----------------------------------------------------------*/

/* Channel --
  
  Type of  variable channel  instance data. Instances  of this  type are
  allocated by "More_OpenBufferChannel()" with "ckalloc()".


  inputBuffer -		The input buffer. NULL if this channel is  write
			only.

  outputBuffer -	The output buffer. NULL if this channel is  read
			only.

  channel -		The  channel token.

  openMode -		The  open   mode  for  the   channel:  an  OR-ed
			combination of TCL_READABLE and TCL_WRITABLE. It
			is used  by the watch function  to register only
			event interests compatible with the channel open
			mode.

  blockingMode -	The blocking  mode configured with [fconfigure],
			"-blocking"     option.     The     value     is
			TCL_MODE_BLOCKING or TCL_MODE_NONBLOCKING.

  eventMask -		Bit mask  of events  of interest to  the generic
			channel  layer.  When  interest in  an  event is
			signaled  to the  driver  with [fileevent],  the
			request   is  registered   here  by   the  watch
			function.

  queuedReader -	Boolean.  True if the  event for  readable input
			buffer  has been  queued and  not  consumed yet.

  queuedWriter -	Boolean. True  if the event  for writable output
			buffer  has been  queued and  not  consumed yet.

  mutex -		Used    to    ensure    exclusive   access    to
			"queuedReader" and "queuedWriter".
*/

typedef struct Channel {
  More_Buffer		inputBuffer;
  More_Buffer		outputBuffer;
  Tcl_Channel		channel;
  int			openMode;
  int			blockingMode;
  int			eventMask;
  int			queuedReader;
  int			queuedWriter;
  Tcl_Mutex		mutex;
} Channel;

/* Event --

  Type used as  event descriptor. Events are used to  queue in the event
  loop notifications to the generic layer about channel events.

    The  event loop  is  the one  of  the thread  to  which the  channel
  belongs, so the event callback function can test for channel existence
  with "Tcl_IsChannelExisting()" safely.

  event -	TCL's event structure.

  channel -	The channel token.

  channelName -	Name of  the channel.  It  is required because  when the
		event  is  consumed  the  channel may be already closed.

  eventMask -	The event mask, an OR-ed combination of TCL_READABLE and
		TCL_WRITABLE.     It   is    used    as   argument    to
		Tcl_NotifyChannel().

*/

typedef struct Event {
  Tcl_Event	event;
  Tcl_Channel	channel;
  char		channelName[CHANNEL_NAME_LENGTH];
  int		eventMask;
} Event;


/** ------------------------------------------------------------
 ** Operational preprocessor macros.
 ** ----------------------------------------------------------*/

#define ChannelName(data)		(Tcl_GetChannelName((data)->channel))
#define ChannelModeIsReadable(data)	((data)->inputBuffer)
#define ChannelModeIsWritable(data)	((data)->outputBuffer)
#define ChannelModeIsBlocking(data)	(TCL_MODE_BLOCKING == (data)->blockingMode)

#define InputBufferEof(data)		(More_BufferEof((data)->inputBuffer))
#define InputBufferEmpty(data)		(More_BufferEmpty((data)->inputBuffer))

#define InterestInReadability(data)	((data)->eventMask & TCL_READABLE)
#define InterestInWritability(data)	((data)->eventMask & TCL_WRITABLE)


/** ------------------------------------------------------------
 ** Debug macros.
 ** ----------------------------------------------------------*/

/* More_OpenBufferChannel() */

#define Debug_InspectOpenBuffer(inputBuffer, outputBuffer, modeMask)	\
  debug("buffer channel for: input %p, output %p, readable %s, writable %s", \
	inputBuffer, outputBuffer,					     \
	question(modeMask&TCL_READABLE), question(modeMask&TCL_WRITABLE));

#define Debug_OpenBufferResult(data)		\
		debug("created channel: name \"%s\", pointer \"%p\"", \
			ChannelName(data), data->channel);

/* BufchanInput() */

#define Debug_InputRequest(data, bufLen)	\
  debug("requested %d bytes through \"%s\", "\
	"blocking? %s, buffer empty? %s, buffer eof? %s",\
	bufLen, ChannelName(data), question(ChannelModeIsBlocking(data)),\
	question(InputBufferEmpty(data)), question(InputBufferEof(data)));

/* BufchanInputBlockForData() */

#define Debug_InputBlockResult(foundEvent, channelExists, isEmpty, notEof) \
  debug("out of loop: found event? %s, channel exists? %s, buffer empty? %s, eof? %s",\
	question(foundEvent), question(channelExists),			\
	question(isEmpty), question(!notEof));


/* BufchanWatch() */

#define Debug_EnterWatchFunction(data, eventMask)	\
	debug("watching \"%s\": readable %d, writable %d",		\
	      ChannelName(data),					\
	      (eventMask&TCL_READABLE), (eventMask&TCL_WRITABLE));

/* BufchanNotifyChannel() */

#define Debug_NotifyChannel(event)	\
  debug("notifying \"%s\": readable %s, writable %s, channel exists? %s", \
	event->channelName,						\
	question(event->eventMask & TCL_READABLE),			\
	question(event->eventMask & TCL_WRITABLE),			\
	question(Tcl_IsChannelExisting(event->channelName)));


/** ------------------------------------------------------------
 ** Function prototypes.
 ** ----------------------------------------------------------*/

/* Channel driver functions. */
static Tcl_DriverCloseProc	BufchanClose;
static Tcl_DriverInputProc	BufchanInput;
static Tcl_DriverOutputProc	BufchanOutput;
static Tcl_DriverWatchProc	BufchanWatch;
static Tcl_DriverGetHandleProc	BufchanGetHandle;
static Tcl_DriverBlockModeProc	BufchanBlockMode;
static Tcl_DriverSetOptionProc	BufchanSetOption;
static Tcl_DriverGetOptionProc	BufchanGetOption;

/* Functions used to notify events to the generic layer. */
static More_CallbackFunction BufchanBufferReadable;
static More_CallbackFunction BufchanBufferWritable;
static void BufchanQueueEvent _ANSI_ARGS_((Tcl_Channel channel, int eventMask));
static int  BufchanNotifyChannel _ANSI_ARGS_((Tcl_Event *D, int flags));

/* Option configuration functions. */
static More_ChannelDriverSetOptionProc SetOptionBufferSize;
static More_ChannelDriverGetOptionProc GetOptionBufferSize;


/** ------------------------------------------------------------
 ** Module's variables.
 ** ----------------------------------------------------------*/

static Tcl_ChannelType Driver = {
  "tclmore_bufchan",
  TCL_CHANNEL_VERSION_3,
  BufchanClose,
  BufchanInput,
  BufchanOutput,
  NULL, /* Seek */
  BufchanSetOption,
  BufchanGetOption,
  BufchanWatch,
  BufchanGetHandle,
  NULL, /* Close2 */
  BufchanBlockMode,
  NULL, /* Flush */
  NULL, /* Handler */
  NULL /* WideSeek */
};

static Tcl_Mutex	counterMutex = NULL;
static int		channelCounter = 0;

static CONST More_ChannelDriverOption optionTable[] = {
  {  "-inputBufferSize", SetOptionBufferSize, GetOptionBufferSize },
  { "-outputBufferSize", SetOptionBufferSize, GetOptionBufferSize },
  { NULL, NULL, NULL }
};


/* More_OpenBufferChannel --

	Opens a new channel interface for a couple of buffers.

	   --------  puts   ----------     
          | TCL    |------>| channel  |----->output buffer
          | interp |<------| instance |<-----input buffer
           --------  read   ---------- 

   Arguments:

	inputBuffer -	The token of the input buffer. Can be NULL if
			the channel is write-only.
	outputBuffer -	The token of the output buffer. Can be NULL if
			the channel is read-only.

   Results:

	Creates the channel linked  to the buffers.  Returns the channel
	token.

*/

Tcl_Channel
More_OpenBufferChannel (inputBuffer, outputBuffer)
     More_Buffer	inputBuffer;
     More_Buffer	outputBuffer;
{
  Channel *	data;
  char		channelName[CHANNEL_NAME_LENGTH];
  int		modeMask = 0;


  assert(inputBuffer || outputBuffer);

  if (inputBuffer)  { modeMask  = TCL_READABLE; }
  if (outputBuffer) { modeMask |= TCL_WRITABLE; }
  
  Debug_InspectOpenBuffer(inputBuffer, outputBuffer, modeMask);

  Tcl_MutexLock(&counterMutex);
  sprintf(channelName, CHANNEL_PATTERN, ++channelCounter);
  Tcl_MutexUnlock(&counterMutex);

  More_AllocStruct(data, Channel);
  data->inputBuffer	= inputBuffer;
  data->outputBuffer	= outputBuffer;
  data->openMode	= modeMask;
  data->channel = Tcl_CreateChannel(&Driver, channelName, data, modeMask);

  Debug_OpenBufferResult(data);
  return data->channel;
}

/* BufchanClose --

	Invoked  by  the  generic   layer  to  clean-up  driver  related
	informations when the channel is closed. 

   Arguments:

	D -		Pointer to the instance structure.
	interp -	If non-NULL: the interpreter used to report
			errors.

   Results:

	Must return zero  if the operation is successful,  or a non-zero
	POSIX error code if an error occurs; always returns zero.

*/

static int
BufchanClose (ClientData D, Tcl_Interp *interp)
{
  Channel *	data = (Channel *) D;


  debug("closing \"%s\"", ChannelName(data));
  if (data->inputBuffer)
    {
      More_BufferDetachReader(data->inputBuffer);
    }
  if (data->outputBuffer)
    {
      More_BufferDetachWriter(data->outputBuffer);
    }
  Tcl_MutexFinalize(&(data->mutex));
  ckfree((char *) data);
  return 0;
}

/* BufchanBlockMode --

	Invoked by  the generic layer  to set blocking and  non blocking
	mode for the channel.

   Arguments:

	D -		Pointer to the instance structure.
	mode -		TCL_MODE_BLOCKING or TCL_MODE_NONBLOCKING.

   Results:

	Registers the  new mode in the instance  structure.  Must return
        zero if the  operation is successful, or a  non-zero POSIX error
        code if an error occurs; always return zero.

*/

static int
BufchanBlockMode (ClientData D, int mode)
{
  Channel *	data = D;

  debug("updating blocking mode for \"%s\": blocking %s",
	ChannelName(data), question(mode == TCL_MODE_BLOCKING));
  data->blockingMode = mode;
  return 0;
}

/* BufchanInput --

	Invoked by  the generic layer to  read data from  the device and
	store it in an internal buffer.

   Arguments:

	D -		Pointer to the instance structure.
	bufPtr -	Pointer to the output buffer.
	bufLen -	Number of bytes to be read.
	errorCodePtr -	Pointer to a variable that must be filled with
			a POSIX error code in case of error.

   Results:

	Must  return a  non-negative integer  indicating how  many bytes
        were read. In case of error  must return -1.

	  If EOF  is found  on the input  channel, the EOF  condition is
	returned to the caller (no bytes read, no error).

	  If no data  is available and blocking mode  is off: the return
	value  is -1  and  the error  code  EAGAIN. This  may cause  the
	"resource temporarily unavailable"  error message to be returned
	(for example when a transformation is stacked on the channel).

	  If no data is available and blocking mode is on: this function
	will loop  until the input  buffer goes in  EOF or some  data is
	available;  for  a  single  threaded process,  this  will  block
	forever.

*/

static int
BufchanInput (ClientData D, char *bufPtr, int bufLen, int *errorCodePtr)
{
  Channel *	data = (Channel *) D;
  More_Block	block;
#ifdef TCLMORE_DEBUGGING
  int		read;
#endif

  
  Debug_InputRequest(data, bufLen);
  if (InputBufferEof(data))
    {
      goto End_Of_File;
    }
  else if (InputBufferEmpty(data))
    {
      if (ChannelModeIsBlocking(data))
	{
	  while (InputBufferEmpty(data))
	    {
	      if (InputBufferEof(data)) { goto End_Of_File; }
	    }
	}
      else /* ! ChannelModeIsBlocking(data) */
	{
	  debug("no bytes and non-blocking mode: returning EAGAIN");
	  *errorCodePtr = EAGAIN;
	  return -1;
	}
    }
  
  *errorCodePtr = 0;
  block.len = bufLen;
  block.ptr = bufPtr;

#ifdef TCLMORE_DEBUGGING
  read = More_BufferRead(data->inputBuffer, block);
  debug("read %d bytes from %s", read, Tcl_GetChannelName(data->channel));
  return read;
#else
  return More_BufferRead(data->inputBuffer, block);
#endif

 End_Of_File:
  debug("returning EOF");
  *errorCodePtr = 0;
  return 0;
}

/* BufchanOutput --

	Invoked by the  generic layer to transfer data  from an internal
	buffer to the output device.

   Arguments:

	D -		Pointer to the instance structure.
	bufPtr -	Pointer to the input buffer.
	bufLen -	Number of bytes to be written.
	errorCodePtr -	Pointer to a variable that must be filled with
			a POSIX error code in case of error.

   Results:

	Must  return a  non-negative integer  indicating how  many bytes
        were written; in case of error  must return -1. EPIPE is used as
        error code if the output buffer has no reader attached.

*/

static int
BufchanOutput (ClientData D, CONST char *bufPtr, int bufLen, int *errorCodePtr)
{
  Channel *	data = (Channel *) D;


  debug("request to write %d bytes through \"%s\" to buffer %p",
	bufLen, ChannelName(data), data->outputBuffer);


  if (More_BufferAlive(data->outputBuffer))
    {
      More_Block	block;

      block.len = bufLen;
      block.ptr = (char *)bufPtr;
      More_BufferWrite(data->outputBuffer, block);
      *errorCodePtr = 0;
      debug("done");
      return bufLen;
    }
  else
    {
      debug("no reader for output buffer, returning EPIPE");
      *errorCodePtr = EPIPE;
      return -1;
    }
}

/* BufchanWatch --

	Invoked by the generic layer to initialise the event notification
	mechanism.

   Arguments:

	D -		Pointer to the instance data.
	eventMask -	An OR-ed combination of TCL_READABLE, TCL_WRITABLE
			and TCL_EXCEPTION. It indicates events the caller
			is interested in.

   Results:

        Registers events of interest  in the data structure: only events
        compatible with the open mode are registered.

	  If  the readable  event  is requested:  the  callback for  the
	reader of  the input  buffer is registered;  this will  cause an
	event to  be queued in the  event loop of the  channel to notify
	the channel of readable events.

	  If the readable  event is NOT requested: the  callback for the
	reader  on the input  buffer, is  removed and  the state  of the
	input buffer requests reset.

	  If  the writable  event  is requested:  the  callback for  the
	writer of  the output buffer  is registered; this will  cause an
	event to  be queued in the  event loop of the  channel to notify
	the channel of  writable events. A buffer is  always writable if
	it has a reader attached.

	  If the writable  event is NOT requested: the  callback for the
	writer  on the ouput  buffer, is  removed and  the state  of the
	output buffer requests reset.

*/

static void
BufchanWatch (ClientData D, int eventMask)
{
  Channel *	data = (Channel *) D;
  More_Callback	callback;


  Debug_EnterWatchFunction(data, eventMask);

  data->eventMask = eventMask & data->openMode;
  if (ChannelModeIsReadable(data))
    {
      if (InterestInReadability(data))
	{
	  debug("registering reader callback");
	  More_CallbackInit(callback, BufchanBufferReadable,
			    (ClientData) data->channel);
	}
      else
	{
	  debug("reset reader callback");
	  More_CallbackReset(callback);
	}
      More_BufferReaderCallback(data->inputBuffer, callback);
    }
  if (ChannelModeIsWritable(data))
    {
      if (InterestInWritability(data))
	{
	  debug("registering writer callback");
	  More_CallbackInit(callback, BufchanBufferWritable,
			    (ClientData) data->channel);
	}
      else
	{
	  debug("reset writer callback");
	  More_CallbackReset(callback);
	}
      More_BufferWriterCallback(data->outputBuffer, callback);
    }
}

/* BufchanBufferReadable --

	Callback invoked by  the input buffer module to  signal that the
	input buffer is readable.

	  This callback must be  registered in the input buffer instance
	if the generic  layer notifies the driver that  it is interested
	in  readable events for  the channel.  The registration  must be
	carried out by the watch function.

	  This callback is invoked only  if the channel has not detached
	itself from the input buffer. So when the invocation happens the
	channel  exists;  it may  no  longer  exist  when the  event  is
	processed by the event loop.

	  This is the only  function testing "queuedReader" and the only
	function setting it to true.

   Arguments:

	D -		The channel token.

   Results:

        Queues an event in the channel's process or thread event loop to
        notify  the generic  layer that  the channel  is  readable.

	  This function  is invoked by the  buffer module, so  it may be
        invoked in a thread different to the one owning the channel.

*/

static void
BufchanBufferReadable (ClientData D)
{
  Tcl_Channel	channel = (Tcl_Channel) D;
  Channel *	data = (Channel *) Tcl_GetChannelInstanceData(channel);


  Tcl_MutexLock(&(data->mutex));
  if (! data->queuedReader)
    {
      BufchanQueueEvent((Tcl_Channel) D, TCL_READABLE);
      data->queuedReader = 1;
    }
  Tcl_MutexUnlock(&(data->mutex));
}

/* BufchanBufferWritable --

	Callback invoked by the output  buffer module to signal that the
	output buffer is writable.

	  This callback must be registered in the output buffer instance
	if the generic  layer notifies the driver that  it is interested
	in writable  events for the  channel.  The registration  must be
	carried out by the watch function.

	  This callback is invoked only  if the channel has not detached
	itself from  the output buffer.  So when the  invocation happens
	the channel  exists; it  may no longer  exist when the  event is
	processed by the event loop.

	  This is the only  function testing "queuedWriter" and the only
	function setting it to true.

   Arguments:

	D -		The channel token.

   Results:

        Queues an event in the channel's process or thread event loop to
        notify  the generic  layer that  the channel  is  writable.

	  This function  is invoked by the  buffer module, so  it may be
        invoked in a thread different to the one owning the channel.

*/

static void
BufchanBufferWritable (ClientData D)
{
  Tcl_Channel	channel = (Tcl_Channel) D;
  Channel *	data = (Channel *) Tcl_GetChannelInstanceData(channel);

  Tcl_MutexLock(&(data->mutex));
  if (! data->queuedWriter)
    {
      BufchanQueueEvent(channel, TCL_WRITABLE);
      data->queuedWriter = 1;
    }
  Tcl_MutexUnlock(&(data->mutex));
}

/* BufchanQueueEvent --

	Appends an  event to the queue  of the process  or thread owning
	the channel. The event will  be used to notify the generic layer
	about the readability or writability of the channel.

   Arguments:

	channel -	The channel token.
	eventMask -	The event mask representing events occurred on
			the channel.

   Results:

        Queues the event.

*/

static void
BufchanQueueEvent (Tcl_Channel channel, int eventMask)
{
  Event *	event;
#ifdef TCL_THREADS
  Tcl_ThreadId	thread;
#endif


  More_AllocStruct(event, Event);
  event->event.proc	= BufchanNotifyChannel;
  event->channel	= channel;
  event->eventMask	= eventMask;
  strcpy(event->channelName, Tcl_GetChannelName(channel));

#ifdef TCL_THREADS
  thread = Tcl_GetChannelThread(channel);
  debug("queuing event for \"%s\" (thread %p), readable %s, writable %s",
	event->channelName, thread,
	question(eventMask & TCL_READABLE), question(eventMask & TCL_WRITABLE));
  Tcl_ThreadQueueEvent(thread, (Tcl_Event *)event, TCL_QUEUE_TAIL);
  Tcl_ThreadAlert(thread);
#else
  debug("queuing event for \"%s\", readable %s, writable %s",
	event->channelName,
	question(eventMask & TCL_READABLE), question(eventMask & TCL_WRITABLE));
  Tcl_QueueEvent((Tcl_Event *)event, TCL_QUEUE_TAIL);
#endif
}

/* BufchanNotifyChannel --

	Invoked   by  the   event  loop   when  the   event   queued  by
	"BufchanQueueEvent()" is consumed: its  purpose is to notify the
	generic layer about channel events.

	  This  is  the only  function  that  resets "queuedReader"  and
	"queuedWriter" to zero.

   Arguments:

	D -		Pointer to the event structure.
	flags -		A mask representing events occurred on the
			channel; it is not the event mask representing
			events in which the generic layer has
			intereset, but is compatible with it.

   Results:

	The notification is performed  by Tcl_NotifyChannel: if an event
	handler matching the mask  has been registered with [fileevent]:
	it will be executed now, before returning from this call.

	  The notification is  not done if there is  no more interest in
	an event.

	  The generic layer is notified of the channel event only if the
	channel still exists when this function is invoked.

	  Returns  one to indicate  that the  notification event  can be
	removed from the event loop queue.

*/

static int
BufchanNotifyChannel (Tcl_Event *D, int flags)
{
  Event *	event = (Event *) D;


  Debug_NotifyChannel(event);
  if (Tcl_IsChannelExisting(event->channelName))
    {
      Channel *	data = (Channel *) Tcl_GetChannelInstanceData(event->channel);

      Tcl_MutexLock(&(data->mutex));
      if (event->eventMask & TCL_READABLE) { data->queuedReader = 0; }
      if (event->eventMask & TCL_WRITABLE) { data->queuedWriter = 0; }
      Tcl_MutexUnlock(&(data->mutex));

      /* Delete events for which there is no more interest. */
      event->eventMask &= (data->eventMask | TCL_READABLE);
      event->eventMask &= (data->eventMask | TCL_WRITABLE);

      if (event->eventMask)
	{
	  Tcl_NotifyChannel(event->channel, event->eventMask);
	}
    }
  return 1;
}

/* BufchanGetHandle --

	Invoked by the generic layer to retrieve the device-specific
	handler for the channel.

   Arguments:

	D -		Pointer to the instance data.
	direction -	The channel direction: TCL_READABLE or
			TCL_WRITABLE.
	handlePtr -	Pointer to the variable that will hold
			the handler.

   Results:

	Returns TCL_ERROR.

*/

static int
BufchanGetHandle (ClientData D, int direction, ClientData *handlePtr)
{
  return TCL_ERROR;
}

/* BufchanSetOption --

	Invoked  by the  generic layer  to set  a channel  type specific
	option to the channel instance.

	  Recognised options are:

	-inputBufferSize SIZE		selects the size of the input
					buffer;

	-outputBufferSize SIZE		selects the size of the output
					buffer.

   Arguments:

	D -		Pointer to the instance structure.
	interp -	If non-NULL: the interpreter to use to
			report errors.
	optionName -	The option name.
	newValue -	The option value.

   Results:

	Stores the  new value in  the instance data.  Returns  TCL_OK or
        TCL_ERROR.   In   case  of  error   leaves  a  message   in  the
        interpreter's result;  if the error is an  invalid option value,
        sets "errno" to EINVAL.

*/

static int
BufchanSetOption (D, interp, optionName, newValue)
     ClientData		D;
     Tcl_Interp *	interp;
     CONST char *	optionName;
     CONST char *	newValue;
{
  Channel *	data = D;

  return More_ChannelDriverSetOption(optionTable, data->channel,
				     D, interp, optionName, newValue);
}

/* SetOptionBufferSize --

	Configures a new buffer size.

   Arguments:

	D -		Pointer to the instance structure.
	interp -	If not NULL, the interpreter in which
			report errors.
	optionName -	Pointer to the option string.
	newValue -	Pointer to a string representing the
			new value.

   Results:

        Returns TCL_OK or TCL_ERROR.

*/

static int
SetOptionBufferSize (D, interp, optionName, newValue)
     ClientData		D;
     Tcl_Interp *	interp;
     CONST char *	optionName;
     CONST char *	newValue;
{
  size_t	bufferSize;
  More_Buffer	buffer;
  Channel *	data = D;
  Tcl_Obj *	valObj;
  int		e;


  valObj = Tcl_NewStringObj(newValue, -1);
  Tcl_IncrRefCount(valObj);
  e = More_GetSizeTFromObj(interp, valObj, &bufferSize);
  Tcl_DecrRefCount(valObj);
  if (TCL_OK != e)
    {
      Tcl_SetErrno(EINVAL);
      return TCL_ERROR;
    }

  buffer = (optionName[1] == 'i')? data->inputBuffer : data->outputBuffer;
  More_BufferSetSize(buffer, bufferSize);
  return TCL_OK;
}

/* BufchanGetOption --

	Invoked by the generic layer  to retrieve the value of a channel
	specific  option.   See   BufchanSetOption()  for  the  list  of
	recognised options.

   Arguments:

	D -		Pointer to the instance structure.
	interp -	If non-NULL: the interpreter used to report
			errors.
	optionName -	The requested option name, NULL if all the
			options are requested.
	optionValue -	Pointer to the output string.

   Results:

        Returns TCL_OK or TCL_ERROR.

*/

static int
BufchanGetOption (D, interp, optionName, optionValue)
     ClientData		D;
     Tcl_Interp *	interp;
     CONST char *	optionName;
     Tcl_DString *	optionValue;
{
  Channel *	data = D;
  return More_ChannelDriverGetOption(optionTable, data->channel,
				     D, interp, optionName, optionValue);
}

/* GetOptionBufferSize --

	Retrieves the current buffer size.

   Arguments:

	D -		Pointer to the instance structure.
	interp -	If non-NULL: the interpreter used to report
			errors.
	optionName -	The requested option name, NULL if all the
			options are requested.
	optionValue -	Pointer to the output string.

   Results:

        Returns TCL_OK or TCL_ERROR.

*/

static int
GetOptionBufferSize (D, interp, optionName, optionValue)
     ClientData		D;
     Tcl_Interp *	interp;
     CONST char *	optionName;
     Tcl_DString *	optionValue;
{
  size_t	bufferSize;
  Channel *	data = (Channel *) D;


  debug("option %s", optionName);
  /* "-inputBufferSize" or "-ouputBufferSize" */
  if (('i' == optionName[1]) && data->inputBuffer)
    {
      bufferSize = More_BufferGetSize(data->inputBuffer);
    }
  else if (('o' == optionName[1]) && data->outputBuffer)
    {
      bufferSize = More_BufferGetSize(data->outputBuffer);
    }
  else
    {
      bufferSize = 0;
    }
  More_DStringAppendSizeT(optionValue, bufferSize);
  return TCL_OK;
}

/* end of file */
/*
Local Variables:
mode: c
page-delimiter: "^$"
End:
*/
