/* buffervariable.c --
   
   Part of: tclmore
   Contents: buffered variable
   Date: Fri Jan 30, 2004
   
   Abstract
   
	This module defines  functions used to link a  TCL variable to a
	shared  buffer. This  interface  is used  in  the [varchan]  and
	[teechan] modules.
   
   Copyright (c) 2004 Marco Maggi
   
   The author  hereby grant permission to use,  copy, modify, distribute,
   and  license this  software  and its  documentation  for any  purpose,
   provided that  existing copyright notices  are retained in  all copies
   and that  this notice  is included verbatim  in any  distributions. No
   written agreement, license, or royalty  fee is required for any of the
   authorized uses.  Modifications to this software may be copyrighted by
   their authors and need not  follow the licensing terms described here,
   provided that the new terms are clearly indicated on the first page of
   each file where they apply.
   
   IN NO  EVENT SHALL THE AUTHOR  OR DISTRIBUTORS BE LIABLE  TO ANY PARTY
   FOR  DIRECT, INDIRECT, SPECIAL,  INCIDENTAL, OR  CONSEQUENTIAL DAMAGES
   ARISING OUT  OF THE  USE OF THIS  SOFTWARE, ITS DOCUMENTATION,  OR ANY
   DERIVATIVES  THEREOF, EVEN  IF THE  AUTHOR  HAVE BEEN  ADVISED OF  THE
   POSSIBILITY OF SUCH DAMAGE.
   
   THE  AUTHOR  AND DISTRIBUTORS  SPECIFICALLY  DISCLAIM ANY  WARRANTIES,
   INCLUDING,   BUT   NOT  LIMITED   TO,   THE   IMPLIED  WARRANTIES   OF
   MERCHANTABILITY,    FITNESS   FOR    A    PARTICULAR   PURPOSE,    AND
   NON-INFRINGEMENT.  THIS  SOFTWARE IS PROVIDED  ON AN "AS  IS" BASIS,
   AND  THE  AUTHOR  AND  DISTRIBUTORS  HAVE  NO  OBLIGATION  TO  PROVIDE
   MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
   
   $Id: buffervariable.c,v 1.1.1.10 2004/02/03 15:53:32 marco Exp marco $
*/


/** ------------------------------------------------------------
 ** Header files.
 ** ----------------------------------------------------------*/

#include <string.h>
#include <assert.h>

/*
#define TCLMORE_DEBUG_MESSAGES
#define TCLMORE_DEBUGGING	1
*/

#include "tclmoreInt.h"


/** ------------------------------------------------------------
 ** Type declarations.
 ** ----------------------------------------------------------*/

/* Variable --

  Type of client data for the variable trace. The operations are:

  * when  the variable  is written:  all the  data is  removed  from the
    variable and appended to the output buffer;

  * when  the variable  is read:  all the  data in  the input  buffer is
    extracted and appended to the variable's content.

    No callback is  registered in the shared buffers:  when the owner of
  the variable wants data it gets it. If one needs notifications, he can
  use a pipe channel.


  variable -		Variable's data.

  inputBuffer -		Pointer to the shared  input buffer, if NULL the
			variable is not traced for reads.

  outputBuffer -	Pointer to the shared output buffer, if NULL the
			variable is not traced for writes.

  trace_flags -		The trace flags.

*/

typedef struct Variable {
  More_Variable	variable;
  More_Buffer *	inputBuffer;
  More_Buffer *	outputBuffer;
  int		trace_flags;
} Variable;


/** ------------------------------------------------------------
 ** Module's function prototypes.
 ** ----------------------------------------------------------*/

static Tcl_VarTraceProc	BufferVariableTrace;
static char * BufferVariableRead	_ANSI_ARGS_((Variable *data));
static char * BufferVariableWrite	_ANSI_ARGS_((Variable *data));

static void BufferVariableFinalise		_ANSI_ARGS_((Variable *data));


/** ------------------------------------------------------------
 ** Preprocessor symbols.
 ** ----------------------------------------------------------*/

#define debug		More_Debug
#define question	More_Question

#define CHUNK_SIZE	4096


/* More_CreateBufferVariable --

	Links an existent variable to  a couple of buffers. The variable
	must be: existent, accessible, free of traces raising errors.

   Arguments:

	variable -	Pointer to a variable token.
	inputBuffer -	Pointer to the shared input buffer. Can be
			NULL.

	outputBuffer -	Pointer to the shared output buffer. Can be
			NULL.

   Results:

	Returns the  client data of the  trace: it can  be used together
	with the "variable" to remove the trace.

	  If data  is present in the  variable and the  output buffer is
	not NULL: it is immediately sent to the output buffer, if any.

	  If both the input and output buffers are NULL, nothing happens
	and the return value is NULL.

*/

ClientData
More_CreateBufferVariable (variable, inputBuffer, outputBuffer)
     More_Variable 	variable;
     More_Buffer	inputBuffer;
     More_Buffer	outputBuffer;
{
  Variable *	data = NULL;
  More_Block	block;
  int		trace_flags = TCL_TRACE_UNSETS;


  if ((! inputBuffer) && (! outputBuffer))
    {
      return data;
    }

  debug("creating buffer variable \"%s\"", More_VariableName(variable));

  data = (Variable *) ckalloc(sizeof(Variable));
  data->inputBuffer	= inputBuffer;
  data->outputBuffer	= outputBuffer;
  More_VariableCopy(data->variable, variable);

  if (outputBuffer)
    {
      block = More_VariableGetBlock(variable);
      if (block.len)
	{
	  More_BufferWrite(outputBuffer, block);
	  More_VariableClear(variable);
	}

      trace_flags |= TCL_TRACE_WRITES;
    }

  if (inputBuffer)
    {
      trace_flags |= TCL_TRACE_READS;
    }

  data->trace_flags = trace_flags;

  Tcl_TraceVar(variable->interp, More_VariableName(variable),
	       trace_flags, BufferVariableTrace, data);

  debug("buffer variable \"%s\" (%p)", More_VariableName(variable), data);
  return data;
}

/* More_DeleteBufferVariable --

	Deletes  the   link  between  a  variable   and  the  associated
	buffers. The variable itself is not touched.

   Arguments:

	variable -	Variable's data.
	D -		The client data of the variable trace.

   Results:

        Removes the  trace on  the variable, causing  the buffers  to be
        detached; this can cause the buffer deletion.

*/

void
More_DeleteBufferVariable (More_Variable variable, ClientData D)
{
  Variable *	data = (Variable *) D;


  debug("delete buffer variable \"%s\" (%p)",
	More_VariableName(data->variable), data);
  Tcl_UntraceVar(variable[0].interp, Tcl_GetString(variable[0].name),
		 data->trace_flags, BufferVariableTrace, D);
  BufferVariableFinalise(data);
}

/* BufferVariableTrace --

	Trace function linked to the variable.

	  If the  variable is unset:  the instance data is  released and
	the   variable  detached   from  both   the  input   and  output
	buffers. This will cause the other  side to be no longer able to
	write data  in its output buffer,  and able to  only consume the
	bytes already in its input buffer.

	  If the variable is read: all the bytes in the input buffer are
	read and appended to the  variable's content. The bytes are seen
	as a ByteArray.

	  If the  variable is  written: all the  bytes in  the variable,
	seen as  a ByteArray, are  extracted end appended to  the output
	buffer.

   Arguments:

	D -		Pointer to the variable instance data.
	interp -	The interpreter the variable belongs to.
	name1 -		Unused.
	name2 -		Unused.
	flags -		The events that caused the trace to be
			invoked.

   Results:

        Returns  NULL  or a  pointer  to  a  statically allocated  error
        message.

   Side effects:

	Reading from the  input buffer may cause the  buffer to queue an
	event to notify the source that the buffer is writable.  Writing
	to the output  buffer may cause the buffer to  queue an event to
	notify the sink that the buffer is readable.

*/

char *
BufferVariableTrace (D, interp, name1, name2, flags)
     ClientData		D;
     Tcl_Interp *	interp;
     CONST char *	name1;
     CONST char *	name2;
     int		flags;
{
  Variable *	data = (Variable *) D;
  char *	errorInfo = NULL;


  debug("trace on: \"%s\" (%p), unset %s, read %s, write %s",
	Tcl_GetString(data->variable[0].name), data,
	question(TCL_TRACE_UNSETS&flags),
	question(TCL_TRACE_READS&flags),
	question(TCL_TRACE_WRITES&flags));
  
  if ((TCL_TRACE_UNSETS & flags) || (TCL_INTERP_DESTROYED & flags))
    {
      BufferVariableFinalise(data);
    }
  else if (TCL_TRACE_READS & flags)
    {
      errorInfo = BufferVariableRead(data);
    }
  else if (TCL_TRACE_WRITES & flags)
    {
      errorInfo = BufferVariableWrite(data);
    }
  /*
    Here the trace may be unset, and "data" may point to invalid memory.
  */
  debug("end of trace");
  return errorInfo;
}

/* BufferVariableRead --

	Invoked to handle the read event on the traced variable.

   Arguments:

	data -		Pointer to the variable data.

   Results:

        Returns  NULL  or a  pointer  to  a  statically allocated  error
        message.

*/

static char *
BufferVariableRead (Variable *data)
{
  More_Block	block;


  debug("input buffer present? %s, buffer empty? %s, buffer eof? %s",
	question(data->inputBuffer),
	question((data->inputBuffer) && More_BufferEmpty(data->inputBuffer)),
	question((data->inputBuffer) && More_BufferEof(data->inputBuffer)));
  
  if (data->inputBuffer)
    {
      if (! More_BufferEmpty(data->inputBuffer))
	{
	  block = More_BufferReadAllBlock(data->inputBuffer);
	  debug("read %d bytes", block.len);
	  More_VariableAppendBlock(data->variable, block);
	  ckfree((char *) block.ptr);
	}
      else
	{
	  More_VariableClear(data->variable);
	  if (More_BufferEof(data->inputBuffer))
	    {
	      debug("detaching from buffer");
	      More_BufferDetachReader(data->inputBuffer);
	      data->inputBuffer = NULL;
	      if (! data->outputBuffer)
		{
		  More_DeleteBufferVariable(data->variable, data);
		}
	    }
	}
    }
  debug("end of read");
  return NULL;
}

/* BufferVariableWrite --

	Invoked to handle the write event on the traced variable.

   Arguments:

	data -		Pointer to the variable data.

   Results:

        Returns  NULL  or a  pointer  to  a  statically allocated  error
        message.

*/

static char *
BufferVariableWrite (Variable *data)
{
  char *	errorInfo = NULL;
  More_Block	output;
  Tcl_Obj *	objPtr;


  debug("output buffer present? %s, buffer alive? %s",
	question(data->outputBuffer),
	question((data->outputBuffer) && More_BufferAlive(data->outputBuffer)));

  if (! data->outputBuffer)
    {
      return NULL;
    }

  if (More_BufferAlive(data->outputBuffer))
    {
      objPtr = More_VariableGet(data->variable);
      output.ptr = Tcl_GetByteArrayFromObj(objPtr, &output.len);
      debug("writing %d bytes to output buffer", output.len);
      More_BufferWrite(data->outputBuffer, output);
      More_VariableClear(data->variable);
    }
  else
    {
      debug("detaching \"%s\" from output buffer",
	    More_VariableName(data->variable));
      More_BufferDetachWriter(data->outputBuffer);
      data->outputBuffer = NULL;
      if (! data->inputBuffer)
	{
	  More_DeleteBufferVariable(data->variable, data);
	}
    }
  return errorInfo;
}

/* BufferVariableFinalise --

	Finalises the data associated to the trace.

   Arguments:

	data -		Pointer to the instance structure.

   Results:

        Detaches the  variable from  the buffers. Releases  the instance
        structure.

*/

void
BufferVariableFinalise (Variable *data)
{
  if (data->inputBuffer)
    {
      debug("detach \"%s\" from input buffer %p",
	    More_VariableName(data->variable), data->inputBuffer);
      More_BufferDetachReader(data->inputBuffer);
    }
  if (data->outputBuffer)
    {
      debug("detach \"%s\" from output buffer %p",
	    More_VariableName(data->variable), data->outputBuffer);
      More_BufferDetachWriter(data->outputBuffer);
    }
  More_VariableFinal(data->variable);
  ckfree((char *) data);
}

		  
/* end of file */
/*
Local Variables:
mode: c
page-delimiter: "^$"
End:
*/
