/* channel.c --
   
   Part of: tclmore
   Contents: channel driver facilities
   Date: Thu Feb 26, 2004
   
   Abstract
   
   
   
   Copyright (c) 2004 Marco Maggi
   
   The author  hereby grant permission to use,  copy, modify, distribute,
   and  license this  software  and its  documentation  for any  purpose,
   provided that  existing copyright notices  are retained in  all copies
   and that  this notice  is included verbatim  in any  distributions. No
   written agreement, license, or royalty  fee is required for any of the
   authorized uses.  Modifications to this software may be copyrighted by
   their authors and need not  follow the licensing terms described here,
   provided that the new terms are clearly indicated on the first page of
   each file where they apply.
   
   IN NO  EVENT SHALL THE AUTHOR  OR DISTRIBUTORS BE LIABLE  TO ANY PARTY
   FOR  DIRECT, INDIRECT, SPECIAL,  INCIDENTAL, OR  CONSEQUENTIAL DAMAGES
   ARISING OUT  OF THE  USE OF THIS  SOFTWARE, ITS DOCUMENTATION,  OR ANY
   DERIVATIVES  THEREOF, EVEN  IF THE  AUTHOR  HAVE BEEN  ADVISED OF  THE
   POSSIBILITY OF SUCH DAMAGE.
   
   THE  AUTHOR  AND DISTRIBUTORS  SPECIFICALLY  DISCLAIM ANY  WARRANTIES,
   INCLUDING,   BUT   NOT  LIMITED   TO,   THE   IMPLIED  WARRANTIES   OF
   MERCHANTABILITY,    FITNESS   FOR    A    PARTICULAR   PURPOSE,    AND
   NON-INFRINGEMENT.  THIS  SOFTWARE IS PROVIDED  ON AN "AS  IS" BASIS,
   AND  THE  AUTHOR  AND  DISTRIBUTORS  HAVE  NO  OBLIGATION  TO  PROVIDE
   MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
   
*/


/** ------------------------------------------------------------
 ** Header files.
 ** ----------------------------------------------------------*/

#if 0
#  define TCLMORE_DEBUG_MESSAGES
#  define TCLMORE_DEBUGGING	1
#endif

#include "tclmoreInt.h"


/** ------------------------------------------------------------
 ** Preprocessor symbols.
 ** ----------------------------------------------------------*/

#define debug		More_Debug



/** ------------------------------------------------------------
 ** Type declarations.
 ** ----------------------------------------------------------*/

typedef More_ChannelDriverOption	Option;


/** ------------------------------------------------------------
 ** Module's function prototypes.
 ** ----------------------------------------------------------*/

static int SelectOption	_ANSI_ARGS_((CONST Option *table,
				     Tcl_Interp *interp,
				     CONST char *optionName,
				     int *indexVar));


/* More_ChannelDriverSetOption --

	Selects the  correct option from  the table and invokes  the set
	option function.

   Arguments:

	table -		Pointer to an array of structures describing
			the supported options. It is used as argument
			for Tcl_GetIndexFromObjStruct().
	channel -	The channel. It is used for transformation
			to handle unknown requests.
	D -		The channel instance data.
	interp -	If not NULL, the interpreter in which report
			errors.
	optionName -	Pointer to a string representing the option
			name.
	optionValue -	Pointer to a string representing the option
			value.

   Results:

        Returns the code returned by the set-function or TCL_ERROR if an
        unsupported option is selected. In case of error leaves an error
        in the interpreter.

*/

int
More_ChannelDriverSetOption (table, channel, D, interp, optionName, optionValue)
     CONST Option *	table;
     Tcl_Channel	channel;
     ClientData		D;
     Tcl_Interp *	interp;
     CONST char *	optionName;
     CONST char *	optionValue;
{
  int					e, index;
  More_ChannelDriverSetOptionProc *	proc;
  Tcl_Channel				subChannel;

  
  e = SelectOption(table, interp, optionName, &index);
      debug("e %d %d", e, (e == TCL_OK));
  if (TCL_OK != e)
    {
      subChannel = Tcl_GetStackedChannel(channel);
      if (subChannel)
	{
	  Tcl_DriverSetOptionProc *	setOptionProc = \
	    Tcl_ChannelSetOptionProc(Tcl_GetChannelType(subChannel));
	  if (setOptionProc)
	    {
	      if (interp) { Tcl_ResetResult(interp); }
	      return setOptionProc(Tcl_GetChannelInstanceData(subChannel),
				   interp, optionName, optionValue);
	    }
	  else
	    {
	      goto Error_Unsettable_Option;
	    }
	}
      else
	{
	  return e;
	}
    }
  debug("index %d", index);
  proc = table[index].setOptionProc;
  if (proc)
    {
      return ((TCL_OK == e)? proc(D, interp, optionName, optionValue) : e);
    }
  else
    {
      goto Error_Unsettable_Option;
    }

 Error_Unsettable_Option:
  if (interp)
    {
      Tcl_SetResult(interp, "this option cannot be set", TCL_STATIC);
      return More_LogicError(interp);
    }
  return TCL_ERROR;
}

/* More_ChannelDriverGetOption --

	Selects the  correct option from  the table and invokes  the get
	option function.

   Arguments:

	table -		Pointer to an array of structures describing
			supported options. It is used as argument
			for Tcl_GetIndexFromObjStruct().
	channel -	The channel. It is used for transformation
			to handle unknown requests.
	D -		The channel instance data.
	interp -	If not NULL, the interpreter in which report
			errors.
	optionName -	Pointer to a string representing the option
			name or NULL. In the second case the function
			will dump all the option/value pairs.
	optionValue -	Pointer to a dynamic string that must be
			filled with the option value or the
			option/value pairs.

   Results:

        Returns the code returned by the get-function or TCL_ERROR if an
        unsupported option is selected. In case of error leaves an error
        in the interpreter.

*/

int
More_ChannelDriverGetOption (table, channel, D, interp, optionName, optionValue)
     CONST Option *	table;
     Tcl_Channel	channel;
     ClientData		D;
     Tcl_Interp *	interp;	
     CONST char *	optionName;
     Tcl_DString *	optionValue;
{
  int					e, index;
  More_ChannelDriverGetOptionProc *	proc;
  Tcl_Channel				subChannel;
  Tcl_DriverGetOptionProc *		getOptionProc;

  
  if (! table)
    {
      if (interp)
	{
	  Tcl_SetResult(interp, "no options", TCL_STATIC);
	  return More_LogicError(interp);
	}
      return TCL_ERROR;
    }
  if (optionName)
    {
      e = SelectOption(table, interp, optionName, &index);
      if (TCL_OK != e)
	{
	  subChannel = Tcl_GetStackedChannel(channel);
	  if (subChannel)
	    {
	      getOptionProc = \
		Tcl_ChannelGetOptionProc(Tcl_GetChannelType(subChannel));
	      if (getOptionProc)
		{
		  if (interp) { Tcl_ResetResult(interp); }
		  return getOptionProc(Tcl_GetChannelInstanceData(subChannel),
				       interp, optionName, optionValue);
		}
	      else
		{
		  return e;
		}
	    }
	  else
	    {
	      return e;
	    }
	}

      proc = table[index].getOptionProc;
      if (proc)
	{
	  return ((TCL_OK == e)? proc(D, interp, optionName, optionValue) : e);
	}
      else
	{
	  Tcl_DStringAppend(optionValue, "{}", 2);
	  return TCL_OK;
	}
    }
  else
    {
      for (; table->name; ++table)
	{
	  debug("len %d options %s", Tcl_DStringLength(optionValue),
		Tcl_DStringValue(optionValue));
	  Tcl_DStringAppend(optionValue, " ", 1);
	  debug("option %s", table->name);
	  Tcl_DStringAppend(optionValue, table->name, -1);
	  Tcl_DStringAppend(optionValue, " ", 1);
	  debug("len %d options %s", Tcl_DStringLength(optionValue),
		Tcl_DStringValue(optionValue));

	  proc = table->getOptionProc;
	  if (proc)
	    {
	      e = proc(D, interp, table->name, optionValue);
	      if (TCL_OK != e)
		{
		  return e;
		}
	    }
	  else
	    {
	      Tcl_DStringAppend(optionValue, "{}", 2);
	    }
	  debug("options %s", Tcl_DStringValue(optionValue));
	}

      /*
	Propagate to the underlying channel.
      */
      subChannel = Tcl_GetStackedChannel(channel);
      if (subChannel)
	{
	  getOptionProc = \
	    Tcl_ChannelGetOptionProc(Tcl_GetChannelType(subChannel));
	  if (getOptionProc)
	    {
	      return getOptionProc(Tcl_GetChannelInstanceData(subChannel),
				   interp, optionName, optionValue);
	    }
	}
      return TCL_OK;
    }
}

/* SelectOption --

	Selects an option from the table.

   Arguments:

	table -		Pointer to an array of structures describing
			supported options. It is used as argument
			for Tcl_GetIndexFromObjStruct().
	interp -	If not NULL, the interpreter in which report
			errors.
	optionName -	Pointer to a string representing the option
			name.
	indexVar -	Pointer to the variable that will hold the
			option index in the table.

   Results:

        Returns TCL_OK or TCL_ERROR. In case of error leaves an error in
        the interpreter.

*/

int
SelectOption (table, interp, optionName, indexVar)
     CONST Option *	table;
     Tcl_Interp *	interp;
     CONST char *	optionName;
     int *		indexVar;
{
  int		e;
  Tcl_Obj *	objPtr;


  objPtr = Tcl_NewStringObj(optionName, -1);
  Tcl_IncrRefCount(objPtr);
  e = Tcl_GetIndexFromObjStruct(interp, objPtr, table, sizeof(Option),
				"option", TCL_EXACT, indexVar);
  Tcl_DecrRefCount(objPtr);
  return (TCL_OK != e)? ((interp)? More_LogicError(interp) : e) : e;
}


/* end of file */
/*
Local Variables:
mode: c
page-delimiter: "^$"
End:
*/
