/* commandinterface.c --
   
   Part of: tclmore
   Contents: interface to TCL commands
   Date: Tue Nov 11, 2003
   
   Abstract
   
	This  module defines  the functions  used to  interface internal
	code to TCL commands.
   
   Copyright (c) 2002, 2003, 2004 Marco Maggi
   
   The author  hereby grant permission to use,  copy, modify, distribute,
   and  license this  software  and its  documentation  for any  purpose,
   provided that  existing copyright notices  are retained in  all copies
   and that  this notice  is included verbatim  in any  distributions. No
   written agreement, license, or royalty  fee is required for any of the
   authorized uses.  Modifications to this software may be copyrighted by
   their authors and need not  follow the licensing terms described here,
   provided that the new terms are clearly indicated on the first page of
   each file where they apply.
   
   IN NO  EVENT SHALL THE AUTHOR  OR DISTRIBUTORS BE LIABLE  TO ANY PARTY
   FOR  DIRECT, INDIRECT, SPECIAL,  INCIDENTAL, OR  CONSEQUENTIAL DAMAGES
   ARISING OUT  OF THE  USE OF THIS  SOFTWARE, ITS DOCUMENTATION,  OR ANY
   DERIVATIVES  THEREOF, EVEN  IF THE  AUTHOR  HAVE BEEN  ADVISED OF  THE
   POSSIBILITY OF SUCH DAMAGE.
   
   THE  AUTHOR  AND DISTRIBUTORS  SPECIFICALLY  DISCLAIM ANY  WARRANTIES,
   INCLUDING,   BUT   NOT  LIMITED   TO,   THE   IMPLIED  WARRANTIES   OF
   MERCHANTABILITY,    FITNESS   FOR    A    PARTICULAR   PURPOSE,    AND
   NON-INFRINGEMENT.  THIS  SOFTWARE IS PROVIDED  ON AN "AS  IS" BASIS,
   AND  THE  AUTHOR  AND  DISTRIBUTORS  HAVE  NO  OBLIGATION  TO  PROVIDE
   MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
*/


/** ------------------------------------------------------------
 ** Header files.
 ** ----------------------------------------------------------*/

/*
#define TCLMORE_DEBUG_MESSAGES
#define TCLMORE_DEBUGGING	1
*/

#include "tclmoreInt.h"
#include <stddef.h>
#include <assert.h>

#define debug			More_Debug
#define question		More_Question


/** ------------------------------------------------------------
 ** Module's function prototypes.
 ** ----------------------------------------------------------*/

static More_Error	CreateNamespace	_ANSI_ARGS_((Tcl_Interp *interp,
						     CONST char *namespName));

static int FindCommandLeafNode	_ANSI_ARGS_((More_CommandTreeNode *rootNodes,
					     Tcl_Interp *interp,
					     int objc,
					     More_ObjStack objv,
					     More_CommandTreeNode **commandVar));
static int FindCommandBetweenNodes _ANSI_ARGS_((Tcl_Interp *interp,
						More_CommandTreeNode *nodesTable,
						Tcl_Obj *cmdObj,
						CONST char *commandDesc,
						More_CommandTreeNode **	commandVar));

static Tcl_ObjCmdProc	CommandProc;
static Tcl_Obj *	NamespaceTailCmd _ANSI_ARGS_((Tcl_Obj *cmdObj));

static int ParseArgs	_ANSI_ARGS_((More_CommandTreeLeaf *cmd,
				     More_CommandFrame cmdFrame));


/* More_CreateCommands --

	Creates a set of commands in a namespace.

   Arguments:

	interp -	The target interpreter.
	namespName -	Namespace name. Can be NULL if the target
			namespace is the root one.
	nodesTable -	Pointer to the declaration table: an array of
			structures; the last of them must have "name"
			field set to NULL.

   Results:

        Creates  the namespace  and the  commands; exports  the commands
        matching  the  pattern  "[a-z]*".   Returns  NULL  or  an  error
        descriptor.

*/

More_Error
More_CreateCommands (interp, namespName, nodesTable)
     Tcl_Interp *		interp;
     CONST char *		namespName;
     More_CommandTreeNode *	nodesTable;
{
  int				safeInterp, len;
  Tcl_Command			token;
  Tcl_DString			commandName;
  More_CommandTreeNode *	command;
  More_Error			error;


  safeInterp = Tcl_IsSafe(interp);
  debug("enter command creator, safe interpreter? %s", question(safeInterp));

  if (namespName)
    {
      error = CreateNamespace(interp, namespName);
      if (error) { return error; }
    }
  
  Tcl_DStringInit(&commandName);
  if (namespName)
    {
      Tcl_DStringAppend(&commandName, namespName, -1);
      Tcl_DStringAppend(&commandName, "::", -1);
    }
  len = Tcl_DStringLength(&commandName);

  for (command = nodesTable; NULL != command->name; ++command)
    {
      if (safeInterp && (! command->isSafeCommand)) {
	debug("skipping unsafe command %s", command->name);
	continue;
      }
      
      Tcl_DStringSetLength(&commandName, len);
      Tcl_DStringAppend(&commandName, command->name, -1);
      token = Tcl_CreateObjCommand(interp, Tcl_DStringValue(&commandName),
				   CommandProc, (ClientData) nodesTable, NULL);
      if (NULL == token)
	{
	  goto Error;
	}
    }
  Tcl_DStringFree(&commandName);

  debug("leave command creator");
  return NULL;

 Error:
  error = More_ErrorNew();
  More_ErrorRuntime(error, More_ObjPrintf("creating command \"%s\"",
					  Tcl_DStringValue(&commandName)));
  Tcl_DStringFree(&commandName);
  return error;
}

/* CreateNamespace --

	Creates a  new namespace and  exports all the  commands matching
	the pattern "[a-z]*".

   Arguments:

	interp -	Pointer to the target interpreter.
	namespName -	Pointer to a string holding the namespace name.

   Results:

        Returns NULL or an error descriptor.

*/

static More_Error
CreateNamespace (Tcl_Interp *interp, CONST char *namespName)
{
  static CONST char *	scriptTemplate = \
    "namespace eval %s { namespace export \\[a-z\\]* }";
  Tcl_Obj *		scriptObj;
  int			e;
  More_Error		error;


  scriptObj = More_ObjPrintf(scriptTemplate, namespName);
  Tcl_IncrRefCount(scriptObj);
  e = Tcl_EvalObjEx(interp, scriptObj, 0);
  Tcl_DecrRefCount(scriptObj);
  if (TCL_OK != e)
    {
      error = More_ErrorNew();
      More_ErrorRuntime(error, Tcl_GetObjResult(interp));
      Tcl_ResetResult(interp);
      return error;
    }
  return NULL;
}

/* CommandProc --

	Callback for a TCL command. Finds the declaration of the command
	in the  table of root command  nodes, the traverses  the tree of
	subcommands until the leaf  node is found; finally parse command
	line  arguments  and  invokes  the  user-defined  implementation
	function.

   Arguments:

	D -		Pointer to the table of root command nodes in
			which this command is declared.
	interp -	The interp to which the TCL command belongs.
	objc -		The number of parameters in the call.
	objv -		The call parameters.

   Results:

        Returns TCL_OK or TCL_ERROR. The return value of the TCL command
        is  the  one of  the  selected  subcommand  or of  the  selected
        implementation function.

*/

int
CommandProc (__TCL_COMMAND_ARGS__)
{
#undef  STATE_SIZE
#define STATE_SIZE (sizeof(int)*6)
  int				e;
  unsigned char			stateStruct[STATE_SIZE];
  unsigned char *		statePtr = stateStruct;
  More_CommandTreeNode *	rootNodes = D;
  More_CommandTreeNode *	command;
  More_CommandTreeLeaf *	definition;
  struct More_CommandFrame	frame = { NULL, D, interp, objc, objv };


  debug("enter command dispatcher, command: %s", Tcl_GetString(objv[0]));

  e = FindCommandLeafNode(rootNodes, interp, objc, objv, &command);
  if (TCL_ERROR == e)
    {
      return e;
    }

  definition = command->definition;
  if (definition->stateSize > STATE_SIZE)
    {
      debug("allocating state structure");
      statePtr = ckalloc(definition->stateSize);
    }
  if (definition->stateSize)
    {
      assert(definition->stateDefaults);
      memcpy(statePtr, definition->stateDefaults, definition->stateSize);
    }
  frame.stateStructPtr = statePtr;
  e = ParseArgs(definition, &frame);
  if (TCL_OK == e)
    {
      e = definition->function(&frame);
    }
  if (definition->stateSize > STATE_SIZE)
    {
      debug("freeing state structure");
      ckfree(statePtr);
    }
  return e;
}

/* FindCommandLeafNode --

	Traverses  a tree  of command  nodes looking  for the  leaf node
	selected by an array of objects.

   Arguments:

	nodesTable -	Pointer to the array of root nodes.
	interp -	Pointer to the interpreter in which report
			errors.
	objc -		The maximum number of objects in "objv".
	objv -		Array of objects seleting a command.
	commandVar -	Pointer to a variable that will hold a pointer
			to the selected command declaration structure.

   Results:

        Returns TCL_OK or  TCL_ERROR. In case of error  leaves a message
        in the interpreter and sets the error code to "LOGIC".

*/

static int
FindCommandLeafNode (nodesTable, interp, objc, objv, commandVar)
     More_CommandTreeNode *	nodesTable;
     Tcl_Interp *		interp;
     int			objc;
     More_ObjStack		objv;
     More_CommandTreeNode **	commandVar;     
{
  Tcl_Obj *			cmdObj;
  int				e, stackIndex, safeInterp;
  More_CommandTreeNode *	command;


  cmdObj = NamespaceTailCmd(objv[0]);
  Tcl_IncrRefCount(cmdObj);
  e = FindCommandBetweenNodes(interp, nodesTable, cmdObj, "command", &command);
  Tcl_DecrRefCount(cmdObj);
  if (TCL_OK != e)
    {
      return e;
    }

  debug("found root command: %s", command->name);

  safeInterp = Tcl_IsSafe(interp);

  stackIndex = 1;
  while (command->subcommands)
    {
      if (objc <= stackIndex)
	{
	  return More_WrongNumArgs(interp, stackIndex, objv,
				   "subcommand ?arg ...?");
	}

      nodesTable = command->subcommands;
      e = FindCommandBetweenNodes(interp, nodesTable, objv[stackIndex],
				  "subcommand", &command);
      if (TCL_OK != e)
	{
	  return e;
	}

      if (safeInterp && (! command->isSafeCommand))
	{
	  goto Error_Unsafe_Command;
	}
      ++stackIndex;
    }

  assert(command->definition);

  debug("found command, entry name \"%s\"", command->name);

  if (safeInterp && (! command->isSafeCommand))
    {
      goto Error_Unsafe_Command;
    }

  *commandVar = command;
  return TCL_OK;

 Error_Unsafe_Command:
  Tcl_SetResult(interp, "unsafe command in safe interpreter", TCL_STATIC);
  return More_RuntimeError(interp);
}

/* FindCommandBetweenNodes --

	Finds a  command declaration structure  in the array  of command
	nodes.

   Arguments:

	interp -		Pointer to the interpreter in which
				report errors.
	nodesTable -		Pointer to the table of nodes.
	cmdObj -		Pointer to the object holding the
				possibly fully qualified command
				name.
	commandDesc -		Pointer to a string describing the command,
				usually: "command" or "subcommand"; it is
				used to report "bad command" errors.
	commandVar -		Pointer to a variable that will hold a
				pointer to the command declaration.

   Results:

        Returns TCL_OK or TCL_ERROR.

*/

static int
FindCommandBetweenNodes (interp, nodesTable, cmdObj, commandDesc, commandVar)
     Tcl_Interp *		interp;
     More_CommandTreeNode *	nodesTable;
     Tcl_Obj *			cmdObj;
     CONST char *		commandDesc;
     More_CommandTreeNode **	commandVar;
{
  int	e, commandIndex;


  e = Tcl_GetIndexFromObjStruct(interp, cmdObj,
				nodesTable, sizeof(More_CommandTreeNode),
				commandDesc, TCL_EXACT, &commandIndex);
  if (TCL_OK != e)
    {
      return More_LogicError(interp);
    }

  *commandVar = &(nodesTable[commandIndex]);
  return e;
}

/* NamespaceTailCmd --

	This chunk of code comes from the TCL core, file "tclNamesp.c".

   Arguments:

	fullName -	Pointer to a string representing a possibly
			qualified command name.

   Results:

        Returns a pointer to the first char of the tail name.

*/

static Tcl_Obj *
NamespaceTailCmd(Tcl_Obj *cmdObj)
{
  CONST char *	fullName;
  CONST char *	ptr;


  fullName = Tcl_GetString(cmdObj);

  for (ptr = fullName; *ptr != '\0'; ptr++) {}
  while (--ptr > fullName)
    {
      if ((*ptr == ':') && (*(ptr-1) == ':'))
	{
	  ptr++;
	  break;
	}
    }
    
  return Tcl_NewStringObj(ptr, -1);
}

/* ParseOptions --

	Parses options  in the invocation  parameters of a  TCL command.
	Invokes the selected extractors for the values.

   Arguments:

	optionTable -	Pointer to the array of structures describing
			supported options. There must be at least "optnum"
			elements.

	D -		Used as	parameter "D" for the invocation of all the
			value extractors.

	args -		The command invocation parameters. The "objv"
			member must reference, as element zero, the
			first option to be parsed. The "objc" member
			must be the number of arguments to parse in
			"objv". The parameters received by the
			"Tcl_ObjCmdProc" function needs to be processed
			to be used here.

   Results:

        Returns "TCL_OK"  if all the arguments  are parsed successfully,
	or "TCL_ERROR" if  an error occurs; in the  latter case an error
	is left in the interpreter and the error code variable is set to
	"LOGIC". It's an error if a  value is missing for an option that
	requires  it.  It's  not  an  error if  no  options are  present
	(objc==0).

*/

static int
ParseOptions (optionTable, extractorFrame)
     CONST More_CommandOption *	optionTable;
     More_ExtractorFrame	extractorFrame;
{
  int			e, i, optionIndex;
  More_CommandFrame	commandFrame = extractorFrame->commandFrame;
  ptrdiff_t		fieldOffset;


  for (i=0; i<commandFrame->objc; ++i)
    {
      debug("option %s", Tcl_GetString(commandFrame->objv[i]));
      e = Tcl_GetIndexFromObjStruct(commandFrame->interp,
				    commandFrame->objv[i],
				    optionTable,
				    sizeof(More_CommandOption), "option",
				    TCL_EXACT, &optionIndex);
      if (TCL_OK != e)
	{
	  return More_LogicError(commandFrame->interp);
	}
      
      if (optionTable[optionIndex].hasValue)
	{
	  if (i+1 >= commandFrame->objc)
	    {
	      return More_OptionRequiresArg(commandFrame->interp,
					    Tcl_GetString(commandFrame->objv[i]));
	    }
	  ++i;
	}

      fieldOffset = optionTable[optionIndex].offsetInStateStruct;
      debug("option %s, offset %d",
	    optionTable[optionIndex].name, fieldOffset);
      extractorFrame->fieldPtr = (0 <= fieldOffset)? \
	commandFrame->stateStructPtr + fieldOffset : NULL;
      extractorFrame->srcObj   = commandFrame->objv[i];
      e = optionTable[optionIndex].extractor(extractorFrame);
      if (TCL_OK != e)
	{
	  return More_LogicError(commandFrame->interp);
	}
    }
  return TCL_OK;
}

/* ParseArgs --

	Parses arguments in the  invocation parameters of a TCL command.
	Invokes the selected extractors for the values.

   Arguments:

	command -	Pointer to a structure holding meta informations
			about a command.

	commandFrame -	Pointer to the command invocation frame.

   Results:

	Returns "TCL_OK"  if all the arguments are  parsed correctly, or
	"TCL_ERROR" if an error occurred; in the latter case an error is
	left in  the interpreter and the  error code variable  is set to
	"LOGIC". It's an error if the number of arguments is lesser than
	the required one ("objc < argnum").

	  After  all the  mandatory  arguments are  parsed, options  are
	parsed  with  "ParseOptions()"  according  to  the  informations
	accessed  through  the  "optionTable"  member of  the  structure
	referenced by "command".

*/

static int
ParseArgs (More_CommandTreeLeaf *command, More_CommandFrame commandFrame)
{
  int				e, i, j, objc;
  ptrdiff_t			fieldOffset;
  struct More_ExtractorFrame	extractorFrame;


  objc = command->stackOffsetOfFirstArgument + command->argumentNumber;

  if ((objc > commandFrame->objc) ||
      ((objc < commandFrame->objc) && (NULL == command->optionTable)))
    {
      return More_WrongNumArgs(commandFrame->interp,
			       command->stackOffsetOfFirstArgument,
			       commandFrame->objv,
			       command->synopsis);
    }

  extractorFrame.commandFrame = commandFrame;
  for (i=0, j=command->stackOffsetOfFirstArgument;
       i<command->argumentNumber;
       ++i, ++j)
    {
      extractorFrame.srcObj = commandFrame->objv[j];
      fieldOffset = command->argumentTable[i].offsetInStateStruct;
      extractorFrame.fieldPtr = (fieldOffset >= 0)? \
	commandFrame->stateStructPtr + fieldOffset : NULL;
      e = (command->argumentTable[i]).extractor(&extractorFrame);
      if (e != TCL_OK)
	{
	  return More_LogicError(commandFrame->interp);
	}
    }

  commandFrame->objc -= objc;
  commandFrame->objv += (ptrdiff_t) objc;
  return ((commandFrame->objc)? \
	  ParseOptions(command->optionTable, &extractorFrame) : TCL_OK);
}

/* end of file */
/*
Local Variables:
mode: c
page-delimiter: "^$"
End:
*/
