/* commands.c --
   
   Part of: tclmore
   Contents: commands definition
   Date: Fri Apr 26, 2002
   
   Abstract:
   
	Definition of the [do] and [loop] and the other commands.
   
   Copyright (c) 2002, 2003, 2004 Marco Maggi
   
   The author  hereby grant permission to use,  copy, modify, distribute,
   and  license this  software  and its  documentation  for any  purpose,
   provided that  existing copyright notices  are retained in  all copies
   and that  this notice  is included verbatim  in any  distributions. No
   written agreement, license, or royalty  fee is required for any of the
   authorized uses.  Modifications to this software may be copyrighted by
   their authors and need not  follow the licensing terms described here,
   provided that the new terms are clearly indicated on the first page of
   each file where they apply.
   
   IN NO  EVENT SHALL THE AUTHOR  OR DISTRIBUTORS BE LIABLE  TO ANY PARTY
   FOR  DIRECT, INDIRECT, SPECIAL,  INCIDENTAL, OR  CONSEQUENTIAL DAMAGES
   ARISING OUT  OF THE  USE OF THIS  SOFTWARE, ITS DOCUMENTATION,  OR ANY
   DERIVATIVES  THEREOF, EVEN  IF THE  AUTHOR  HAVE BEEN  ADVISED OF  THE
   POSSIBILITY OF SUCH DAMAGE.
   
   THE  AUTHOR  AND DISTRIBUTORS  SPECIFICALLY  DISCLAIM ANY  WARRANTIES,
   INCLUDING,   BUT   NOT  LIMITED   TO,   THE   IMPLIED  WARRANTIES   OF
   MERCHANTABILITY,    FITNESS   FOR    A    PARTICULAR   PURPOSE,    AND
   NON-INFRINGEMENT.  THIS  SOFTWARE IS PROVIDED  ON AN "AS  IS" BASIS,
   AND  THE  AUTHOR  AND  DISTRIBUTORS  HAVE  NO  OBLIGATION  TO  PROVIDE
   MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
*/


/** ------------------------------------------------------------
 ** Header files.
 ** ----------------------------------------------------------*/

#include "tclmoreInt.h"


/** ------------------------------------------------------------
 ** Preprocessor symbols.
 ** ----------------------------------------------------------*/

#define debug		More_Debug
#define question	More_Question

#define StateStruct(type)	 type *	state = frame->stateStructPtr


/** ------------------------------------------------------------
 ** Commands from XML document.
 ** ----------------------------------------------------------*/

#include "declaration.c"


int
MoreDoFunction (More_CommandFrame frame)
{
  StateStruct(DoState);
  int		e, boolean;
  char		infomsg[32 + TCL_INTEGER_SPACE];


  do
    {
      e = Tcl_EvalObjEx(frame->interp, state->body, 0);

      switch (e)
	{
	case TCL_OK:
	  break;

	case TCL_CONTINUE:
	  continue;

	case TCL_BREAK:
	  return TCL_OK;

	case TCL_ERROR:
	  sprintf(infomsg, "\n    (\"do\" body line %d)", frame->interp->errorLine);
	  Tcl_AddErrorInfo(frame->interp, infomsg);
	  return TCL_ERROR;

	default:
	  return e;
	}

      Tcl_ResetResult(frame->interp);

      e = Tcl_ExprBooleanObj(frame->interp, state->expr, &boolean);
      if (e != TCL_OK)
	{
	  return e;
	}
    }
  while (boolean);
  return TCL_OK;
}

int
MoreLoopFunction (More_CommandFrame frame)
{
  StateStruct(LoopState);
  int	e;
  char	infomsg[32 + TCL_INTEGER_SPACE];


  for (; state->times; --state->times)
    {
      e = Tcl_EvalObjEx(frame->interp, state->body, 0);
      switch (e)
	{
	case TCL_OK:
	case TCL_CONTINUE:
	  break;

	case TCL_BREAK:
	  return TCL_OK;

	case TCL_ERROR:
	  sprintf(infomsg, "\n    (\"loop\" body line %d)", frame->interp->errorLine);
	  Tcl_AddErrorInfo(frame->interp, infomsg);
	  return TCL_ERROR;

	default:
	  return e;
	}
    }

  return TCL_OK;
}

int
MoreVarchanFunction (More_CommandFrame frame)
{
  StateStruct(VarchanState);
  int		e;
  Tcl_Channel	channel;
  More_Variable	variable;

  
  debug("channel for variable \"%s\", mode: readable %s, writable %s",
	Tcl_GetString(state->varName),
	question(state->mode&TCL_READABLE), question(state->mode&TCL_WRITABLE));

  More_VariableInit(variable, frame->interp, state->varName);
  if (More_VariableExists(variable))
    {
      e = More_VariableTest(variable);
      if (TCL_OK != e)
	{
	  return More_LogicError(frame->interp);
	}
    }
  else
    {
      debug("creating variable");
      More_VariableSet(variable, Tcl_NewStringObj(NULL, 0));
    }

  channel = More_OpenVarChannel(variable, state->mode);
  Tcl_RegisterChannel(frame->interp, channel);
  Tcl_SetResult(frame->interp, (char *) Tcl_GetChannelName(channel), TCL_STATIC);

  More_VariableFinal(variable);
  debug("registered channel \"%s\"", Tcl_GetChannelName(channel));
  return TCL_OK;
}

int
MorePipechanFunction (More_CommandFrame frame)
{
  StateStruct(PipechanState);
  int		e;
  Tcl_Channel	channelOne,  channelTwo;
  More_Variable variableOne, variableTwo;

  
  More_VariableInit(variableOne, frame->interp, state->varOne);
  if (More_VariableExists(variableOne))
    {
      e = More_VariableTest(variableOne);
      if (TCL_OK != e)
	{
	  More_VariableFinal(variableOne);
	  return e;
	}
    }
  
  More_VariableInit(variableTwo, frame->interp, state->varTwo);
  if (More_VariableExists(variableTwo))
    {
      e = More_VariableTest(variableTwo);
      if (TCL_OK != e)
	{
	  More_VariableFinal(variableTwo);
	  return e;
	}
    }
  
  debug("opening pipe channel");

  channelOne = More_OpenPipeChannel(state->mode, &channelTwo);
  Tcl_RegisterChannel(frame->interp, channelOne);
  Tcl_RegisterChannel(frame->interp, channelTwo);

  More_VariableSet(variableOne,
		   Tcl_NewStringObj(Tcl_GetChannelName(channelOne), -1));
  More_VariableSet(variableTwo,
		   Tcl_NewStringObj(Tcl_GetChannelName(channelTwo), -1));

  More_VariableFinal(variableOne);
  More_VariableFinal(variableTwo);

  debug("registered channel one \"%s\"", Tcl_GetChannelName(channelOne));
  debug("registered channel two \"%s\"", Tcl_GetChannelName(channelTwo));
  return TCL_OK;
}

int
MoreTeechanFunction (More_CommandFrame frame)
{
  StateStruct(ChannelState);
  Tcl_Channel	channel;

  
  debug("request to stack channel on top of \"%s\"",
	Tcl_GetChannelName(state->channel.token));

  channel = More_StackTeeChannel(frame->interp,
				 state->channel.token, state->channel.modeMask);
  /*
    Calling Tcl_GetChannelName() with the stacked channel seems to crash
    TCL.
  */
  Tcl_SetResult(frame->interp,
		(char *) Tcl_GetChannelName(state->channel.token), TCL_STATIC);
  return TCL_OK;
}

int
MoreUnstackFunction (More_CommandFrame frame)
{
  StateStruct(ChannelState);
  return Tcl_UnstackChannel(frame->interp, state->channel.token);
}

int
GetObjFromArg (More_ExtractorFrame frame)
{
  return More_GetObjFromArg(frame);
}
int
GetUnsignedFromArg (More_ExtractorFrame frame)
{
  return More_GetUnsignedFromArg(frame);
}
int
GetChannelFromArg (More_ExtractorFrame frame)
{
  return More_GetChannelFromObj(frame->commandFrame->interp,
				frame->srcObj,
				&(((More_ChannelArg *) frame->fieldPtr)->token),
				&(((More_ChannelArg *) frame->fieldPtr)->modeMask));
}
int
GetOpenModeFromArg (More_ExtractorFrame frame)
{
  return More_GetOpenModeFromObj(frame->commandFrame->interp,
				 frame->srcObj, 0, (int *) frame->fieldPtr);
}


/* end of file */
/*
  Local Variables:
  page-delimiter: "^"
  End:
*/
