/*
  This file is meant to be included (with #include) in a source
  file provided by you that implements the command functions. Your
  file must include the TCL header ("tcl.h") first and define the
  functions later.

    The reason of this inclusion is that here are declared the struct
  typedefs of the command invocation states. These types must be
  visible by the command functions. The stuff declared in this file
  uses the "static" keyword as much as possible.

    The command functions should be small: they should do the job by
  invoking other functions, the ones that the extension should export
  in its stub table.
*/


/** ------------------------------------------------------------
 ** Header files.
 ** ----------------------------------------------------------*/

#include "tclmore.h"



/** ------------------------------------------------------------
 ** Type declarations.
 ** ----------------------------------------------------------*/

/* DoState --

  This structure is used to hold the invocation state of the
  following commands:

                    [more do   body expr]
*/
typedef struct DoState {
  Tcl_Obj *               body;
  Tcl_Obj *               expr;
} DoState;

static CONST DoState DoStateDefaults = {
  NULL, NULL
};

/* LoopState --

  This structure is used to hold the invocation state of the
  following commands:

                    [more loop   integet body]
*/
typedef struct LoopState {
  unsigned                times;
  Tcl_Obj *               body;
} LoopState;

static CONST LoopState LoopStateDefaults = {
  0, NULL
};

/* PipechanState --

  This structure is used to hold the invocation state of the
  following commands:

                    [more pipechan   chanVar1 chanVar2 mode]
*/
typedef struct PipechanState {
  Tcl_Obj *               varOne;
  Tcl_Obj *               varTwo;
  int                     mode;
} PipechanState;

static CONST PipechanState PipechanStateDefaults = {
  NULL, NULL, 0
};

/* ChannelState --

  This structure is used to hold the invocation state of the
  following commands:

                    [more teechan   channel]
                    [more unstack   channel]
*/
typedef struct ChannelState {
  More_ChannelArg         channel;
} ChannelState;

static CONST ChannelState ChannelStateDefaults = {
  MORE_CHANNEL_ARGUMENT_DEFAULT
};

/* VarchanState --

  This structure is used to hold the invocation state of the
  following commands:

                    [more varchan   varname mode]
*/
typedef struct VarchanState {
  Tcl_Obj *               varName;
  int                     mode;
} VarchanState;

static CONST VarchanState VarchanStateDefaults = {
  NULL, 0
};



/** ------------------------------------------------------------
 ** Implementation functions prototypes.
 ** ----------------------------------------------------------*/

/*
  You must provide the implementation of each of these functions in
  this file. When the function is invoked, the state invocation
  structure has been already filled with values from the command line
  arguments and options, with no errors.

    The function should look like the following:

  int
  MyCommandFunction (More_CommandFrame frame)
  {
     MyStateStruct *  state = (MyStateStruct *) frame->stateStructPtr;
     Tcl_Obj *        result;

     *
       Make use of the fields of "*state" to build a result by
       invoking functions from your library.
     *
     result = Mylib_DoSomething(state->value, state->another_value);

     Tcl_SetObjResult(frame->interp, result);
     return TCL_OK;
  }

    Being the implementation of a TCL command, this function is
  declared "static": it is useless to share this function among
  different modules. Remember that the role of a "More_CmdFunc"
  is to be an interface between the state invocation structure and
  the functions of your library that do the "real work".
*/

static More_CmdFunc         MoreDoFunction;
static More_CmdFunc         MoreLoopFunction;
static More_CmdFunc         MoreVarchanFunction;
static More_CmdFunc         MorePipechanFunction;
static More_CmdFunc         MoreTeechanFunction;
static More_CmdFunc         MoreUnstackFunction;


/** ------------------------------------------------------------
 ** Extractor functions prototypes.
 ** ----------------------------------------------------------*/

/*
  You must provide the implementation of each of the static extractors
  in this file. Functions declared EXTERN can be implemented in other
  modules of your extension or in other libraries that are dependencies
  of this one (for example the TCLMORE library implements some common
  extractors).

    The functions must be available as pointers: you cannot directly use
  functions exported in a stub table; this is because the pointer must
  be included in a statically allocated data structure. Remember that
  stub function names are available as macros, each macro references
  an element in a vector of function pointers in another library: the
  value of the function pointer is available only at execution time.

    The only way to use an extractor from a stub table is to declare
  another extractor (with a similar name) that wraps the real function.
  Example:

     More_Extractor MyExtractor;
     ...
     int
     MyExtractor (More_ExtractorFrame frame)
     {
       return RealExtractor(frame);
     }

  "RealExtractor" can be the macro of a stub function.

    If a prototype is omitted: it is your responsibility to declare it
  before the line that includes this file in your source module. Obviously
  the declaration can be in an external header file; this is particularly
  useful if your extension is using a wrapper extractor for a real extractor
  from another extension, and the extractor is used in more than one
  set of commands.

    A common extractor has the following structure, lets assume that
  "Type" is the type of value we want:

     EXTERN int Mylib_GetTypeFromObj _ANSI_ARGS_((Tcl_Interp * interp,
                                                  Tcl_Obj *    srcObj,
                                                  Type *       valueVar));
     ...
     int
     TypeExtractor (More_ExtractorFrame frame)
     {
        return Mylib_GetTypeFromObj(frame->commandFrame->interp,
                                    frame->srcObj,
                                    (Type *) frame->fieldPtr);       
     }

  or:

     int
     TypeExtractor (More_ExtractorFrame frame)
     {
        int   e;

        e = Mylib_GetTypeFromObj(frame->commandFrame->interp,
                                 frame->srcObj,
                                 (Type *) frame->fieldPtr);       
        return ((TCL_OK == e)? e : More_LogicError(interp));
     }

  the implementation of a "GetTypeFromObj" function to do the real
  work is recommended because such functions can be used by other
  modules, and even exported in the stub table.

    Remember that the field pointer can reference a whole structure
  embedded in the state invocation one.

    At present an extractor can validate arguments only with complete
  independence from the current process state, that is it can test that
  a value is in the correct space and range. Coherence tests between
  argument values and process state must be performed by the command
  implementation functions. This is not exactly true, but it is the
  recommended way to do things.
*/

static More_Extractor      GetObjFromArg;
static More_Extractor      GetUnsignedFromArg;
static More_Extractor      GetOpenModeFromArg;
static More_Extractor      GetChannelFromArg;


/** ------------------------------------------------------------
 ** Command arguments tables.
 ** ----------------------------------------------------------*/

#define more_do_ARGUMENT_NUMBER	2
static CONST More_CommandArgument MoreDoArguments[2] = {
  { GetObjFromArg,       offsetof(struct DoState, body) },
  { GetObjFromArg,       offsetof(struct DoState, expr) }
};

#define more_loop_ARGUMENT_NUMBER	2
static CONST More_CommandArgument MoreLoopArguments[2] = {
  { GetUnsignedFromArg,  offsetof(struct LoopState, times) },
  { GetObjFromArg,       offsetof(struct LoopState, body) }
};

#define more_varchan_ARGUMENT_NUMBER	2
static CONST More_CommandArgument MoreVarchanArguments[2] = {
  { GetObjFromArg,       offsetof(struct VarchanState, varName) },
  { GetOpenModeFromArg,  offsetof(struct VarchanState, mode) }
};

#define more_pipechan_ARGUMENT_NUMBER	3
static CONST More_CommandArgument MorePipechanArguments[3] = {
  { GetObjFromArg,       offsetof(struct PipechanState, varOne) },
  { GetObjFromArg,       offsetof(struct PipechanState, varTwo) },
  { GetOpenModeFromArg,  offsetof(struct PipechanState, mode) }
};

#define more_teechan_ARGUMENT_NUMBER	1
static CONST More_CommandArgument MoreTeechanArguments[1] = {
  { GetChannelFromArg,   offsetof(struct ChannelState, channel) }
};

#define more_unstack_ARGUMENT_NUMBER	1
static CONST More_CommandArgument MoreUnstackArguments[1] = {
  { GetChannelFromArg,   offsetof(struct ChannelState, channel) }
};



/** ------------------------------------------------------------
 ** Command options tables.
 ** ----------------------------------------------------------*/

/* No options for [more do]. */

/* No options for [more loop]. */

/* No options for [more varchan]. */

/* No options for [more pipechan]. */

/* No options for [more teechan]. */

/* No options for [more unstack]. */


/** ------------------------------------------------------------
 ** Leaf commands array.
 ** ----------------------------------------------------------*/

/*
  There is a leaf structure for each "More_CmdFunc". Each structure
  has references to the argument and option tables. It is not your
  business to mess with this structure.
*/

static More_CommandTreeLeaf commandTreeLeaves[] = {
  /* (0) synopsis: more do body expr */
  { 2, MoreDoFunction, sizeof(DoState), (CONST ClientData)&DoStateDefaults,
    more_do_ARGUMENT_NUMBER, MoreDoArguments, NULL,
    "body expr" },

  /* (1) synopsis: more loop integet body */
  { 2, MoreLoopFunction, sizeof(LoopState), (CONST ClientData)&LoopStateDefaults,
    more_loop_ARGUMENT_NUMBER, MoreLoopArguments, NULL,
    "integet body" },

  /* (2) synopsis: more varchan varname mode */
  { 2, MoreVarchanFunction, sizeof(VarchanState), (CONST ClientData)&VarchanStateDefaults,
    more_varchan_ARGUMENT_NUMBER, MoreVarchanArguments, NULL,
    "varname mode" },

  /* (3) synopsis: more pipechan chanVar1 chanVar2 mode */
  { 2, MorePipechanFunction, sizeof(PipechanState), (CONST ClientData)&PipechanStateDefaults,
    more_pipechan_ARGUMENT_NUMBER, MorePipechanArguments, NULL,
    "chanVar1 chanVar2 mode" },

  /* (4) synopsis: more teechan channel */
  { 2, MoreTeechanFunction, sizeof(ChannelState), (CONST ClientData)&ChannelStateDefaults,
    more_teechan_ARGUMENT_NUMBER, MoreTeechanArguments, NULL,
    "channel" },

  /* (5) synopsis: more unstack channel */
  { 2, MoreUnstackFunction, sizeof(ChannelState), (CONST ClientData)&ChannelStateDefaults,
    more_unstack_ARGUMENT_NUMBER, MoreUnstackArguments, NULL,
    "channel" }
};


/** ------------------------------------------------------------
 ** Node commands array.
 ** ----------------------------------------------------------*/

/*
  These are node structures that declare subcommands with subcommands.
  It is not your business to mess with this structure.
*/

static More_CommandTreeNode MoreCommandNode[] = {
  { "do", TCLMORE_SAFE_COMMAND,  NULL, &commandTreeLeaves[0] },
  { "loop", TCLMORE_SAFE_COMMAND,  NULL, &commandTreeLeaves[1] },
  { "varchan", TCLMORE_SAFE_COMMAND,  NULL, &commandTreeLeaves[2] },
  { "pipechan", TCLMORE_SAFE_COMMAND,  NULL, &commandTreeLeaves[3] },
  { "teechan", TCLMORE_SAFE_COMMAND,  NULL, &commandTreeLeaves[4] },
  { "unstack", TCLMORE_SAFE_COMMAND,  NULL, &commandTreeLeaves[5] },
  { NULL, TCLMORE_UNSAFE_COMMAND, NULL, NULL }
};



/** ------------------------------------------------------------
 ** Root commands table.
 ** ----------------------------------------------------------*/

/*
  This array of structures declares the root commands. There is one of
  them for each <maincommand> and <command> markups with no ancestors
  in the original XML document. It is not your business to mess with
  the content of this array.

    You may create commands declared in this array with the following
  chunk of code:

    error = More_CreateCommands(interp, "namesp", More_Commands);
    if (error)
      {
	return More_ErrorResult(interp, error);
      }

  where "namesp" is the name of the namespace in which the command
  must be created. If the namespace is the root one: you can use NULL.
*/

More_CommandTreeNode More_Commands[] = {
  { "more", TCLMORE_SAFE_COMMAND, MoreCommandNode, NULL },
  { NULL, 0, NULL, NULL }
};


/* end of file */
/*
Local Variables:
mode: c
page-delimiter: "^$"
End:
*/

