/*  delayedscript.c --
   
   Part of: tclmore
   Contents: delayed execution of script in the event loop
   Date: Tue Mar 16, 2004
   
   Abstract
   
   
   
   Copyright (c) 2004 Marco Maggi
   
   The author  hereby grant permission to use,  copy, modify, distribute,
   and  license this  software  and its  documentation  for any  purpose,
   provided that  existing copyright notices  are retained in  all copies
   and that  this notice  is included verbatim  in any  distributions. No
   written agreement, license, or royalty  fee is required for any of the
   authorized uses.  Modifications to this software may be copyrighted by
   their authors and need not  follow the licensing terms described here,
   provided that the new terms are clearly indicated on the first page of
   each file where they apply.
   
   IN NO  EVENT SHALL THE AUTHOR  OR DISTRIBUTORS BE LIABLE  TO ANY PARTY
   FOR  DIRECT, INDIRECT, SPECIAL,  INCIDENTAL, OR  CONSEQUENTIAL DAMAGES
   ARISING OUT  OF THE  USE OF THIS  SOFTWARE, ITS DOCUMENTATION,  OR ANY
   DERIVATIVES  THEREOF, EVEN  IF THE  AUTHOR  HAVE BEEN  ADVISED OF  THE
   POSSIBILITY OF SUCH DAMAGE.
   
   THE  AUTHOR  AND DISTRIBUTORS  SPECIFICALLY  DISCLAIM ANY  WARRANTIES,
   INCLUDING,   BUT   NOT  LIMITED   TO,   THE   IMPLIED  WARRANTIES   OF
   MERCHANTABILITY,    FITNESS   FOR    A    PARTICULAR   PURPOSE,    AND
   NON-INFRINGEMENT.  THIS  SOFTWARE IS PROVIDED  ON AN "AS  IS" BASIS,
   AND  THE  AUTHOR  AND  DISTRIBUTORS  HAVE  NO  OBLIGATION  TO  PROVIDE
   MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
   
*/


/** ------------------------------------------------------------
 ** Header files.
 ** ----------------------------------------------------------*/

#if 0
#  define TCLMORE_DEBUGGING		1
#  define TCLMORE_DEBUG_MESSAGES	1
#endif

#include "tclmoreInt.h"


/** ------------------------------------------------------------
 ** Module's type declarations.
 ** ----------------------------------------------------------*/

/* DelayedScript --

  Type used to define  a script to be evaluated in the  event loop of an
  interpreter, possibly with additional arguments appended.

  interp -		Pointer to the interpreter in which the script
			will be evaluated.

  body -		The body of the script.

  timer -		The token of the timer used to queue the script.
			We use a timer rather than an event directly
			queued because, with the token, we can remove
			it from the queue if the subject of the script
			is detroyed.

  callback -		Callback invoked to notify the originating
			module that the event has been consumed.

*/

typedef struct DelayedScript {
  Tcl_Interp *		interp;
  Tcl_Obj *		body;
  Tcl_TimerToken	timer;
  More_Callback		callback;
} DelayedScript;


/** ------------------------------------------------------------
 ** Preprocessor symbols.
 ** ----------------------------------------------------------*/

#define TimerIsQueued(script)		((script)->timer)

#define debug			More_Debug
#define question		More_Question


/** ------------------------------------------------------------
 ** Module's prototypes.
 ** ----------------------------------------------------------*/

static void	EvalScript	_ANSI_ARGS_((DelayedScript *script));



/* More_DScriptInit --

	Allocates and  initialises a new  delayed script. The  script is
	not queued.

	  The  caller  may  keep   a  reference  to  the  instance  with
	"Tcl_Preserve()".

	  The target interpreter is preserved because when the script is
	evaluated it may no longer exist: we need a way to test if it is
	still there or not, so we use "Tcl_InterpDeleted()".

   Arguments:

	interp -	Pointer to the interpreter in which
			the script will be evaluated.
	body -		Pointer to the object holding the
			script body.

   Results:

        Returns a reference to the structure.

*/

More_DScript
More_DScriptInit (Tcl_Interp *interp, Tcl_Obj *body)
{
  DelayedScript *	script;


  debug("creating script: %s", Tcl_GetString(body));
  Tcl_Preserve(interp);
  Tcl_IncrRefCount(body);
  More_AllocStruct(script, DelayedScript);
  script->interp	= interp;
  script->body		= body;
  return (More_DScript) script;  
}

/* More_DScriptCopy --

	Duplicates an  instance. This function  is useful if we  need to
	queue a script multiple times.

   Arguments:

	token -		Original instance's token.

   Results:

        Returns the token of the new instance.

*/

More_DScript
More_DScriptCopy (More_DScript token)
{
  DelayedScript	*	script = (DelayedScript *) token;
  DelayedScript	*	copy;
  

  debug("duplicating script %p", script);
  copy = (DelayedScript *) ckalloc(sizeof(DelayedScript));  
  *copy = *script;
  Tcl_Preserve(copy->interp);
  Tcl_IncrRefCount(copy->body);
  copy->timer = NULL;
  debug("copy (%p): %s", copy, Tcl_GetString(copy->body));
  return (More_DScript) copy;
}

/* More_DScriptFinal --

	Finalises an instance. This function may be used:

	* to release an instance that is not queued;

	* to abort a queued script;

	* to free the resources after the timer event has been consumed.

   Arguments:

	token -		Instance's token.

   Results:

        Releases all the resources.

*/

void
More_DScriptFinal (More_DScript token)
{
  DelayedScript	*	script = (DelayedScript *) token;

  
  debug("finalising script (%p): %s", script, Tcl_GetString(script->body));
  if (TimerIsQueued(script))
    {
      Tcl_DeleteTimerHandler(script->timer);
    }
  Tcl_DecrRefCount(script->body);
  Tcl_Release((ClientData) script->interp);
  ckfree((char *) script);
  debug("script finalised");
}

/* More_DScriptQueue --

	Queues the script in the event loop.

   Arguments:

	token -		Instance's token.
	args -		Additional arguments to be appended
			to the script.

   Results:

        Appends the arguments to the script body the Returns ...

*/

void
More_DScriptQueue (More_DScript token, More_Objects args)
{
  DelayedScript	*	script = (DelayedScript *) token;  
  int			i;


  debug("queuing script: (%p) %s", script, Tcl_GetString(script->body));
  if (args.objc && Tcl_IsShared(script->body))
    {
      script->body = Tcl_DuplicateObj(script->body);
      Tcl_IncrRefCount(script->body);
    }
  for (i=0; i<args.objc; ++i)
    {
      Tcl_ListObjAppendElement(NULL, script->body, args.objv[i]);
    }
  script->timer = Tcl_CreateTimerHandler(1, (Tcl_TimerProc *) EvalScript,
					 (ClientData) token);
  debug("done");
}

/* More_DScriptCallback --

	Registers a  callback to  be invoked after  the script  has been
	consumed.

   Arguments:

	token -		Instance's token.
	callback -	The callback.

   Results:

        Updates the instance context.

*/

void
More_DScriptCallback (More_DScript token, More_Callback callback)
{
  DelayedScript	*	script = (DelayedScript *) token;
  script->callback = callback;
}

/* EvalScript --

	Callback  procedure invoked  by the  event loop  when  the timer
	event  is   consumed.   Evaluates  the  script   in  the  global
	namespace.

   Arguments:

	script -	Instance's token.

   Results:

        Ignore   the  errors   and  eventually   frees   the  instance's
        structure.

          The "timer"  field is reset  here because the  original module
        that  queued  the  script  may  have kept  a  reference  to  the
        structure (preserved with "Tcl_Preserve()"):  in this case we do
        not  want "More_DScriptFinal()"  to  try to  release an  already
        consumed   timer  event   when  the   original   module  invokes
        "Tcl_Release()" on this instance.

*/

static void
EvalScript (DelayedScript *script)
{
  debug("evaluating: (%p) %s", script, Tcl_GetString(script->body));
  if (! Tcl_InterpDeleted(script->interp))
    {
      Tcl_GlobalEvalObj(script->interp, script->body);
    }
  script->timer = (Tcl_TimerToken) NULL;
  if (More_CallbackPresent(script->callback))
    {
      More_CallbackInvoke(script->callback);
    }
  Tcl_EventuallyFree((ClientData) script, (Tcl_FreeProc *) More_DScriptFinal);
  debug("script evaluated");
}

/* end of file */
/*
Local Variables:
mode: c
page-delimiter: "^$"
End:
*/
