/* error.c --
   
   Part of: tclmore
   Contents: error handling functions
   Date: Tue Nov 11, 2003
   
   Abstract
   
	Declares error handling functions.   
   
   Copyright (c) 2002, 2003 Marco Maggi
   
   The author  hereby grant permission to use,  copy, modify, distribute,
   and  license this  software  and its  documentation  for any  purpose,
   provided that  existing copyright notices  are retained in  all copies
   and that  this notice  is included verbatim  in any  distributions. No
   written agreement, license, or royalty  fee is required for any of the
   authorized uses.  Modifications to this software may be copyrighted by
   their authors and need not  follow the licensing terms described here,
   provided that the new terms are clearly indicated on the first page of
   each file where they apply.
   
   IN NO  EVENT SHALL THE AUTHOR  OR DISTRIBUTORS BE LIABLE  TO ANY PARTY
   FOR  DIRECT, INDIRECT, SPECIAL,  INCIDENTAL, OR  CONSEQUENTIAL DAMAGES
   ARISING OUT  OF THE  USE OF THIS  SOFTWARE, ITS DOCUMENTATION,  OR ANY
   DERIVATIVES  THEREOF, EVEN  IF THE  AUTHOR  HAVE BEEN  ADVISED OF  THE
   POSSIBILITY OF SUCH DAMAGE.
   
   THE  AUTHOR  AND DISTRIBUTORS  SPECIFICALLY  DISCLAIM ANY  WARRANTIES,
   INCLUDING,   BUT   NOT  LIMITED   TO,   THE   IMPLIED  WARRANTIES   OF
   MERCHANTABILITY,    FITNESS   FOR    A    PARTICULAR   PURPOSE,    AND
   NON-INFRINGEMENT.  THIS  SOFTWARE IS PROVIDED  ON AN "AS  IS" BASIS,
   AND  THE  AUTHOR  AND  DISTRIBUTORS  HAVE  NO  OBLIGATION  TO  PROVIDE
   MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
*/


/** ------------------------------------------------------------
 ** Header files.
 ** ----------------------------------------------------------*/

#include "tclmoreInt.h"


/** ------------------------------------------------------------
 ** Module's variables.
 ** ----------------------------------------------------------*/

static CONST char *	LOGIC	= "LOGIC";
static CONST char *	RUNTIME	= "RUNTIME";


/** ------------------------------------------------------------
 ** Module's function prototypes.
 ** ----------------------------------------------------------*/

static void FillDescriptor _ANSI_ARGS_((More_Error error,
					Tcl_Obj *errorInfo, int type));


/* More_ErrorNew --

	Instantiates a new error descriptor.

   Arguments:

	None.

   Results:

        Returns the error descriptor handler.

*/

More_Error
More_ErrorNew (void)
{
  More_Error	error;

  error = (More_Error) ckalloc(sizeof(struct More_ErrorTag));
  memset(error, '\0', sizeof(struct More_ErrorTag));
  return error;
}

/* More_ErrorForget --

	Releases all the resources associated to an error descriptor.

   Arguments:

	error -		The error descriptor.

   Results:

	Frees the descriptor.

   Side effects:

        None.

*/

void
More_ErrorForget (More_Error error)
{
  Tcl_DecrRefCount(error->errorInfo);
  ckfree((char *) error);
}

/* More_ErrorResult --

	Stores the  error informations in the  interpreters result, then
	frees the error descriptor.

   Arguments:

	interp -	If not NULL, the interpreter in which store
			the result.
	error -		The error descriptor.

   Results:

        Sets new values for the error code and error info variables.

*/

int
More_ErrorResult (Tcl_Interp *interp, More_Error error)
{
  if (interp)
    {
      Tcl_ResetResult(interp);
      Tcl_SetErrorCode(interp, error->errorCode, NULL);
      Tcl_SetObjResult(interp, error->errorInfo);
    }
  More_ErrorForget(error);
  return TCL_ERROR;
}

/* More_ErrorLogic --

	Configures an error descriptor  to represent a logic error.

   Arguments:

	error -		Error descriptor.
	errorInfo -	Error information.

   Results:

        Fills the descriptor.

*/

void
More_ErrorLogic (More_Error error, Tcl_Obj *errorInfo)
{
  FillDescriptor(error, errorInfo, 0);
}

/* More_ErrorRuntime --

	Configures an error descriptor  to represent a runtime error.

   Arguments:

	error -		Error descriptor.
	errorInfo -	Error information.

   Results:

        Fills the descriptor.

*/

void
More_ErrorRuntime (More_Error error, Tcl_Obj *errorInfo)
{
  FillDescriptor(error, errorInfo, 1);
}

/* FillDescriptor --

	Fills the members of a descriptor.

   Arguments:

	error -		The descriptor.
	errorInfo -	Pointer to the info object.
	type -		False for logic error, true for run
			time error.

   Results:

        Returns 

   Side effects:

        None.

*/

static void
FillDescriptor (More_Error error, Tcl_Obj *errorInfo, int type)
{
  More_ErrorPrepend(error, errorInfo);
  error->errorCode = type? RUNTIME : LOGIC;
}

/* More_ErrorPrepend --

	Prepends  a string  to the  error information  in  a descriptor,
	separating the two with a colon and a space.

   Arguments:

	error -		The descriptor.
	errorInfo -	Pointer to the info object.

   Results:

        Updates the descriptor.

*/

void
More_ErrorPrepend (More_Error error, Tcl_Obj *errorInfo)
{
  if (error->errorInfo)
    {
      if (Tcl_IsShared(errorInfo))
	{
	  errorInfo = Tcl_DuplicateObj(errorInfo);
	}
      Tcl_IncrRefCount(errorInfo);
      Tcl_AppendToObj(errorInfo, ": ", -1);
      Tcl_AppendObjToObj(errorInfo, error->errorInfo);
      Tcl_DecrRefCount(error->errorInfo);
    }
  else
    {
      Tcl_IncrRefCount(errorInfo);
    }
  error->errorInfo = errorInfo;
}

/* More_ErrorLogicStr --

	Configures an error descriptor to signal a logic error.

   Arguments:

	error -		Error descriptor.
	errorInfo -	Error information.

   Results:

        Updates the descriptor.

*/

void
More_ErrorLogicStr (More_Error error, CONST char *errorInfo)
{
  More_ErrorLogic(error, Tcl_NewStringObj(errorInfo, -1));
}

/* More_ErrorRuntimeStr --

	Configures an error descriptor to signal a runtime error.

   Arguments:

	error -		Error descriptor.
	errorInfo -	Error information.

   Results:

	Updates the descriptor.

*/

void
More_ErrorRuntimeStr (More_Error error, CONST char *errorInfo)
{
  More_ErrorRuntime(error, Tcl_NewStringObj(errorInfo, -1));
}

/* More_ErrorPrependStr --

	Prepends  a string  to the  error information  in  a descriptor,
	separating the two with a colon and a space.

   Arguments:

	error -		Error descriptor.
	errorInfo -	Error information.

   Results:

	Updates the descriptor.

*/

void
More_ErrorPrependStr (More_Error error, CONST char *errorInfo)
{
  return More_ErrorPrepend(error, Tcl_NewStringObj(errorInfo, -1));
}

/* More_ErrorIsLogic --

	Tests if the error code is logic.

   Arguments:

	error -		The error descriptor.

   Results:

        Returns true if the error is logic, else returns false.

*/

int
More_ErrorIsLogic (More_Error error)
{
  return (LOGIC == error->errorCode);
}

/* More_ErrorIsRuntime --

	Tests if the error code is runtime.

   Arguments:

	error -		The error descriptor.

   Results:

        Returns true if the error is runtime, else returns false.

*/

int
More_ErrorIsRuntime (More_Error error)
{
  return (RUNTIME == error->errorCode);
}

/* More_ErrorSetLogic --

	Set the error code to logic.

   Arguments:

	error -		The error descriptor.

   Results:

        Updates the error code member.

*/

void
More_ErrorSetLogic (More_Error error)
{
  error->errorCode = LOGIC;
}

/* More_ErrorSetRuntime --

	Set the error code to runtime.

   Arguments:

	error -		The error descriptor.

   Results:

        Updates the error code member.

*/

void
More_ErrorSetRuntime (More_Error error)
{
  error->errorCode = RUNTIME;
}

/* More_ErrorNoMemory --

	Wrapper  for  "More_ErrorRuntimeSTR()"  to  return an  error  in
	memory allocation.

   Arguments:

	None.

   Results:

        Returns the new error descriptor.

*/

More_Error
More_ErrorNoMemory (void)
{
  More_Error	error;

  error = More_ErrorNew();
  More_ErrorRuntimeStr(error, "not enough memory");
  return error;
}

/* More_ErrorErrno --

	Builds   an   error   descriptor   with  the   current   "errno"
	informations.

   Arguments:

	None.

   Results:

        Returns the error descriptor.

*/

More_Error
More_ErrorErrno (void)
{
  More_Error	error;

  error = More_ErrorNew();
  More_ErrorRuntimeStr(error, Tcl_ErrnoMsg(Tcl_GetErrno()));
  More_ErrorSetInt(error, Tcl_GetErrno());
  return error;
}

int
More_ErrorCodeAndForget (More_Error error)
{
  int	errorCode;

  errorCode = More_ErrorGetInt(error);
  More_ErrorForget(error);
  return errorCode;
}

/* More_LogicError --

	Signals a logic error and returns an error code.

   Arguments:

	interp -	the interpreter in which report the error

   Results:

        Sets "LOGIC" as error code and returns TCL_ERROR.

*/

int
More_LogicError (Tcl_Interp *interp)
{
  if (interp)
    {
      Tcl_SetErrorCode(interp, LOGIC, NULL);
    }
  return TCL_ERROR;
}

/* More_LogicErrorStr --

	Wrapper for More_LogicError that sets the error information.

   Arguments:

	interp -	Pointer to the interpreter, can be NULL.
	errorInfo -	Pointer to the error message string.

   Results:

        Returns TCL_ERROR.

*/

int
More_LogicErrorStr (Tcl_Interp *interp, CONST char *errorInfo)
{
  if (interp)
    {
      Tcl_SetResult(interp, (char *)errorInfo, TCL_STATIC);
    }
  return More_LogicError(interp);
}

/* More_RuntimeError --

	Signals a runtime error and returns an error code.

   Arguments:

	interp -	The interpreter in which report the
			error, can be NULL.

   Results:

        Sets "RUNTIME" as error code and returns TCL_ERROR.

*/

int
More_RuntimeError (Tcl_Interp *interp)
{
  if (interp)
    {
      Tcl_SetErrorCode(interp, RUNTIME, NULL);
    }
  return TCL_ERROR;
}

/* More_RuntimeErrorStr --

	Wrapper for More_RuntimeError that sets the error information.

   Arguments:

	interp -	Pointer to the interpreter, can be NULL.
	errorInfo -	Pointer to the error message string.

   Results:

        Returns TCL_ERROR.

*/

int
More_RuntimeErrorStr (Tcl_Interp *interp, CONST char *errorInfo)
{
  if (interp)
    {
      Tcl_SetResult(interp, (char *)errorInfo, TCL_STATIC);
    }
  return More_RuntimeError(interp);
}

/* More_WrongNumArgs --

	Sets up the "wrong # args" error message and sets the error code
	to logic.

   Arguments:

	interp -	pointer to the interpreter structure in which
			report the error
	objc -		number of leading arguments from "objv" to
			include in error messages
	objv -		arguments to command that had the wrong number
			of arguments
	message -	additional error informations to print,
			may be NULL

   Results:

        Returns TCL_ERROR and  leaves an error message as  result in the
        interpreter.

   Side effects:

        None.

*/

int
More_WrongNumArgs	(interp, objc, objv, message)
     Tcl_Interp *	interp;
     int		objc;
     Tcl_Obj *CONST	objv[];
     CONST char *	message;
{
  Tcl_WrongNumArgs(interp, objc, objv, message);
  return More_LogicError(interp);
}

/* More_OptionRequiresArg --

	Builds the "option requires argument" error message.

   Arguments:

	interp -	pointer to the interpreter in which report
			the error
	option -	pointer to a string representing the option
			name that was used without argument

   Results:

        Returns TCL_ERROR and  leaves an error message as  result in the
        interpreter.

   Side effects:

        None.

*/

int
More_OptionRequiresArg (Tcl_Interp *interp, CONST char *option)
{
  Tcl_SetObjResult(interp,
		   More_ObjPrintf("option \"%s\" requires argument", option));
  return More_LogicError(interp);
}


/* end of file */
/*
Local Variables:
mode: c
page-delimiter: "^$"
End:
*/
