/* extractor.c --
   
   Part of: tclmore
   Contents: extractor functions used to parse command arguments
   Date: Thu Feb 12, 2004
   
   Abstract
   
   
   
   Copyright (c) 2004 Marco Maggi
   
   The author  hereby grant permission to use,  copy, modify, distribute,
   and  license this  software  and its  documentation  for any  purpose,
   provided that  existing copyright notices  are retained in  all copies
   and that  this notice  is included verbatim  in any  distributions. No
   written agreement, license, or royalty  fee is required for any of the
   authorized uses.  Modifications to this software may be copyrighted by
   their authors and need not  follow the licensing terms described here,
   provided that the new terms are clearly indicated on the first page of
   each file where they apply.
   
   IN NO  EVENT SHALL THE AUTHOR  OR DISTRIBUTORS BE LIABLE  TO ANY PARTY
   FOR  DIRECT, INDIRECT, SPECIAL,  INCIDENTAL, OR  CONSEQUENTIAL DAMAGES
   ARISING OUT  OF THE  USE OF THIS  SOFTWARE, ITS DOCUMENTATION,  OR ANY
   DERIVATIVES  THEREOF, EVEN  IF THE  AUTHOR  HAVE BEEN  ADVISED OF  THE
   POSSIBILITY OF SUCH DAMAGE.
   
   THE  AUTHOR  AND DISTRIBUTORS  SPECIFICALLY  DISCLAIM ANY  WARRANTIES,
   INCLUDING,   BUT   NOT  LIMITED   TO,   THE   IMPLIED  WARRANTIES   OF
   MERCHANTABILITY,    FITNESS   FOR    A    PARTICULAR   PURPOSE,    AND
   NON-INFRINGEMENT.  THIS  SOFTWARE IS PROVIDED  ON AN "AS  IS" BASIS,
   AND  THE  AUTHOR  AND  DISTRIBUTORS  HAVE  NO  OBLIGATION  TO  PROVIDE
   MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
   
*/


/** ------------------------------------------------------------
 ** Header files.
 ** ----------------------------------------------------------*/

#if 0
#  define TCLMORE_DEBUG_MESSAGES
#  define TCLMORE_DEBUGGING	1
#endif

#include <limits.h>
#include <stddef.h>
#include <float.h>
#include "tclmoreInt.h"


/** ------------------------------------------------------------
 ** Preprocessor symbols.
 ** ----------------------------------------------------------*/

#define debug		More_Debug
#define question	More_Question


/* More_GetObjFromArg --

	Extracts  a pointer  to an  object from  a command  argument and
	stores it into the selected struct field.

   Arguments:

	frame -		Pointer to the invocation arguments.

   Results:

        Returns TCL_OK or  TCL_ERROR. In case of error  leaves a message
        in the interpreter and sets the error code to LOGIC.

*/

int
More_GetObjFromArg (More_ExtractorFrame frame)
{
  Tcl_Obj **	objPtr;

  objPtr = (Tcl_Obj **) frame->fieldPtr;
  *objPtr = frame->srcObj;
  return TCL_OK;
}

/* More_GetStringFromArg --

	Extracts a string from an  object command argument and stores it
	into the selected struct field. The string field is allocated as
	a "More_String" instance.

   Arguments:

	frame -		Pointer to the invocation arguments.

   Results:

        Returns TCL_OK or  TCL_ERROR. In case of error  leaves a message
        in the interpreter and sets the error code to LOGIC.

*/

int
More_GetStringFromArg (More_ExtractorFrame frame)
{
  More_String	string;
  More_String *	stringVar;


  string.ptr = Tcl_GetStringFromObj(frame->srcObj, &(string.len));
  stringVar  = (More_String *) frame->fieldPtr;
  *stringVar = string;
  return TCL_OK;
}

/* More_GetAStringFromArg --

	Extracts a non-empty string  from an object command argument and
	stores it  into the selected  struct field. The string  field is
	allocated as a "More_String" instance.

   Arguments:

	frame -		Pointer to the invocation arguments.

   Results:

        Returns TCL_OK or  TCL_ERROR. In case of error  leaves a message
        in the interpreter and sets the error code to LOGIC.

*/

int
More_GetAStringFromArg (More_ExtractorFrame frame)
{
  More_String	string;
  More_String *	stringVar;


  string.ptr = Tcl_GetStringFromObj(frame->srcObj, &(string.len));
  if (string.len == 0)
    {
      Tcl_Interp *	interp = frame->commandFrame->interp;

      Tcl_SetResult(interp, "expected non-empty string", TCL_STATIC);
      return More_LogicError(interp);
    }
  debug("string: %s, pointer: %p", string.ptr, frame->fieldPtr);
  stringVar  = (More_String *) frame->fieldPtr;
  *stringVar = string;
  return TCL_OK;
}

/* More_GetBlockFromArg --

	Extracts a byte array from an object command argument and stores
	it into the  selected struct field. The field  is allocated as a
	"More_Block" instance.

   Arguments:

	frame -		Pointer to the invocation arguments.

   Results:

        Returns TCL_OK or  TCL_ERROR. In case of error  leaves a message
        in the interpreter and sets the error code to LOGIC.

*/

int
More_GetBlockFromArg (More_ExtractorFrame frame)
{
  More_Block	string;
  More_Block *	stringVar;


  string.ptr = Tcl_GetStringFromObj(frame->srcObj, &(string.len));
  stringVar  = (More_Block *) frame->fieldPtr;
  *stringVar = string;
  return TCL_OK;
}

/* More_GetABlockFromArg --

	Extracts a non-empty byte  array from an object command argument
	and  stores it  into the  selected struct  field.  The  field is
	allocated as a "More_Block" instance.

   Arguments:

	frame -		Pointer to the invocation arguments.

   Results:

        Returns TCL_OK or  TCL_ERROR. In case of error  leaves a message
        in the interpreter and sets the error code to LOGIC.

*/

int
More_GetABlockFromArg (More_ExtractorFrame frame)
{
  More_Block	string;
  More_Block *	stringVar;


  string.ptr = Tcl_GetStringFromObj(frame->srcObj, &(string.len));
  if (string.len == 0)
    {
      Tcl_Interp *	interp = frame->commandFrame->interp;

      Tcl_SetResult(interp, "expected non-empty block", TCL_STATIC);
      return More_LogicError(interp);
    }
  stringVar  = (More_Block *) frame->fieldPtr;
  *stringVar = string;
  return TCL_OK;
}

/* More_GetIntFromArg --

	Extracts an integer  from a command argument and  stores it into
	the selected struct field.

   Arguments:

	frame -		Pointer to the invocation arguments.

   Results:

        Returns TCL_OK or  TCL_ERROR. In case of error  leaves a message
        in the interpreter and sets the error code to LOGIC.

*/

int
More_GetIntFromArg (More_ExtractorFrame frame)
{
  int		e;
  Tcl_Interp *	interp = frame->commandFrame->interp;

  e = Tcl_GetIntFromObj(interp, frame->srcObj, (int *) frame->fieldPtr);
  return ((TCL_OK == e)? e : More_LogicError(interp));
}

/* More_GetWideIntFromArg --

	Extracts an integer  from a command argument and  stores it into
	the selected struct field.

   Arguments:

	frame -		Pointer to the invocation arguments.

   Results:

        Returns TCL_OK or  TCL_ERROR. In case of error  leaves a message
        in the interpreter and sets the error code to LOGIC.

*/

int
More_GetWideIntFromArg (More_ExtractorFrame frame)
{
  int		e;
  Tcl_Interp *	interp = frame->commandFrame->interp;

  e = Tcl_GetWideIntFromObj(interp, frame->srcObj,
			    (Tcl_WideInt *) frame->fieldPtr);
  return ((TCL_OK == e)? e : More_LogicError(interp));
}

/* More_GetUnsignedFromArg --

	Extracts an unsigned integer  from a command argument and stores
	it into the selected struct field.

   Arguments:

	frame -		Pointer to the invocation arguments.

   Results:

        Returns TCL_OK or  TCL_ERROR. In case of error  leaves a message
        in the interpreter and sets the error code to LOGIC.

*/

int
More_GetUnsignedFromArg (More_ExtractorFrame frame)
{
  int		e;
  Tcl_Interp *	interp = frame->commandFrame->interp;

  e = More_GetUnsignedFromObj(interp, frame->srcObj,
			      (unsigned *) frame->fieldPtr);
  return ((TCL_OK == e)? e : More_LogicError(interp));
}

/* More_GetFloatFromArg --

	Extracts a  float number from  a command argument and  stores it
	into the selected struct field.

   Arguments:

	frame -		Pointer to the invocation arguments.

   Results:

        Returns TCL_OK or  TCL_ERROR. In case of error  leaves a message
        in the interpreter and sets the error code to LOGIC.

*/

int
More_GetFloatFromArg (More_ExtractorFrame frame)
{
  int		e;
  Tcl_Interp *	interp = frame->commandFrame->interp;

  e = More_GetFloatFromObj(interp, frame->srcObj, (float *) frame->fieldPtr);
  return ((TCL_OK == e)? e : More_LogicError(interp));
}

/* More_GetDoubleFromArg --

	Extracts a double  number from a command argument  and stores it
	into the selected struct field.

   Arguments:

	frame -		Pointer to the invocation arguments.

   Results:

        Returns TCL_OK or  TCL_ERROR. In case of error  leaves a message
        in the interpreter and sets the error code to LOGIC.

*/

int
More_GetDoubleFromArg (More_ExtractorFrame frame)
{
  int		e;
  Tcl_Interp *	interp = frame->commandFrame->interp;

  e = Tcl_GetDoubleFromObj(interp, frame->srcObj, (double *) frame->fieldPtr);
  return ((TCL_OK == e)? e : More_LogicError(interp));
}

/* More_GetSizeTFromArg --

	Extracts a  float number from  a command argument and  stores it
	into the selected struct field.

   Arguments:

	frame -		Pointer to the invocation arguments.

   Results:

        Returns TCL_OK or  TCL_ERROR. In case of error  leaves a message
        in the interpreter and sets the error code to LOGIC.

*/

int
More_GetSizeTFromArg (More_ExtractorFrame frame)
{
  int		e;
  Tcl_Interp *	interp = frame->commandFrame->interp;

  e = More_GetSizeTFromObj(interp, frame->srcObj, (size_t *) frame->fieldPtr);
  return ((TCL_OK == e)? e : More_LogicError(interp));
}




/* end of file */
/*
Local Variables:
mode: c
page-delimiter: "^$"
End:
*/
