/* identifier.c --
   
   Part of: tclmore
   Contents: identifiers table functions
   Date: Mon Aug 25, 2003
   
   Abstract
   
	This  is a  collection of  common functions  to be  used  in TCL
	extension.
   
   Copyright (c) 2002, 2003 Marco Maggi
   
   The author  hereby grant permission to use,  copy, modify, distribute,
   and  license this  software  and its  documentation  for any  purpose,
   provided that  existing copyright notices  are retained in  all copies
   and that  this notice  is included verbatim  in any  distributions. No
   written agreement, license, or royalty  fee is required for any of the
   authorized uses.  Modifications to this software may be copyrighted by
   their authors and need not  follow the licensing terms described here,
   provided that the new terms are clearly indicated on the first page of
   each file where they apply.
   
   IN NO  EVENT SHALL THE AUTHOR  OR DISTRIBUTORS BE LIABLE  TO ANY PARTY
   FOR  DIRECT, INDIRECT, SPECIAL,  INCIDENTAL, OR  CONSEQUENTIAL DAMAGES
   ARISING OUT  OF THE  USE OF THIS  SOFTWARE, ITS DOCUMENTATION,  OR ANY
   DERIVATIVES  THEREOF, EVEN  IF THE  AUTHOR  HAVE BEEN  ADVISED OF  THE
   POSSIBILITY OF SUCH DAMAGE.
   
   THE  AUTHOR  AND DISTRIBUTORS  SPECIFICALLY  DISCLAIM ANY  WARRANTIES,
   INCLUDING,   BUT   NOT  LIMITED   TO,   THE   IMPLIED  WARRANTIES   OF
   MERCHANTABILITY,    FITNESS   FOR    A    PARTICULAR   PURPOSE,    AND
   NON-INFRINGEMENT.  THIS  SOFTWARE IS PROVIDED  ON AN "AS  IS" BASIS,
   AND  THE  AUTHOR  AND  DISTRIBUTORS  HAVE  NO  OBLIGATION  TO  PROVIDE
   MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
   
   $Id: identifier.c,v 1.1.1.19 2003/12/23 08:42:42 marco Exp $
*/



/** ------------------------------------------------------------
 ** Header files.
 ** ----------------------------------------------------------*/

#include "tclmoreInt.h"
#include <stddef.h>
#include <string.h>


/* More_InitIdTable --

	Initialises the identifiers table.

   Arguments:

	data -		pointer to an already allocated table struct
	template -	pointer to a string holding the identifier
			template, something like "file%d"
	destructor -	pointer to a function that could be used
			to destroy the data associated to an
			identifier

   Results:

        The   data  structure   is  initialised.   The  hash   table  is
        initialised.

   Side effects:

        None.

*/

void
More_InitIdTable (data, template, destructor)
     More_IdTable *		data;
     CONST char *		template;
     More_IdDestructor		destructor;
{
  Tcl_InitHashTable(&(data->table), TCL_STRING_KEYS);
  data->counter		= 0;
  data->tmplPtr		= template;
  data->tmplLen		= sizeof(char)*(strlen(template)+32+1);
  data->destructor	= destructor;
}

/* More_DeleteIdTable --

	Deletes all the objects in the table.

   Arguments:

	data -		pointer to the table data structure

   Results:

        Extracts all  the data from the  table and destroys  it with the
        registered  destructor (if not  NULL).

   Side effects:

        None.

*/

void
More_DeleteIdTable (More_IdTable *data)
{
  Tcl_HashEntry *	entry;
  Tcl_HashSearch	search;
  Tcl_HashTable *	table = &(data->table);
  More_IdDestructor	destructor = data->destructor;


  for (entry = Tcl_FirstHashEntry(table, &search);
       entry != NULL;
       entry = Tcl_NextHashEntry(&search))
    {
      if (destructor != NULL)
	{
	  destructor(Tcl_GetHashValue(entry));
	}
      Tcl_DeleteHashEntry(entry);
    }
  
  Tcl_DeleteHashTable(table);
}

/* More_GetDataFromId --

	Extracts the data associated to a key in the identifiers table.

   Arguments:

	data -		pointer to a table structure
	idObj -		pointer to the object holding the
			identifier's string

   Results:

        Returns the data or NULL if the identifier is not present in the
        table.

   Side effects:

        None.

*/

ClientData
More_GetDataFromId (More_IdTable *data, Tcl_Obj *idObj)
{
  Tcl_HashEntry *	entry;


  entry = Tcl_FindHashEntry(&(data->table), Tcl_GetString(idObj));
  return ((entry != NULL)? Tcl_GetHashValue(entry) : NULL);
}

/* More_AttachId --

	Inserts a new value in the table.

   Arguments:

	data -		pointer to the table data structure
	D -		the data

   Results:

        Builds  and  returns a  new  identifier's  string,  the data  is
        associated to it in the hash table.

   Side effects:

        None.

*/

Tcl_Obj *
More_AttachId (More_IdTable *data, ClientData D)
{
  int			new;
  Tcl_HashEntry *	entry;
  char			threadId[data->tmplLen];
  

  do
    {
      memset(threadId, '\0', data->tmplLen);
      sprintf(threadId, data->tmplPtr, ++(data->counter));
      entry = Tcl_CreateHashEntry(&(data->table), threadId, &new);
    }
  while (!new);

  Tcl_SetHashValue(entry, D);
  return Tcl_NewStringObj(threadId, -1);
}

/* More_DetachId --

	Extracts an identifier from the table.

   Arguments:

	data -		pointer to the table data structure
	idObj -		pointer to the object holding the
			identifier's string

   Results:

        Extracts the  data from the  table; the data is  just extracted,
        not destroyed.  If the identifier  is not in the  table, nothing
        happens.

   Side effects:

        None.

*/

void
More_DetachId (More_IdTable * data, Tcl_Obj *idObj)
{
  Tcl_HashEntry *	entry;


  entry = Tcl_FindHashEntry(&(data->table), Tcl_GetString(idObj));
  if (entry != NULL)
    {
      Tcl_DeleteHashEntry(entry);
    }
}

/* end of file */
/*
  Local Variables:
  page-delimiter: "^"
  End:
*/
