/* miscellaneous.c --
   
   Part of: tclmore
   Contents: miscellaneous functions
   Date: Mon Feb 23, 2004
   
   Abstract
   
   
   
   Copyright (c) 2004 Marco Maggi
   
   The author  hereby grant permission to use,  copy, modify, distribute,
   and  license this  software  and its  documentation  for any  purpose,
   provided that  existing copyright notices  are retained in  all copies
   and that  this notice  is included verbatim  in any  distributions. No
   written agreement, license, or royalty  fee is required for any of the
   authorized uses.  Modifications to this software may be copyrighted by
   their authors and need not  follow the licensing terms described here,
   provided that the new terms are clearly indicated on the first page of
   each file where they apply.
   
   IN NO  EVENT SHALL THE AUTHOR  OR DISTRIBUTORS BE LIABLE  TO ANY PARTY
   FOR  DIRECT, INDIRECT, SPECIAL,  INCIDENTAL, OR  CONSEQUENTIAL DAMAGES
   ARISING OUT  OF THE  USE OF THIS  SOFTWARE, ITS DOCUMENTATION,  OR ANY
   DERIVATIVES  THEREOF, EVEN  IF THE  AUTHOR  HAVE BEEN  ADVISED OF  THE
   POSSIBILITY OF SUCH DAMAGE.
   
   THE  AUTHOR  AND DISTRIBUTORS  SPECIFICALLY  DISCLAIM ANY  WARRANTIES,
   INCLUDING,   BUT   NOT  LIMITED   TO,   THE   IMPLIED  WARRANTIES   OF
   MERCHANTABILITY,    FITNESS   FOR    A    PARTICULAR   PURPOSE,    AND
   NON-INFRINGEMENT.  THIS  SOFTWARE IS PROVIDED  ON AN "AS  IS" BASIS,
   AND  THE  AUTHOR  AND  DISTRIBUTORS  HAVE  NO  OBLIGATION  TO  PROVIDE
   MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
   
*/


/** ------------------------------------------------------------
 ** Header files.
 ** ----------------------------------------------------------*/

/*
#define TCLMORE_DEBUG_MESSAGES
#define TCLMORE_DEBUGGING	1
*/

#include "tclmoreInt.h"
#include <stddef.h>
#include <assert.h>

#define debug			More_Debug


/* More_EqualVarnames --

	Compares  a  variable  name  in  a  string  with  a  pair  array
	name/array key.

   Arguments:

	name -		Pointer to the variable name.
	var -		Pointer to the array name.
	key -		Pointer to the key name.

   Results:

        Returns  true  if  the  two  variable  names  are  equal,  false
        otherwise.

*/

int
More_EqualVarnames (CONST char *name, CONST char *var, CONST char *key)
{
  unsigned	len, len1, len2;


  if (key == NULL)
    {
      return (strcmp(name, var) == 0)? 1 : 0;
    }

  len1 = strlen(var);
  len2 = strlen(key);
  len = len1 + len2 + 2;
  if (strlen(name) != len)
    {
      return 0;
    }
  if ((name[len1] != '(') || (name[len1+len2+1] != '('))
    {
      return 0;
    }
  if (strncmp(name, var, len1) != 0)
    {
      return 0;
    }
  if (strncmp(name+((ptrdiff_t)(len1+1)), var, len2) != 0)
    {
      return 0;
    }
  return 1;
}

/* More_MakeName --

	Builds a new unique name for the interpreter.

   Arguments:

	interp -	The interpreter to which the channel
			counter is associated.
	template -	Pointer to the template string.

   Results:

        Returns a pointer to  a dynamically allocated string. The buffer
        is allocated with a call to ckalloc().

*/

char *
More_MakeName (Tcl_Interp *interp, CONST char *template)
{
  More_AssocData *	data;
  Tcl_HashTable *	table;
  Tcl_HashEntry *	entry;
  char *		buffer = NULL;
  size_t		buffer_size = 24, char_count;
  int			new;
  unsigned		counter;


  data  = More_GetAssocData(interp);
  table = &(data->channelCounters);
  if (! data->channelCountersInitialised)
    {
      debug("initialising hash table");
      Tcl_InitHashTable(table, TCL_STRING_KEYS);
      data->channelCountersInitialised = 1;
    }

  entry = Tcl_CreateHashEntry(table, template, &new);
  if (new)
    {
      debug("creating new element for template: %s", template);
      counter = 1;
      Tcl_SetHashValue(entry, (ClientData) counter);
    }
  else
    {
      counter = (int) Tcl_GetHashValue(entry);
      ++counter;
      Tcl_SetHashValue(entry, (ClientData) counter);      
      debug("incremented counter for template: %s, %u", template, counter);
    }

  do
    {
      buffer_size <<= 1;
      buffer = ckrealloc(buffer, buffer_size);
      char_count = snprintf(buffer, buffer_size, template, counter);
    }
  while (char_count == (buffer_size-1));

  debug("new channel name %s", buffer);
  return buffer;
}

/* More_DupAllocString --

	Duplicates a string into dynamically allocated buffer.

   Arguments:

	string -	Pointer to the source string.

   Results:

        Returns the pointer to the new buffer, allocated with ckalloc().

*/

char *
More_DupAllocString (CONST char *string)
{
  char *	new;

  new = ckalloc(strlen(string) + sizeof(char));
  return strcpy(new, string);
}

/* More_Asprintf --

	An  "sprintf()" version  that  allocates enough  memory for  the
	string.

   Arguments:

	format -	Pointer to the format string.
	... -		Format arguments.

   Results:

        Returns  a   pointer  to  the  output   string,  allocated  with
        "ckalloc()".

*/

char *
More_Asprintf (CONST char *format, ...)
{
  char *	buffer;
  va_list	ap;


  if (NULL == format)
    {
      return NULL;
    }

  va_start(ap, format);
  buffer = More_Bsprintf(format, ap);
  va_end(ap);
  return buffer;
}

/* More_ObjPrintf --

	An  "asprintf()" version  that  outputs the  string  into a  new
	string object.

   Arguments:

	format -	Pointer to the format string.
	... -		Format arguments.

   Results:

        Returns  a pointer  to  the new  string  object, with  reference
        counter set to zero. If  "format" is NULL the returned object is
        empty.

*/

Tcl_Obj *
More_ObjPrintf (CONST char *format, ...)
{
  char *	buffer;
  Tcl_Obj *	objPtr;
  va_list	ap;


  if (! format)
    {
      return Tcl_NewStringObj(NULL, 0);
    }

  va_start(ap, format);
  buffer = More_Bsprintf(format, ap);
  va_end(ap);

  objPtr = Tcl_NewStringObj(buffer, -1);
  ckfree(buffer);
  return objPtr;
}

/* More_Bsprintf --

	Utility function for "More_Asprintf()".

   Arguments:

	format -	Pointer to the format string.
	ap -		Format arguments.

   Results:

        Returns  a   pointer  to  the  output   string,  allocated  with
        "ckalloc()".

*/

char *
More_Bsprintf (CONST char *CONST format, va_list ap)
{
  char *	buffer = NULL;
  size_t	bufferSize = 128, charCount;


  buffer = ckalloc(bufferSize);
  charCount = vsnprintf(buffer, bufferSize, format, ap);
  if (charCount >= bufferSize)
    {
      bufferSize = charCount + 1;
      buffer = ckrealloc(buffer, bufferSize);
      vsnprintf(buffer, bufferSize, format, ap);
    }
  else if (charCount < bufferSize);
    {
      buffer = ckrealloc(buffer, charCount+1);
    }
  return buffer;
}


/* end of file */
/*
Local Variables:
mode: c
page-delimiter: "^$"
End:
*/
