/* object.c --
   
   Part of: tclmore
   Contents: object value extractors
   Date: Sat Jan 17, 2004
   
   Abstract
   
	This  module  collects functions  used  to  extract values  from
	Tcl_Obj instances.
   
   Copyright (c) 2004 Marco Maggi
   
   The author  hereby grant permission to use,  copy, modify, distribute,
   and  license this  software  and its  documentation  for any  purpose,
   provided that  existing copyright notices  are retained in  all copies
   and that  this notice  is included verbatim  in any  distributions. No
   written agreement, license, or royalty  fee is required for any of the
   authorized uses.  Modifications to this software may be copyrighted by
   their authors and need not  follow the licensing terms described here,
   provided that the new terms are clearly indicated on the first page of
   each file where they apply.
   
   IN NO  EVENT SHALL THE AUTHOR  OR DISTRIBUTORS BE LIABLE  TO ANY PARTY
   FOR  DIRECT, INDIRECT, SPECIAL,  INCIDENTAL, OR  CONSEQUENTIAL DAMAGES
   ARISING OUT  OF THE  USE OF THIS  SOFTWARE, ITS DOCUMENTATION,  OR ANY
   DERIVATIVES  THEREOF, EVEN  IF THE  AUTHOR  HAVE BEEN  ADVISED OF  THE
   POSSIBILITY OF SUCH DAMAGE.
   
   THE  AUTHOR  AND DISTRIBUTORS  SPECIFICALLY  DISCLAIM ANY  WARRANTIES,
   INCLUDING,   BUT   NOT  LIMITED   TO,   THE   IMPLIED  WARRANTIES   OF
   MERCHANTABILITY,    FITNESS   FOR    A    PARTICULAR   PURPOSE,    AND
   NON-INFRINGEMENT.  THIS  SOFTWARE IS PROVIDED  ON AN "AS  IS" BASIS,
   AND  THE  AUTHOR  AND  DISTRIBUTORS  HAVE  NO  OBLIGATION  TO  PROVIDE
   MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
   
   $Id: object.c,v 1.1.1.8 2004/02/02 19:12:23 marco Exp marco $
*/


/** ------------------------------------------------------------
 ** Header files.
 ** ----------------------------------------------------------*/

/*
#define TCLMORE_DEBUG_MESSAGES
#define TCLMORE_DEBUGGING	1
*/

#include <limits.h>
#include <stddef.h>
#include <float.h>
#include "tclmoreInt.h"


/** ------------------------------------------------------------
 ** Preprocessor symbols.
 ** ----------------------------------------------------------*/

#define TCLMORE_FLOAT_MAX	FLT_MAX
#define TCLMORE_FLOAT_MIN	FLT_MIN

#define debug			More_Debug

/* More_NewUnsignedObj --

	Builds a new object holding an "unsigned int" value.

   Arguments:

	value -		The value.

   Results:

        Returns a pointer  to the new object with  reference counter set
        to zero.

*/

Tcl_Obj *
More_NewUnsignedObj(unsigned value)
{
  char *	string;
  Tcl_Obj *	objPtr;


  string = More_Asprintf("%u", value);
  objPtr = Tcl_NewStringObj(string, -1);
  ckfree(string);
  return objPtr;
}

/* More_GetUnsignedFromObj --

	Extracts an unsigned integer from an object.

   Arguments:

	interp -	If not NULL: the interpreter used
			to report errors.
	objPtr -	Pointer to the source object.
	valVar -	Pointer to the target variable.

   Results:

        Returns TCL_OK or TCL_ERROR.

*/

int
More_GetUnsignedFromObj (interp, objPtr, valVar)
     Tcl_Interp *	interp;
     Tcl_Obj *		objPtr;
     unsigned *		valVar;
{
  Tcl_WideInt	value;
  int		e;
  char *	errorString;


  e = Tcl_GetWideIntFromObj(interp, objPtr, &value);
  if ((TCL_OK != e) || (value < 0) || (value > UINT_MAX))
    {
      goto Error;
    }

  *valVar = (unsigned) value;
  return TCL_OK;

 Error:
  if (interp)
    {
      errorString = More_Asprintf \
	("expected unsigned but got \"%s\"", Tcl_GetString(objPtr));
      Tcl_SetResult(interp, (char *) errorString, TCL_DYNAMIC);
      Tcl_SetErrorCode(interp, "LOGIC", NULL);
    }
  return TCL_ERROR;
}

/* More_GetUnsignedInRangeFromObj --

	Extracts an unsigned integer inside a range from an object.

   Arguments:

	interp -	If not NULL: the interpreter used
			to report errors.
	objPtr -	Pointer to the source object.
	min -		The minimum value.
	max -		The maximum value.
	valVar -	Pointer to the target variable.

   Results:

        Returns TCL_OK or TCL_ERROR.

*/

int
More_GetUnsignedInRangeFromObj (interp, objPtr, min, max, valVar)
     Tcl_Interp *	interp;
     Tcl_Obj *		objPtr;
     unsigned		min;
     unsigned		max;
     unsigned *		valVar;
{
  unsigned	value;
  int		e;
  CONST char *	errorString;


  e = More_GetUnsignedFromObj(interp, objPtr, &value);
  if ((TCL_OK != e) || (value < min) || (max < value))
    {
      goto Error;
    }

  *valVar = value;
  return TCL_OK;

 Error:
  if (interp)
    {
      errorString = More_Asprintf \
	("expected unsigned integer between %d and %d but got \"%s\"",
	 min, max, Tcl_GetString(objPtr));
      Tcl_SetResult(interp, (char *) errorString, TCL_DYNAMIC);
      Tcl_SetErrorCode(interp, "LOGIC", NULL);
    }
  return TCL_ERROR;
}

/* More_GetIntInRangeFromObj --

	Extracts an integer inside a range from an object.

   Arguments:

	interp -	If not NULL: the interpreter used
			to report errors.
	objPtr -	Pointer to the source object.
	min -		The minimum value.
	max -		The maximum value.
	valVar -	Pointer to the target variable.

   Results:

        Returns TCL_OK or TCL_ERROR.

*/

int
More_GetIntInRangeFromObj (interp, objPtr, min, max, valVar)
     Tcl_Interp *	interp;
     Tcl_Obj *		objPtr;
     int		min;
     int		max;
     int *		valVar;
{
  int		value;
  int		e;
  CONST char *	errorString;


  e = Tcl_GetIntFromObj(interp, objPtr, &value);
  if ((TCL_OK != e) || (value < min) || (max < value))
    {
      goto Error;
    }

  *valVar = value;
  return TCL_OK;

 Error:
  if (interp)
    {
      errorString = More_Asprintf \
	("expected integer between %d and %d but got \"%s\"",
	 min, max, Tcl_GetString(objPtr));
      Tcl_SetResult(interp, (char *) errorString, TCL_DYNAMIC);
      Tcl_SetErrorCode(interp, "LOGIC", NULL);
    }
  return TCL_ERROR;
}

/* More_GetWideIntInRangeFromObj --

	Extracts a wide integer inside a range from an object.

   Arguments:

	interp -	If not NULL: the interpreter used
			to report errors.
	objPtr -	Pointer to the source object.
	min -		The minimum value.
	max -		The maximum value.
	valVar -	Pointer to the target variable.

   Results:

        Returns TCL_OK or TCL_ERROR.

*/

int
More_GetWideIntInRangeFromObj (interp, objPtr, min, max, valVar)
     Tcl_Interp *	interp;
     Tcl_Obj *		objPtr;
     Tcl_WideInt	min;
     Tcl_WideInt	max;
     Tcl_WideInt *	valVar;
{
  Tcl_WideInt	value;
  int		e;
  CONST char *	errorString;


  e = Tcl_GetWideIntFromObj(interp, objPtr, &value);
  if ((TCL_OK != e) || (value < min) || (max < value))
    {
      goto Error;
    }

  *valVar = value;
  return TCL_OK;

 Error:
  if (interp)
    {
      Tcl_Obj *	minObj;
      Tcl_Obj * maxObj;

      minObj = Tcl_NewWideIntObj(min);
      maxObj = Tcl_NewWideIntObj(max);
      Tcl_IncrRefCount(minObj);
      Tcl_IncrRefCount(maxObj);

      errorString = More_Asprintf \
	("expected a wide integer between %s and %s but got \"%s\"",
	 Tcl_GetString(minObj), Tcl_GetString(maxObj), Tcl_GetString(objPtr));

      Tcl_DecrRefCount(minObj);
      Tcl_DecrRefCount(maxObj);

      Tcl_SetResult(interp, (char *) errorString, TCL_DYNAMIC);
      Tcl_SetErrorCode(interp, "LOGIC", NULL);
    }
  return TCL_ERROR;
}

/* More_NewSizeTObj --

	Builds a new object holding an "size_t" value.

   Arguments:

	value -		The value.

   Results:

        Returns a pointer  to the new object with  reference counter set
        to zero.

*/

Tcl_Obj *
More_NewSizeTObj(size_t value)
{
  char *	string;
  Tcl_Obj *	objPtr;


  string = More_Asprintf(TCLMORE_SIZE_T_PRINTF, value);
  objPtr = Tcl_NewStringObj(string, -1);
  ckfree(string);
  return objPtr;
}

/* More_GetSizeTFromObj --

	Extracts a size_t number from an object.

   Arguments:

	interp -	If not NULL: the interpreter used
			to report errors.
	objPtr -	Pointer to the source object.
	min -		The minimum value.
	max -		The maximum value.
	valVar -	Pointer to the target variable.

   Results:

        Returns TCL_OK or TCL_ERROR.

*/

int
More_GetSizeTFromObj (interp, objPtr, valVar)
     Tcl_Interp *	interp;
     Tcl_Obj *		objPtr;
     size_t *		valVar;
{
  Tcl_WideInt	value;
  int		e;
  CONST char *	errorString;


  e = Tcl_GetWideIntFromObj(interp, objPtr, &value);
  if ((TCL_OK != e) || (value < 0))
    {
      goto Error;
    }

#ifdef TCLMORE_SIZE_T_IS_USINT
  if (value > USINT_MAX)
    {
      goto Error;
    }
#elif (defined TCLMORE_SIZE_T_IS_UINT)
  if (value > UINT_MAX)
    {
      goto Error;
    }
#elif (defined TCLMORE_SIZE_T_IS_ULONG)
  if (value > ULONG_MAX)
    {
      goto Error;
    }
#elif (defined TCLMORE_SIZE_T_IS_ULONGLONG)
  if (value > ULONG_LONG_MAX)
    {
      goto Error;
    }
#else
  if (interp)
    {
      errorString = More_Asprintf \
	("cannot extract size integer from \"%s\"", Tcl_GetString(objPtr));
      Tcl_SetResult(interp, (char *) errorString, TCL_DYNAMIC);
      Tcl_SetErrorCode(interp, "LOGIC", NULL);
    }
  return TCL_ERROR;
#endif
  
  /*
    If we are here and none  of the ifs triggered a "size_t" is probably
    an "unsigned long long int" so it's ok to return it. (Isn't it?).
  */

  *valVar = (size_t) value;
  return TCL_OK;

 Error:
  if (interp)
    {
      errorString = More_Asprintf \
	("expected size integer but got \"%s\"", Tcl_GetString(objPtr));
      Tcl_SetResult(interp, (char *) errorString, TCL_DYNAMIC);
      Tcl_SetErrorCode(interp, "LOGIC", NULL);
    }
  return TCL_ERROR;
}

/* More_GetSizeTInRangeFromObj --

	Extracts a "size_t" in a range from an object.

   Arguments:

	interp -	If not NULL: the interpreter used
			to report errors.
	objPtr -	Pointer to the source object.
	min -		The minimum value, inclusive.
	max -		The maximum value, inclusive.
	valVar -	Pointer to the target variable.

   Results:

        Returns TCL_OK or TCL_ERROR.

*/

int
More_GetSizeTInRangeFromObj (interp, objPtr, min, max, valVar)
     Tcl_Interp *	interp;
     Tcl_Obj *		objPtr;
     size_t		min;
     size_t		max;
     size_t *		valVar;
{
  size_t	value;
  int		e;


  e = More_GetSizeTFromObj(interp, objPtr, &value);
  if ((TCL_OK != e) || (value < min) || (max < value))
    {
      goto Error;
    }

  *valVar = value;
  return TCL_OK;

 Error:
  if (interp)
    {
      Tcl_SetResult(interp, "size integer out of range", TCL_STATIC);
      Tcl_SetErrorCode(interp, "LOGIC", NULL);
    }
  return TCL_ERROR;
}

/* More_NewFloatObj --

	Builds a new object holding an "float" value.

   Arguments:

	value -		The value.

   Results:

        Returns a pointer  to the new object with  reference counter set
        to zero.

*/

Tcl_Obj *
More_NewFloatObj(float value)
{
  char *	string;
  Tcl_Obj *	objPtr;


  string = More_Asprintf("%f", value);
  objPtr = Tcl_NewStringObj(string, -1);
  ckfree(string);
  return objPtr;
}

/* More_GetFloatFromObj --

	Extracts a float from an object.

   Arguments:

	interp -	If not NULL: the interpreter used
			to report errors.
	objPtr -	Pointer to the source object.
	valVar -	Pointer to the target variable.

   Results:

        Returns TCL_OK or TCL_ERROR.

*/

int
More_GetFloatFromObj (interp, objPtr, valVar)
     Tcl_Interp *	interp;
     Tcl_Obj *		objPtr;
     float *		valVar;
{
  double	value;
  int		e;
  CONST char *	errorString;


  e = Tcl_GetDoubleFromObj(interp, objPtr, &value);
  if ((TCL_OK != e) ||
      (value < TCLMORE_FLOAT_MIN) || (TCLMORE_FLOAT_MAX < value))
    {
      goto Error;
    }

  *valVar = (float) value;
  return TCL_OK;

 Error:
  if (interp)
    {
      errorString = More_Asprintf \
	("expected float number but got \"%s\"", Tcl_GetString(objPtr));
      Tcl_SetResult(interp, (char *) errorString, TCL_DYNAMIC);
      Tcl_SetErrorCode(interp, "LOGIC", NULL);
    }
  return TCL_ERROR;
}

/* More_GetFloatInRangeFromObj --

	Extracts a float in a range from an object.

   Arguments:

	interp -	If not NULL: the interpreter used
			to report errors.
	objPtr -	Pointer to the source object.
	min -		The minimum value, inclusive.
	max -		The maximum value, inclusive.
	valVar -	Pointer to the target variable.

   Results:

        Returns TCL_OK or TCL_ERROR.

*/

int
More_GetFloatInRangeFromObj (interp, objPtr, min, max, valVar)
     Tcl_Interp *	interp;
     Tcl_Obj *		objPtr;
     float		min;
     float		max;
     float *		valVar;
{
  double	value;
  int		e;
  CONST char *	errorString;


  e = Tcl_GetDoubleFromObj(interp, objPtr, &value);
  if ((TCL_OK != e) || (value < min) || (max < value))
    {
      goto Error;
    }

  *valVar = (float) value;
  return TCL_OK;

 Error:
  if (interp)
    {
      errorString = More_Asprintf \
	("expected float number between %f and %f but got \"%s\"",
	 min, max, Tcl_GetString(objPtr));
      Tcl_SetResult(interp, (char *) errorString, TCL_DYNAMIC);
      Tcl_SetErrorCode(interp, "LOGIC", NULL);
    }
  return TCL_ERROR;
}

/* More_GetOpenModeFromObj --

	Extracts the open mode from an object.

   Arguments:

	interp -	The interpreter in which report errors.
	objPtr -	Pointer to the source object.
	flags -		Zero or TCLMORE_READ_OR_WRITE.
	modeVar -	Pointer to the target variable.

   Results:

        Returns TCL_OK or TCL_ERROR. If an error occurs an error message
        is left in  the interpreter result and the error  code is set to
        "LOGIC".

*/

int
More_GetOpenModeFromObj (interp, objPtr, flags, modeVar)
     Tcl_Interp *	interp;
     Tcl_Obj *		objPtr;
     int		flags;
     int *		modeVar;
{
  static CONST char * modes[] = { "RDWR", "RDONLY", "WRONLY", NULL };
  CONST char **	modesPtr = modes;
  int		e, index;


  if (flags & TCLMORE_READ_OR_WRITE)
    {
      ++modesPtr;
    }
  e = Tcl_GetIndexFromObj(interp, objPtr, modesPtr, "mode", TCL_EXACT, &index);
  if (TCL_OK != e)
    {
      return More_LogicError(interp);
    }
  if (flags & TCLMORE_READ_OR_WRITE)
    {
      *modeVar = (index)? TCL_WRITABLE : TCL_READABLE;
    }
  else
    {
      switch (index)
	{
	case 0: /* RDWR */
	  *modeVar = TCL_READABLE|TCL_WRITABLE;
	  break;

	case 1: /* RDONLY */
	  *modeVar = TCL_READABLE;
	  break;

	case 2: /* WRONLY */
	  *modeVar = TCL_WRITABLE;
	  break;
	}
    }
  return TCL_OK;
}

/* More_GetTransformOpenModeFromObj --

	Extracts a transformation open mode from an object. The mode has
	to be compatible with the mode of the underlying channel.

   Arguments:

	interp -	The interpreter in which report errors.
	objPtr -	Pointer to the source object.
	subModeMask -	The underlying channel mode.
	flags -		Zero or TCLMORE_READ_OR_WRITE.
	modeMaskVar -	Pointer to the target variable.

   Results:

        Returns TCL_OK or TCL_ERROR. If an error occurs an error message
        is left in  the interpreter result and the error  code is set to
        "LOGIC".

*/

int
More_GetTransformOpenModeFromObj (interp, objPtr, subModeMask, flags, modeMaskVar)
     Tcl_Interp *	interp;
     Tcl_Obj *		objPtr;
     int		subModeMask;
     int		flags;
     int *		modeMaskVar;
{
  int	e, modeMask;


  e = More_GetOpenModeFromObj(interp, objPtr, flags, &modeMask);
  if (TCL_OK != e)
    {
      return e;
    }

  if (! (subModeMask & TCL_READABLE))
    {
      modeMask &= (~ TCL_READABLE);
    }
  if (! (subModeMask & TCL_WRITABLE))
    {
      modeMask &= (~ TCL_WRITABLE);
    }
  if (! modeMask)
    {
      Tcl_SetResult(interp, "null mode for transformation", TCL_STATIC);
      return More_LogicError(interp);
    }

  *modeMaskVar = modeMask;
  return TCL_OK;
}

/* More_GetChannelFromObj --

	Extracts  a channel  token from  an object  holding  the channel
	name.

   Arguments:

	interp -	Pointer to the interpreter to which the
			channel belongs. Cannot be NULL.
	objPtr -	Pointer to the source object.
	channelVar -	If not NULL: pointer to the variable that
			will hold the channel token.
	modeMaskVar -	If not NULL: pointer to the variable that
			will hold the mode mask.

   Results:

        Returns TCL_OK or TCL_ERROR. If an error occurs an error message
        is left in  the interpreter result and the error  code is set to
        "LOGIC".

*/

int
More_GetChannelFromObj (interp, objPtr, channelVar, modeMaskVar)
     Tcl_Interp *	interp;
     Tcl_Obj *		objPtr;
     Tcl_Channel *	channelVar;
     int *		modeMaskVar;
{
  Tcl_Channel	channel;
  int		modeMask;


  if (! Tcl_IsChannelExisting(Tcl_GetString(objPtr)))
    {
      Tcl_SetResult(interp, "unexistent channel", TCL_STATIC);
      return More_LogicError(interp);
    }

  channel = Tcl_GetChannel(interp, Tcl_GetString(objPtr), &modeMask);
  if (! channel)
    {
      Tcl_SetResult(interp, "channel not associated to this interpreter",
		    TCL_STATIC);
      return More_LogicError(interp);
    }

  if (channelVar)
    {
      *channelVar = channel;
    }
  if (modeMaskVar)
    {
      *modeMaskVar = modeMask;
    }
  return TCL_OK;
}


/* end of file */
/*
Local Variables:
mode: c
page-delimiter: "^$"
End:
*/
