/* sharedbuffer.c --
   
   Part of: tclmore
   Contents: shared buffer functions
   Date: Sun Feb  1, 2004
   
   Abstract
   
   
   
   Copyright (c) 2004 Marco Maggi
   
   The author  hereby grant permission to use,  copy, modify, distribute,
   and  license this  software  and its  documentation  for any  purpose,
   provided that  existing copyright notices  are retained in  all copies
   and that  this notice  is included verbatim  in any  distributions. No
   written agreement, license, or royalty  fee is required for any of the
   authorized uses.  Modifications to this software may be copyrighted by
   their authors and need not  follow the licensing terms described here,
   provided that the new terms are clearly indicated on the first page of
   each file where they apply.
   
   IN NO  EVENT SHALL THE AUTHOR  OR DISTRIBUTORS BE LIABLE  TO ANY PARTY
   FOR  DIRECT, INDIRECT, SPECIAL,  INCIDENTAL, OR  CONSEQUENTIAL DAMAGES
   ARISING OUT  OF THE  USE OF THIS  SOFTWARE, ITS DOCUMENTATION,  OR ANY
   DERIVATIVES  THEREOF, EVEN  IF THE  AUTHOR  HAVE BEEN  ADVISED OF  THE
   POSSIBILITY OF SUCH DAMAGE.
   
   THE  AUTHOR  AND DISTRIBUTORS  SPECIFICALLY  DISCLAIM ANY  WARRANTIES,
   INCLUDING,   BUT   NOT  LIMITED   TO,   THE   IMPLIED  WARRANTIES   OF
   MERCHANTABILITY,    FITNESS   FOR    A    PARTICULAR   PURPOSE,    AND
   NON-INFRINGEMENT.  THIS  SOFTWARE IS PROVIDED  ON AN "AS  IS" BASIS,
   AND  THE  AUTHOR  AND  DISTRIBUTORS  HAVE  NO  OBLIGATION  TO  PROVIDE
   MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
   
   $Id: sharedbuffer.c,v 1.1.1.9 2004/02/03 09:44:42 marco Exp $
*/


/** ------------------------------------------------------------
 ** Header files.
 ** ----------------------------------------------------------*/


#define TCLMORE_DEBUG_MESSAGES
#define TCLMORE_DEBUGGING	1


#include "tclmoreInt.h"
#include <assert.h>
#include <stddef.h>
#include <string.h>


/** ------------------------------------------------------------
 ** Module's type declarations.
 ** ----------------------------------------------------------*/

#define Byte		More_Byte
#define BytePtr		More_BytePtr

/* SharedBuffer --

  Type used as base for shared  buffers. When a buffer is shared between
  two instances of some type,  possibly owned by different threads, this
  type allows exclusive access and reference counting.


  mutex -		Mutex  used  for  exclusive  acces. It  is  used
			automatically the shared buffer functions.

  readerAttached -	Boolean: true if an entity has registered itself
			to use the buffer as its input.

  writerAttached -	Boolean: true if an entity has registered itself
			to use the buffer as its output.

  readableEventQueued - Boolean: set to false after a read operation has
			been  executed; set  to true   after  the reader
			callback has been invoked.

  writableEventQueued -	Boolean:  set to false  after a  write operation
			has been executed; set  to true after the writer
			callback has been invoked.

  buffer -		Pointer to the first buffer in a chain.

  bufferSize -		Dimension of  the data area for  the next buffer
			allocated.

  readerCallback -	Callback  function  used  to notify  the  entity
			attached to  the output side that the  buffer is
			readable.

  writerCallback -	Callback  function  used  to notify  the  entity
			attached to  the input  side that the  buffer is
			writable.


    A controller is  responsible to create an instance  of this type and
  send references to it to a couple of entities: one using the buffer as
  input and one using the buffer  as output. When the source or the sink
  both detach themself from the buffer, the buffer is freed.

*/

typedef struct SharedBuffer {
  Tcl_Mutex		mutex;
  More_Buffer		buffer;
  size_t		bufferSize;
  int			readerAttached;
  int			writerAttached;
  int			readableEventQueued;
  int			writableEventQueued;
  More_Callback		readerCallback;
  More_Callback		writerCallback;
} SharedBuffer;


/** ------------------------------------------------------------
 ** Preprocessor symbols.
 ** ----------------------------------------------------------*/

#define debug			More_Debug
#define DEFAULT_BUFFER_SIZE	4096


/** ------------------------------------------------------------
 ** Module's function prototypes.
 ** ----------------------------------------------------------*/

static void	SharedBufferFinal _ANSI_ARGS_((SharedBuffer *base));

static void	SharedBufferNotify _ANSI_ARGS_((SharedBuffer *base));


/* More_SharedBufferAlloc --

	Allocates a new shared buffer.

   Arguments:

	bufferPtr -		Pointer to an already allocated
				buffer. If NULL a new buffer is
				allocated with default size.

   Results:

        Returns the pointer to the base structure.

*/

More_SharedBuffer
More_SharedBufferAlloc (More_Buffer bufferPtr)
{
  SharedBuffer *	base;


  base = (SharedBuffer *) ckalloc(sizeof(SharedBuffer));
  memset(base, '\0', sizeof(SharedBuffer));

  base->buffer = (bufferPtr)? bufferPtr : More_BufferAlloc(NULL, 0);
  base->readerAttached	= 1;
  base->writerAttached	= 1;
  base->bufferSize	= DEFAULT_BUFFER_SIZE;
  return base;
}

/* More_SharedBufferFree --

	Releases a shared buffer.

   Arguments:

	buffer -	The buffer token.

   Results:

        Detaches both the reader and the writer.

*/

void
More_SharedBufferFree (More_SharedBuffer buffer)
{
  More_SharedBufferDetachReader(buffer);
  More_SharedBufferDetachWriter(buffer);
}

/* More_SharedBufferNotifyReader --

	Registers the callback used to notify the entity attached to the
	reader side that the buffer is readable.

   Arguments:

	bufferPtr -	Pointer to the buffer.
	function -	Pointer to the function.
	data -		Pointer to the function's argument.

   Results:

	Stores values  in the instance function fields;  if the function
	pointer is NULL: the callback is reset.

	  The  notification  function  is  invoked.  If  the  reader  is
	attached  and  data  is   available:  the  function  is  invoked
	immediately.

*/

void
More_SharedBufferNotifyReader (bufferPtr, function, data)
     More_SharedBuffer		bufferPtr;
     More_CallbackFunction *	function;
     ClientData			data;
{
  SharedBuffer *	base = bufferPtr;

  Tcl_MutexLock(&(base->mutex));
  More_CallbackInit(base->readerCallback, function, data);
  Tcl_MutexUnlock(&(base->mutex));
  SharedBufferNotify(base);
}

/* More_SharedBufferNotifyWriter --

	Registers the callback used to notify the entity attached to the
	input side that the buffer is writable.

   Arguments:

	bufferPtr -	Pointer to the buffer.
	function -	Pointer to the function.
	data -		Pointer to the function's argument.

   Results:

	Stores values  in the instance function fields:  if the function
	pointer is NULL: the callback is reset.

	  The  notification  function  is  invoked.  If  the  writer  is
	attached: the  function is invoked immediately  because a buffer
	is always writable.

*/

void
More_SharedBufferNotifyWriter (bufferPtr, function, data)
     More_SharedBuffer		bufferPtr;
     More_CallbackFunction *	function;
     ClientData			data;
{
  SharedBuffer *	base = bufferPtr;

  Tcl_MutexLock(&(base->mutex));
  More_CallbackInit(base->writerCallback, function, data);
  Tcl_MutexUnlock(&(base->mutex));
  SharedBufferNotify(base);
}

/* More_SharedBufferDetachReader --

	Detaches the sink entity from the buffer.

   Arguments:

	basePtr -	Pointer to the base buffer.

   Results:

        Resets  the reader  flag.  If the  source  has already  detached
        itself: the buffer is released.

*/

void
More_SharedBufferDetachReader (More_SharedBuffer basePtr)
{
  SharedBuffer *	base = (SharedBuffer *) basePtr;


  Tcl_MutexLock(&(base->mutex));
  debug("detaching reader from %p", base);
  base->readerAttached = 0;
  if (! base->writerAttached)
    {
      Tcl_MutexUnlock(&(base->mutex));
      SharedBufferFinal(base);
      return;
    }
  Tcl_MutexUnlock(&(base->mutex));
}

/* More_SharedBufferDetachWriter --

	Detaches the source entity from the buffer.

   Arguments:

	basePtr -	Pointer to the base buffer.

   Results:

        Resets  the  writer flag.   If  the  sink  has already  detached
        itself: the buffer is released.

*/

void
More_SharedBufferDetachWriter (More_SharedBuffer basePtr)
{
  SharedBuffer *	base = (SharedBuffer *) basePtr;


  Tcl_MutexLock(&(base->mutex));
  debug("detaching writer from %p", base);
  base->writerAttached = 0;
  if (! base->readerAttached)
    {
      Tcl_MutexUnlock(&(base->mutex));
      SharedBufferFinal(base);
      return;
    }
  Tcl_MutexUnlock(&(base->mutex));
}

/* SharedBufferFinal --

	Finalises a buffer releasing all the resources.

   Arguments:

	base -		Pointer to the buffer.

   Results:

        Finalises the mutex and frees memory.

*/

void
SharedBufferFinal (SharedBuffer *base)
{
  debug("finalising %p", base);
  Tcl_MutexFinalize(&(base->mutex));
  More_BufferFree(base->buffer);
  ckfree((char *) base);
  debug("finalised %p", base);  
}

/* More_SharedBufferRead --

	Wrapper for "More_BufferRead()".

   Arguments:

	basePtr -	The base buffer token.
	outLen -	Number of bytes to read.
	outPtr -	Pointer to the output block.

   Results:

        Returns the number of bytes read.

*/

int
More_SharedBufferRead (basePtr, outLen, outPtr)
     More_SharedBuffer 	basePtr;
     int		outLen;
     More_BytePtr	outPtr;
{
  SharedBuffer *	base = (SharedBuffer *) basePtr;
  int	e;

  Tcl_MutexLock(&(base->mutex));
  e = More_BufferRead(&(base->buffer), outLen, outPtr);
  base->readableEventQueued = 0;
  SharedBufferNotify(base);
  Tcl_MutexUnlock(&(base->mutex));
  return e;
}

/* More_SharedBufferWrite --

	Wrapper for "More_BufferWrite()". Writes data to the buffer.

   Arguments:

	basePtr -	Shared buffer token.
	inLen -		Number of bytes to write.
	inPtr -		Pointer to the input block.

   Results:

        Fills the buffer with data.

*/

void
More_SharedBufferWrite (basePtr, inLen, inPtr)
     More_SharedBuffer	basePtr;
     int		inLen;
     CONST More_BytePtr inPtr;
{
  SharedBuffer *	base = (SharedBuffer *) basePtr;


  Tcl_MutexLock(&(base->mutex));
  More_BufferWrite(base->buffer, inLen, inPtr, base->bufferSize);
  base->writableEventQueued = 0;
  SharedBufferNotify(base);
  Tcl_MutexUnlock(&(base->mutex));
}

/* More_SharedBufferEmpty --

	Tests if data is available in the buffer.

   Arguments:

	basePtr -	Shared buffer token.

   Results:

        Returns true if the buffer is empty, otherwise returns false.

*/

int
More_SharedBufferEmpty (More_SharedBuffer basePtr)
{
  SharedBuffer *	base = (SharedBuffer *) basePtr;
  int			e;


  Tcl_MutexLock(&(base->mutex));
  e = More_BufferEmpty(base->buffer);
  Tcl_MutexUnlock(&(base->mutex));
  return e;
}

/* SharedBufferNotify --

	Notifies entities  attached to the  input and output  ends about
	the state of the buffer.

   Arguments:

	base -		Pointer to the buffer.

   Results:

        For the  reader, if:  it is attached,  no notification  was done
        after the  last read  operation, there is  data, a  callback has
        been  registered: the  callback is  invoked;  a flag  is set  to
        signal the  notification: it will  be cleared by the  first read
        operation on the channel.

        For the  writer, if:  it is attached,  no notification  was done
        after the last write  operation, a callback has been registered:
        the  callback  is   invoked;  a  flag  is  set   to  signal  the
        notification: it will be cleared by the first write operation.

	Note  that notifications  to the  reader  are sent  even if  the
	writer has detached itself; the same for inverted roles. This is
	to allow the  reader and the writer to be  warned of the current
	state of the buffer.

*/

void
SharedBufferNotify (SharedBuffer *base)
{
  if (base->readerAttached &&
      (! base->readableEventQueued) &&
      (! More_BufferEmpty(base->buffer)) &&
      More_CallbackPresent(base->readerCallback))
    {
      base->readableEventQueued = 1;
      More_CallbackInvoke(base->readerCallback);
    }
  if (base->writerAttached &&
      (! base->writableEventQueued) &&
      More_CallbackPresent(base->writerCallback))
    {
      base->writableEventQueued = 1;
      More_CallbackInvoke(base->writerCallback);
    }
}

/* More_SharedBufferSetSize --

	Selects a new value for the next buffer allocated.

   Arguments:

	buffer -	Buffer token.
	bufferSize -	New buffer size.

   Results:

        Stores the value.

*/

void
More_SharedBufferSetSize (More_SharedBuffer buffer, size_t bufferSize)
{
  SharedBuffer *	base = (SharedBuffer *) buffer;

  Tcl_MutexLock(&(base->mutex));
  base->bufferSize = bufferSize;
  Tcl_MutexUnlock(&(base->mutex));
}

/* More_SharedBufferGetSize --

	Acquires the current buffer size.

   Arguments:

	buffer -	Buffer token.

   Results:

        Returns the buffer size.

*/

size_t
More_SharedBufferGetSize (More_SharedBuffer buffer)
{
  SharedBuffer *	base = (SharedBuffer *) buffer;
  size_t		bufferSize;

  Tcl_MutexLock(&(base->mutex));
  bufferSize = base->bufferSize;
  Tcl_MutexUnlock(&(base->mutex));
  return bufferSize;
}

/* More_SharedBufferEof --

	Tests  the "end  of file"  condition  on a  shared buffer.   The
	condition  is true  if the  writer  has detached  itself and  no
	unread data is in the buffer.

   Arguments:

	buffer -	Shared buffer token.

   Results:

        Returns true if the writer has detached itself.

*/

int
More_SharedBufferEof (More_SharedBuffer buffer)
{
  SharedBuffer *	base = (SharedBuffer *) buffer;
  int			bool;

  Tcl_MutexLock(&(base->mutex));
  bool = (! base->writerAttached) && More_BufferEmpty(base->buffer);
  Tcl_MutexUnlock(&(base->mutex));
  return bool;
}

/* More_SharedBufferAlive --

	Tests the  effect of writing data  to the buffer.  If the reader
	has detached itself the operation is useless.

   Arguments:

	buffer -	Shared buffer token.

   Results:

        Returns true if the reader has not detached itself.

*/

int
More_SharedBufferAlive (More_SharedBuffer buffer)
{
  SharedBuffer *	base = (SharedBuffer *) buffer;
  int			bool;

  Tcl_MutexLock(&(base->mutex));
  bool = base->readerAttached;
  Tcl_MutexUnlock(&(base->mutex));
  return bool;
}


/* end of file */
/*
Local Variables:
mode: c
page-delimiter: "^$"
End:
*/
