/*  streamtransform.c --
   
   Part of: tclmore
   Contents: stream transformation
   Date: Fri Mar 12, 2004
   
   Abstract
   
	This  module  implements the  transformation  channel used  with
	streams.  A  tclmore stream is defined  by the More_StreamDriver
	interface and can be used  with compression libraries: it is the
	base for bztcl and ztcl transformations.
   
   Copyright (c) 2004 Marco Maggi
   
   The author  hereby grant permission to use,  copy, modify, distribute,
   and  license this  software  and its  documentation  for any  purpose,
   provided that  existing copyright notices  are retained in  all copies
   and that  this notice  is included verbatim  in any  distributions. No
   written agreement, license, or royalty  fee is required for any of the
   authorized uses.  Modifications to this software may be copyrighted by
   their authors and need not  follow the licensing terms described here,
   provided that the new terms are clearly indicated on the first page of
   each file where they apply.
   
   IN NO  EVENT SHALL THE AUTHOR  OR DISTRIBUTORS BE LIABLE  TO ANY PARTY
   FOR  DIRECT, INDIRECT, SPECIAL,  INCIDENTAL, OR  CONSEQUENTIAL DAMAGES
   ARISING OUT  OF THE  USE OF THIS  SOFTWARE, ITS DOCUMENTATION,  OR ANY
   DERIVATIVES  THEREOF, EVEN  IF THE  AUTHOR  HAVE BEEN  ADVISED OF  THE
   POSSIBILITY OF SUCH DAMAGE.
   
   THE  AUTHOR  AND DISTRIBUTORS  SPECIFICALLY  DISCLAIM ANY  WARRANTIES,
   INCLUDING,   BUT   NOT  LIMITED   TO,   THE   IMPLIED  WARRANTIES   OF
   MERCHANTABILITY,    FITNESS   FOR    A    PARTICULAR   PURPOSE,    AND
   NON-INFRINGEMENT.  THIS  SOFTWARE IS PROVIDED  ON AN "AS  IS" BASIS,
   AND  THE  AUTHOR  AND  DISTRIBUTORS  HAVE  NO  OBLIGATION  TO  PROVIDE
   MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
   
*/


/** ------------------------------------------------------------
 ** Header files.
 ** ----------------------------------------------------------*/

#if 0
#  define TCLMORE_DEBUG_MESSAGES
#  define TCLMORE_DEBUGGING		1
#endif

#include "tclmoreInt.h"
#include <assert.h>
#include <errno.h>


/** ------------------------------------------------------------
 ** Module's type declarations.
 ** ----------------------------------------------------------*/

typedef struct Stream {
  ClientData		streamToken;
  int			eos;
  Tcl_TimerToken	timer;
} Stream;

typedef struct Transform {
  CONST More_StreamDriver *	driver;
  Stream			input;
  Stream			output;
  Tcl_Channel			channel;
  int				openModeMask;
  int				blocking;
  int				eventMask;
} Transform;


/** ------------------------------------------------------------
 ** Preprocessor symbols.
 ** ----------------------------------------------------------*/

#define debug			More_Debug
#define question		More_Question

#define min(a, b)		(((a) < (b))? (a) : (b))

#define D(T)			((T)->driver)


/** ------------------------------------------------------------
 ** Input stream macros.
 ** ----------------------------------------------------------*/

#define IStream(T)			(T)->input.streamToken

#define IStream_Exists(T)		(NULL != (IStream(T)))

#define IStream_GetAvailableOutput(T)	D(T)->output(IStream(T))
#define IStream_CommitRead(T, number)	D(T)->read(IStream(T), (number))
#define IStream_Write(T, block)		StreamWrite(T, &(T->input), block)
#define IStream_Flush(T)		D(T)->flush(IStream(T))
#define IStream_Finish(T)		D(T)->finish(IStream(T))
#define IStream_Final(T)		D(T)->final(IStream(T))


/** ------------------------------------------------------------
 ** Output stream macros.
 ** ----------------------------------------------------------*/

#define OStream(T)			(T)->output.streamToken

#define OStream_Exists(T)		(NULL != (OStream(T)))

#define OStream_GetAvailableOutput(T)	D(T)->output(OStream(T))
#define OStream_CommitRead(T, number)	D(T)->read(OStream(T), (number))
#define OStream_Write(T, block)		StreamWrite(T, &(T->output), block)
#define OStream_Flush(T)		D(T)->flush(OStream(T))
#define OStream_Finish(T)		D(T)->finish(OStream(T))
#define OStream_Final(T)		D(T)->final(OStream(T))


/** ------------------------------------------------------------
 ** Function prototypes.
 ** ----------------------------------------------------------*/

static Tcl_DriverCloseProc	StreamtranClose;
static Tcl_DriverInputProc	inputFunction;
static Tcl_DriverOutputProc	outputFunction;
static Tcl_DriverWatchProc	StreamtranWatch;
static Tcl_DriverBlockModeProc	StreamtranBlockMode;
static Tcl_DriverGetHandleProc	StreamtranGetHandle;
static Tcl_DriverSetOptionProc	StreamtranSetOption;
static Tcl_DriverGetOptionProc	StreamtranGetOption;
static Tcl_DriverHandlerProc	StreamtranHandler;

static void	NotifyReadability	_ANSI_ARGS_((ClientData));
static void	NotifyWritability	_ANSI_ARGS_((ClientData));

static int	StreamWrite		_ANSI_ARGS_((Transform *T,
						     Stream *stream,
						     More_Block blockPtr));

static More_ChannelDriverSetOptionProc SetOptionFlushOrFinishStream;
static More_ChannelDriverGetOptionProc GetOptionEndOfStream;


/** ------------------------------------------------------------
 ** Module's variable.
 ** ----------------------------------------------------------*/

static Tcl_ChannelType TransformDriver = {
  "tclmore_stream_transformation_version_1",
  TCL_CHANNEL_VERSION_3,
  StreamtranClose,
  inputFunction,
  outputFunction,
  NULL, /* Seek */
  StreamtranSetOption,
  StreamtranGetOption,
  StreamtranWatch,
  StreamtranGetHandle,
  NULL, /* Close2 */
  StreamtranBlockMode,
  NULL, /* Flush */
  StreamtranHandler,
  NULL /* WideSeek */
};

static CONST More_ChannelDriverOption optionTable[] = {
  { "-flush",  SetOptionFlushOrFinishStream, NULL },
  { "-finish", SetOptionFlushOrFinishStream, NULL },
  { "-endOfInputStream", NULL, GetOptionEndOfStream },
  { "-endOfOutputStream", NULL, GetOptionEndOfStream },
  { NULL, NULL, NULL }
};


Tcl_Channel
More_MakeStreamTransform(driver, input, output, subChannel)
     CONST More_StreamDriver *	driver;
     ClientData			input;
     ClientData			output;
     Tcl_Channel		subChannel;
{
  Transform * 	T;
  int		openModeMask = \
    ((input? TCL_READABLE : 0) | (output? TCL_WRITABLE : 0));
  

  debug("creating transform, readable %s, writable %s",
	question(openModeMask & TCL_READABLE),
	question(openModeMask & TCL_WRITABLE));

  More_AllocStruct(T, Transform);
  T->driver		= driver;
  T->input.streamToken	= input;
  T->output.streamToken	= output;
  T->openModeMask	= openModeMask;

  T->channel = Tcl_StackChannel(NULL, &TransformDriver, T, openModeMask, subChannel);
  assert(T->channel);
  debug("success: transform %p, subchannel %p, subchannel %p, name %s)",
	T->channel, Tcl_GetStackedChannel(T->channel),
	subChannel, Tcl_GetChannelName(subChannel));
  return T->channel;
}

static int
StreamtranClose (ClientData D, Tcl_Interp *interp)
{
  Transform * 	T = (Transform *) D;


  debug("finalising transformation");
  if (T->input.timer)
    {
      Tcl_DeleteTimerHandler(T->input.timer);
    }
  if (IStream(T))
    {
      IStream_Final(T);
    }
  if (OStream(T))
    {
      OStream_Final(T);
    }
  ckfree((char *) (T));
  return 0;
}

static int
inputFunction (ClientData D, char *requestedPtr, int requestedLen, int *errorCodePtr)
{
  Transform * 	T = (Transform *) D;
  More_Block	available;
  int		furnished;
  Tcl_Channel	subChannel = Tcl_GetStackedChannel(T->channel);
# undef		CHUNK_SIZE
# define	CHUNK_SIZE 4096
  char		buffer[CHUNK_SIZE];
  More_Block	block;
  More_Error	error;


  if (! (T->openModeMask & TCL_READABLE))
    {
      *errorCodePtr = EACCES;
      return -1;
    }

  available = IStream_GetAvailableOutput(T);
  debug("enter input function, requested %d bytes, available %d",
	requestedLen, available.len);

  while ((available.len < requestedLen) && (! Tcl_Eof(subChannel)))
    {
      block.ptr = buffer;
      block.len = Tcl_ReadRaw(subChannel, block.ptr, CHUNK_SIZE);
      debug("read %d bytes from channel", block.len);
      if (block.len < 0)
	{
	  *errorCodePtr = Tcl_GetErrno();
	  if ((! T->blocking) && (EAGAIN == *errorCodePtr))
	    {
	      break;
	    }
	  else
	    {
	      return block.len;	      
	    }
	}
      else if (block.len > 0)
	{
	  *errorCodePtr = IStream_Write(T, block);
	  if (*errorCodePtr)
	    {
	      return -1;
	    }
	  else if (T->input.eos)
	    {
	      break;
	    }
	}
      else if (T->blocking) /* block.len == 0 */
	{
	  break;
	}
      available = IStream_GetAvailableOutput(T);
      debug("available %d", available.len);
    }

/*   if (Tcl_Eof(subChannel) && (! T->input.eos)) */
  if (Tcl_Eof(subChannel) || T->input.eos)
    {
      debug("eof, finishing input stream");
      error = IStream_Finish(T);
      T->input.eos = 1;
      if (error)
	{
	  return More_ErrorCodeAndForget(error);
	}

      available = IStream_GetAvailableOutput(T);
    }

  furnished = min(available.len, requestedLen);
  if (furnished)
    {
      memcpy(requestedPtr, available.ptr, (size_t) furnished);
      IStream_CommitRead(T, furnished);
    }
  
  debug("returning %d bytes read from output buffer, left %d",
	furnished, available.len - furnished);

  if ((! T->blocking) && (0 == furnished))
    {
      *errorCodePtr = EAGAIN;
      return -1;
    }
  else
    {
      *errorCodePtr = 0;
      return furnished;
    }
}

static int
outputFunction (ClientData D, CONST char *bufPtr, int bufLen, int *errorCodePtr)
{
  Transform * 	T = (Transform *) D;
  More_Block	block;


  debug("enter output function, request to write %d bytes", bufLen);
  if (! (T->openModeMask & TCL_WRITABLE))
    {
      *errorCodePtr = EACCES;
      return -1;
    }


  if (bufLen)
    {
      block.len = bufLen;
      block.ptr = (More_Byte *) bufPtr;
      *errorCodePtr = OStream_Write(T, block);
      if (*errorCodePtr)
	{
	  return -1;
	}
    }


  block = OStream_GetAvailableOutput(T);
  debug("available output: %d bytes", block.len);
  if (block.len)
    {
      Tcl_Channel	subChannel = Tcl_GetStackedChannel(T->channel);
      int		written;


      if (T->blocking)
	{
	  debug("writing in loop");
	  do
	    {
	      written = Tcl_WriteRaw(subChannel, (char *)block.ptr, block.len);
	      if (written < 0) { break; }
	    }
	  while (written < block.len);
	}
      else
	{
	  written = Tcl_WriteRaw(subChannel, (char *)block.ptr, block.len);	  
	}

      if (written > 0)
	{
	  debug("written %d bytes to the underlying channel", written);
	  OStream_CommitRead(T, written);
	}
      else
	{
	  *errorCodePtr = Tcl_GetErrno();
	  if (T->blocking || (EAGAIN != *errorCodePtr))
	    {
	      debug("error writing channel: %s", Tcl_ErrnoMsg(Tcl_GetErrno()));
	      return written;
	    }
	  *errorCodePtr = 0;
	}
    }
  else 
    {
      if ((T->eventMask & TCL_WRITABLE) && (! T->output.timer))
	{
	  debug("queuing writability event");
	  T->output.timer = Tcl_CreateTimerHandler(1, NotifyWritability, T);
	}
    }

  debug("returning, signaling %d bytes read", bufLen);
  return bufLen;
}

static void
StreamtranWatch (ClientData D, int mask)
{
  Transform *		T = (Transform *) D;
  Tcl_Channel		subChannel;
  Tcl_DriverWatchProc *	subWatchProc;
  ClientData		subInstanceData;
  More_Block		block;


  debug("interest in: readability %s, writability %s",
	question(mask & TCL_READABLE), question(mask & TCL_WRITABLE));

  T->eventMask = T->openModeMask & mask;

  subChannel		= Tcl_GetStackedChannel(T->channel);
  subWatchProc		= Tcl_ChannelWatchProc(Tcl_GetChannelType(subChannel));
  subInstanceData	= Tcl_GetChannelInstanceData(subChannel);
  subWatchProc(subInstanceData, mask);

  if (T->openModeMask & TCL_READABLE)
    {
      block = IStream_GetAvailableOutput(T);
      if ((! (mask & TCL_READABLE)) || (! block.len))
	{
	  if (T->input.timer)
	    {
	      debug("deleting read timer");
	      Tcl_DeleteTimerHandler(T->input.timer);
	      T->input.timer = (Tcl_TimerToken) NULL;
	    }
	}
      else
	{
	  if (! T->input.timer)
	    {
	      debug("creating read timer");
	      T->input.timer = Tcl_CreateTimerHandler(1, NotifyReadability, T);
	    }
	}
    }

  if (T->openModeMask & TCL_WRITABLE)
    {
      if ((! (mask & TCL_WRITABLE)) && (T->output.timer))
	{
	  debug("deleting write timer");
	  Tcl_DeleteTimerHandler(T->output.timer);
	  T->output.timer = (Tcl_TimerToken) NULL;
	}
    }
}

static void
NotifyReadability (ClientData D)
{
  Transform * 	T = (Transform *) D;

  debug("notifying channel readability");
  Tcl_NotifyChannel(T->channel, TCL_READABLE);
  T->input.timer = NULL;
}
static void
NotifyWritability (ClientData D)
{
  Transform * 	T = (Transform *) D;

  debug("notifying channel readability");
  Tcl_NotifyChannel(T->channel, TCL_WRITABLE);
  T->output.timer = NULL;
}

static int
StreamtranBlockMode (ClientData D, int mode)
{
  Transform *	T = (Transform *) D;

  T->blocking = (TCL_MODE_BLOCKING == mode);
  debug("blocking? %s", question(T->blocking));
  return 0;
}

static int
StreamtranGetHandle (ClientData D, int direction, ClientData *handlePtr)
{
  return TCL_ERROR;
}

static int
StreamtranHandler (ClientData D, int eventMask)
{
  return eventMask;
}

static int
StreamtranSetOption (D, interp, optionName, optionValue)
     ClientData		D;
     Tcl_Interp *	interp;
     CONST char *	optionName;
     CONST char *	optionValue;
{

  Transform * T = (Transform *) D;
  CONST More_ChannelDriverOption *option;
  More_StreamIO	streams = { T->input.streamToken, T->output.streamToken };


  debug("setting option %s %s", optionName, optionValue);
  option = &(optionTable[0]);
  while (option->name)
    {
      if (strcmp(option->name, optionName) == 0)
	{
	  return (option->setOptionProc)? \
	    option->setOptionProc(D, interp, optionName, optionValue) : \
	    More_LogicErrorStr(interp, "unupported operation");
	}
      ++option;
    }

  return More_ChannelDriverSetOption(T->driver->optionTable, T->channel,
				     (ClientData) &streams,
				     interp, optionName, optionValue);
}

static int
StreamtranGetOption (D, interp, optionName, optionValue)
     ClientData		D;
     Tcl_Interp *	interp;
     CONST char *	optionName;
     Tcl_DString *	optionValue;
{
  Transform *T = (Transform *) D;
  CONST More_ChannelDriverOption *option;
  More_StreamIO streams = { T->input.streamToken, T->output.streamToken };


  if (optionName)
    {
      option = &(optionTable[0]);
      while (option->name)
	{
	  if (strcmp(option->name, optionName) == 0)
	    {
	      return (option->getOptionProc)? \
		option->getOptionProc(D, interp, optionName, optionValue) : \
		More_LogicErrorStr(interp, "unupported operation");
	    }
	  ++option;
	}
    }

  return More_ChannelDriverGetOption(T->driver->optionTable, T->channel,
				     (ClientData) &streams,
				     interp, optionName, optionValue);
}

static int
SetOptionFlushOrFinishStream (D, interp, optionName, optionValue)
     ClientData		D;
     Tcl_Interp *	interp;
     CONST char *	optionName;
     CONST char *	optionValue;
{
  Transform *	T = (Transform *) D;
  More_Error	error = NULL;
  int		errorCode, written;
  

  debug("flush or finish");

  if (! optionValue)
    {
      return More_LogicErrorStr(interp, "null option value");      
    }
  else if (strcmp(optionValue, "input") == 0)
    {
      if (IStream_Exists(T))
	{
	  if (strcmp(optionName, "-flush") == 0)
	    {
	      debug("flushing input");
	      error = IStream_Flush(T);
	    }
	  else
	    {
	      debug("finishing input (finished: %s)", question(T->input.eos));
	      error = IStream_Finish(T);
	    }
	}
      else
	{
	  return More_RuntimeErrorStr(interp, "this channel has no input stream");
	}
    }
  else if (strcmp(optionValue, "output") == 0)
    {
      if (OStream_Exists(T))
	{
	  if (strcmp(optionName, "-flush") == 0)
	    {
	      debug("flushing output");
	      error = OStream_Flush(T);
	    }
	  else
	    {
	      debug("finishing output (finished: %s)", question(T->output.eos));
	      error = OStream_Finish(T);	      
	    }
	  if (! error)
	    {
	      written = outputFunction((ClientData) T, NULL, 0, &errorCode);
	      if (written < 0)
		{
		  Tcl_SetErrno(errorCode);
		  error = More_ErrorErrno();
		}
	    }
	}
      else
	{
	  return More_RuntimeErrorStr(interp, "this channel has no output stream");
	}
    }
  else
    {
      return More_LogicErrorStr(interp, "bad option value");
    }

  debug("done");
  return (! error)? TCL_OK : \
    ((! interp)? TCL_ERROR : More_ErrorResult(interp, error));
}

static int
GetOptionEndOfStream (D, interp, optionName, optionValue)
     ClientData		D;
     Tcl_Interp *	interp;
     CONST char *	optionName;
     Tcl_DString *	optionValue;
{
  Transform *	T = (Transform *) D;
  Stream *	stream;


  /* "-endOfInputStream" or "-endOfOutputStream" */
  stream = ('I' == optionName[6])? &(T->input) : &(T->output);
  if (! stream)
    {
      return More_LogicErrorStr(interp, "requested stream not present");
    }
  Tcl_DStringAppend(optionValue, ((stream->eos)? "1" : "0"), -1);
  return TCL_OK;
}

static int
StreamWrite (Transform *T, Stream *stream, More_Block block)
{
  int		e = 0;
  More_Error	error;


  debug("writing %d bytes to stream", block.len);
  error = T->driver->write(stream->streamToken, &block);
  debug("written, unread %d bytes", block.len);
  if (error)
    {
      return More_ErrorCodeAndForget(error);
    }

  if (block.len)
    {
      stream->eos = 1;
      /* Ignore data past the end of the stream. */
      /* return EINVAL; */
    }
  return e;
}

/* end of file */
/*
Local Variables:
mode: c
page-delimiter: "^$"
End:
*/
