/* teechan.c --
   
   Part of: tclmore
   Contents: stacked channel interface for debugging purposes
   Date: Tue Jan  6, 2004
   
   Abstract
   
	A  teechan is  a  stacked channel  that  saves all  the data  in
	flowing in one direction in a variable, and all the data flowing
	in the other direction in another variable.
   
   Copyright (c) 2004 Marco Maggi
   
   The author  hereby grant permission to use,  copy, modify, distribute,
   and  license this  software  and its  documentation  for any  purpose,
   provided that  existing copyright notices  are retained in  all copies
   and that  this notice  is included verbatim  in any  distributions. No
   written agreement, license, or royalty  fee is required for any of the
   authorized uses.  Modifications to this software may be copyrighted by
   their authors and need not  follow the licensing terms described here,
   provided that the new terms are clearly indicated on the first page of
   each file where they apply.
   
   IN NO  EVENT SHALL THE AUTHOR  OR DISTRIBUTORS BE LIABLE  TO ANY PARTY
   FOR  DIRECT, INDIRECT, SPECIAL,  INCIDENTAL, OR  CONSEQUENTIAL DAMAGES
   ARISING OUT  OF THE  USE OF THIS  SOFTWARE, ITS DOCUMENTATION,  OR ANY
   DERIVATIVES  THEREOF, EVEN  IF THE  AUTHOR  HAVE BEEN  ADVISED OF  THE
   POSSIBILITY OF SUCH DAMAGE.
   
   THE  AUTHOR  AND DISTRIBUTORS  SPECIFICALLY  DISCLAIM ANY  WARRANTIES,
   INCLUDING,   BUT   NOT  LIMITED   TO,   THE   IMPLIED  WARRANTIES   OF
   MERCHANTABILITY,    FITNESS   FOR    A    PARTICULAR   PURPOSE,    AND
   NON-INFRINGEMENT.  THIS  SOFTWARE IS PROVIDED  ON AN "AS  IS" BASIS,
   AND  THE  AUTHOR  AND  DISTRIBUTORS  HAVE  NO  OBLIGATION  TO  PROVIDE
   MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
*/


/** ------------------------------------------------------------
 ** Header files.
 ** ----------------------------------------------------------*/

#include <string.h>
#include <errno.h>
#include <stddef.h>
#include <assert.h>

/*
#define TCLMORE_DEBUG_MESSAGES
#define TCLMORE_DEBUGGING	1
*/

#include "tclmoreInt.h"


/** ------------------------------------------------------------
 ** Module's type declarations.
 ** ----------------------------------------------------------*/

/* LogBuffer --

  Type  representing the  buffer used  to log  data flowing  through the
  layer.

  buffer -		Pointer   to   the    buffer.   If   NULL:   the
			transformation  is  not  logging  data  for  the
			associated direction.

  name -		Pointer to a dynamically allocated block used to
			store  the variable  name.  It  is here  only to
			respond to [fconfigure]  queries.

			  It must  not be replaced  with a pointer  to a
			shared Tcl_Obj: the variable and the channel may
			end up  to belong to  different threads, objects
			must not be shared among threads.

*/

typedef struct LogBuffer {
  More_Buffer	buffer;
  char *	name;
} LogBuffer;

/* Channel --

  Type representing the state of the tee channel.

  channel -		The  channel  token  associated to  the  stacked
			channel.

  input -		Input log variable.

  output -		Output log variable.

  openModeMask -	The open  mode of the stacked  channel: an OR-ed
			combination of TCL_READABLE and TCL_WRITABLE.


  Note. There is no "watching events mask" in this structure (a bit mask
  representing the events  of interest for the generic  layer). The mask
  handed  to the  watch function  by the  generic layer  is sent  to the
  underlying channel. The tee stacked channel does not handle file event
  requests.

*/

typedef struct Channel {
  Tcl_Channel	channel;
  LogBuffer	input;
  LogBuffer	output;
  int		openModeMask;
} Channel;


/** ------------------------------------------------------------
 ** Preprocessor symbols.
 ** ----------------------------------------------------------*/

#define debug			More_Debug
#define question		More_Question

#define LogBufferActive(logbuf)	((logbuf).buffer)


/** ------------------------------------------------------------
 ** Function prototypes.
 ** ----------------------------------------------------------*/

static Tcl_DriverCloseProc	TeechanClose;
static Tcl_DriverInputProc	TeechanInput;
static Tcl_DriverOutputProc	TeechanOutput;
static Tcl_DriverWatchProc	TeechanWatch;
static Tcl_DriverGetHandleProc	TeechanGetHandle;
static Tcl_DriverSetOptionProc	TeechanSetOption;
static Tcl_DriverGetOptionProc	TeechanGetOption;
static Tcl_DriverHandlerProc	TeechanHandler;

static int SetupLogBuffer	_ANSI_ARGS_((Tcl_Interp *interp,
					     CONST char *variableName,
					     LogBuffer *var));

static void LogBufferInit	_ANSI_ARGS_((LogBuffer *logBufPtr,
					     More_Variable variable));
static void LogBufferFinal	_ANSI_ARGS_((LogBuffer *logBufPtr));

static More_ChannelDriverSetOptionProc SetOptionLogVariable;
static More_ChannelDriverGetOptionProc GetOptionLogVariable;


/** ------------------------------------------------------------
 ** Module's variables.
 ** ----------------------------------------------------------*/

static Tcl_ChannelType Driver = {
  "tclmore_teechan",
  TCL_CHANNEL_VERSION_3,
  TeechanClose,
  TeechanInput,
  TeechanOutput,
  NULL, /* Seek */
  TeechanSetOption,
  TeechanGetOption,
  TeechanWatch,
  TeechanGetHandle,
  NULL, /* Close2 */
  NULL, /* Block */
  NULL, /* Flush */
  TeechanHandler,
  NULL /* WideSeek */
};

static CONST More_ChannelDriverOption optionTable[] = {
  {  "-invar", SetOptionLogVariable, GetOptionLogVariable },
  { "-outvar", SetOptionLogVariable, GetOptionLogVariable },
  { NULL, NULL, NULL }
};


/* More_StackTeeChannel --

	Stacks a new channel on top of a selected one.

   Arguments:

	interp -	The interpreter to which the channel is
			associated and into which error messages
			are set.
	subChannel -	The token of the underlying channel.
	modeMask -	An OR-ed combination of TCL_READABLE and
			TCL_WRITABLE.

   Results:

	Stacks the channel. Returns the channel token.

*/

Tcl_Channel
More_StackTeeChannel (interp, subChannel, modeMask)
     Tcl_Interp *	interp;
     Tcl_Channel	subChannel;
     int		modeMask;
{
  Channel *	data;


  debug("stacking channel on top of \"%s\"", Tcl_GetChannelName(subChannel));

  More_AllocStruct(data, Channel);
  data->openModeMask	= modeMask;
  data->channel		= Tcl_StackChannel(interp, &Driver, data,
					   modeMask, subChannel);

  debug("stacked channel, pointer \"%p\"", data->channel);
  return data->channel;
}

/* TeechanClose --

	Invoked  by  the  generic   layer  to  clean-up  driver  related
	informations when the channel is closed. 

   Arguments:

	D -		Pointer to the instance structure.
	interp -	If non-NULL: the interpreter used to report
			errors.

   Results:

	Frees all the resources  associated to the channel.  Must return
	zero if the  operation is successful, or a  non-zero POSIX error
	code if an error occurs; always return zero.

*/

int
TeechanClose (ClientData D, Tcl_Interp *interp)
{
  Channel *	data = D;


  debug("entering close for stacked channel \"%s\"",
	Tcl_GetChannelName(data->channel));
  if (LogBufferActive(data->input))
    {
      LogBufferFinal(&(data->input));
    }
  if (LogBufferActive(data->output))
    {
      LogBufferFinal(&(data->output));
    }
  ckfree((char *) data);
  debug("resources freed, return success");
  return 0;
}

/* TeechanInput --

	Invoked by the generic layer to read data from the device.

	  Reads raw bytes  from the underlying channel and  stores it in
	the output  block supplied  by the generic  layer.  The  data is
	also written to the buffer  linked to the input variable, if one
	is currently registered.

   Arguments:

	D -		Pointer to the instance structure.
	bufPtr -	Pointer to the output buffer.
	bufLen -	Number of bytes to be read.
	errorCodePtr -	Pointer to a variable that must be filled with
			a POSIX error code in case of error.

   Results:

        Must  return a  non-negative integer  indicating how  many bytes
        were read. In case of error  must return -1.  If an error occurs
        in the  underlying channel: the  same error is returned.

	  If an  error occurs  writing the variable:  EINVAL is  used as
        POSIX  error code. In  this case:  data has  been read  from the
        underlying channel and so it will be lost.

	  If  the log buffer  of the  input variable  is not  alive (the
	variable  was  unset  causing  the  buffer  to  have  no  reader
	attached) the writer is detached  from the buffer, causing it to
	be finalised.

*/

int
TeechanInput (ClientData D, char *bufPtr, int bufLen, int *errorCodePtr)
{
  Channel *	data = D;


  debug("entering input for channel \"%s\", requested %d bytes",
	Tcl_GetChannelName(data->channel), bufLen);

  /*
    We do this  check ourselves because TCL 8.4.5  does not register the
    open mode for transformations, so it does not block wrong requests.
  */
  if (! (data->openModeMask & TCL_READABLE))
    {
      debug("wrong request: input proc on non readable transformation");
      *errorCodePtr = EACCES;
      return -1;
    }

  bufLen = Tcl_ReadRaw(Tcl_GetStackedChannel(data->channel),
		       bufPtr, bufLen);
  if (bufLen < 0)
    {
      debug("error reading: %s", Tcl_ErrnoMsg(Tcl_GetErrno()));
      *errorCodePtr = Tcl_GetErrno();
      return -1;
    }

  if ((bufLen > 0) && (LogBufferActive(data->input)))
    {
      if (More_BufferAlive(data->input.buffer))
	{
	  More_Block	block;
	  
	  block.len = bufLen;
	  block.ptr = bufPtr;
	  More_BufferWrite(data->input.buffer, block);
	  debug("written %d bytes to input log buffer %p",
		block.len, data->input.buffer);
	}
      else
	{
	  debug("detaching writer from input log buffer %p",
		data->input.buffer);
	  LogBufferFinal(&(data->input));
	}
    }
  
  debug("leaving input function %d bytes, success", bufLen);
  *errorCodePtr = 0;
  return bufLen;
}

/* TeechanOutput --

	Invoked by the  generic layer to transfer data  from an internal
	buffer to the output device.

	  Writes  raw bytes  to the  underlying channel  from  the input
	block supplied by the generic  layer.  The data is also appended
	to the buffer linked to the output variable, if one is currently
	registered.

   Arguments:

	D -		Pointer to the instance structure.
	bufPtr -	Pointer to the input buffer.
	bufLen -	Number of bytes to be written.
	errorCodePtr -	Pointer to a variable that must be filled with
			a POSIX error code in case of error.

   Results:

	Must  return a  non-negative integer  indicating how  many bytes
        were  written. In  case of  error must  return -1.  If  an error
        occurs in the underlying channel the same error is returned.

	  If an  error occurs  writing the variable:  EINVAL is  used as
        POSIX error  code. In this case:  some data has  been written to
        the underlying channel with no errors.

	  If the  log buffer  of the output  variable is not  alive (the
	variable  was  unset  causing  the  buffer  to  have  no  reader
	attached) the writer is detached  from the buffer, causing it to
	be finalised.

*/

int
TeechanOutput (ClientData D, CONST char *bufPtr, int bufLen, int *errorCodePtr)
{
  Channel *	data = D;


  debug("entering output for channel \"%s\", requested writing %d chars",
	Tcl_GetChannelName(data->channel), bufLen);

  /*
    We do this  check ourselves because TCL 8.4.5  does not register the
    open mode for transformations, so it does not block wrong requests.
  */
  if (! (data->openModeMask & TCL_WRITABLE))
    {
      debug("wrong request: output proc on non writable transformation");
      *errorCodePtr = EACCES;
      return -1;
    }

  bufLen = Tcl_WriteRaw(Tcl_GetStackedChannel(data->channel),
			bufPtr, bufLen);
  if (bufLen < 0)
    {
      debug("error writing: %s", Tcl_ErrnoMsg(Tcl_GetErrno()));
      *errorCodePtr = Tcl_GetErrno();
      return -1;
    }

  debug("output buffer present? %s", question(data->output.buffer));
  if ((bufLen > 0) && (LogBufferActive(data->output)))
    {
      if (More_BufferAlive(data->output.buffer))
	{
	  More_Block	block;

	  block.len = bufLen;
	  block.ptr = (More_BytePtr) bufPtr;
	  More_BufferWrite(data->output.buffer, block);
	  debug("written %d bytes to output log buffer %p",
		block.len, data->output.buffer);
	}
      else
	{
	  debug("detaching writer from buffer %p", data->output.buffer);
	  LogBufferFinal(&(data->input));
	}
    }

  debug("leaving output function, success");
  *errorCodePtr = 0;      
  return bufLen;
}

/* TeechanWatch --

	Invoked by the generic layer to initialise the event notification
	mechanism.

   Arguments:

	D -		Pointer to the instance data.
	mask -		An OR-ed combination of TCL_READABLE, TCL_WRITABLE
			and TCL_EXCEPTION. It indicates events the caller
			is interested in.

   Results:

        If the  mask is null: this  function hands it  to the underlying
        channel.

	  If the mask is not null and the request is compatible with the
        open  mode of the  transformation: notifies  the request  to the
        underlying channel.

	  If the mask is not null and the request is not compatible with
        the open mode of the transformation: the request is removed from
        the mask, then the mask  is sent to the underlying channel. That
        way the  underlying channel can clear  registered event requests
        not compatible with the open mode of the transformation.

*/

void
TeechanWatch (ClientData D, int mask)
{
  Channel *		data = D;
  Tcl_DriverWatchProc *	subwatch;
  ClientData		subdata;
  Tcl_Channel		subChannel;


  debug("enabling events: readable %d, writable %d, exception %d",
	 (mask&TCL_READABLE), (mask&TCL_WRITABLE), (mask&TCL_EXCEPTION));

  if (mask)
    {
      mask &= data->openModeMask;
    }

  subChannel = Tcl_GetStackedChannel(data->channel);
  subwatch = Tcl_ChannelWatchProc(Tcl_GetChannelType(subChannel));
  if (NULL != subwatch)
    {
      subdata  = Tcl_GetChannelInstanceData(subChannel);
      debug("propagating event request to subdriver");
      subwatch(subdata, mask);
    }
}

/* TeechanGetHandle --

	Invoked by the generic layer to retrieve the device-specific
	handler for the channel.

   Arguments:

	D -		Pointer to the instance data.
	direction -	The channel direction: TCL_READABLE or
			TCL_WRITABLE.
	handlePtr -	Pointer to the variable that will hold
			the handler.

   Results:

        Returns TCL_ERROR.

*/

int
TeechanGetHandle (ClientData D, int direction, ClientData *handlePtr)
{
  return TCL_ERROR;
}

/* TeechanHandler --

	Invoked by the generic  layer to notify the transformation about
	events on the underlying channel.

   Arguments:

	D -		Pointer to the transformation instance.
	eventMask -	The event mask.

   Results:

        Returns the event mask unchanged.

*/

int
TeechanHandler (ClientData D, int eventMask)
{
  return eventMask;
}

/* TeechanSetOption --

	Invoked  by the  generic layer  to set  a channel  type specific
	option to the channel instance.

	  Supports the following options:

	"-invar VARNAME"	selects the input log variable; if the
				variable name is the empty string: the
				writer is detached from the input log
				buffer;

	"-outvar VARNAME"	selects the output log variable; if the
				variable name is the empty string: the
				writer is detached from the output log
				buffer.

   Arguments:

	D -		Pointer to the instance structure.
	interp -	If non-NULL: the interpreter to use to
			report errors.
	optionName -	The option name.
	newValue -	The option value.

   Results:

	Stores the new value in the instance data, creates the requested
	log buffers, creates and links the variables to the log buffers.

          Returns TCL_OK or TCL_ERROR. In case of error leaves a message
	in the interpreter's  result; if the error is  an invalid option
	value, sets "errno" to EINVAL.

*/

int
TeechanSetOption (D, interp, optionName, newValue)
     ClientData		D;
     Tcl_Interp *	interp;
     CONST char *	optionName;
     CONST char *	newValue;
{
  Channel *	data = D;

  return More_ChannelDriverSetOption(optionTable, data->channel,
				     D, interp, optionName, newValue);
}

/* SetOptionLogVariable --

	Configures a new log variable.

   Arguments:

	D -		Pointer to the instance structure.
	interp -	If not NULL, the interpreter in which
			report errors.
	optionName -	Pointer to the option string.
	newValue -	Pointer to a string representing the
			new variable name.

   Results:

        Returns TCL_OK or TCL_ERROR.

*/

int
SetOptionLogVariable (D, interp, optionName, newValue)
     ClientData		D;
     Tcl_Interp *	interp;
     CONST char *	optionName;
     CONST char *	newValue;
{
  Channel *	data = D;

  /* "-invar" or "-outvar" */
  return SetupLogBuffer(interp, newValue,
		('i' == optionName[1])? &(data->input) : &(data->output));
}

/* TeechanGetOption --

	Invoked by the generic layer  to retrieve the value of a channel
	specific  option.   See  TeechanSetOption()  for   the  list  of
	recognised options.

   Arguments:

	D -		Pointer to the instance structure.
	interp -	If non-NULL: the interpreter used to report
			errors.
	optionName -	The requested option name, NULL if 
	optionValue -	Pointer to the output string.

   Results:

	Appends to  the supplied  string the requested  values.  Returns
	TCL_OK or TCL_ERROR.

*/

int
TeechanGetOption (D, interp, optionName, optionValue)
     ClientData		D;
     Tcl_Interp *	interp;
     CONST char *	optionName;
     Tcl_DString *	optionValue;
{
  Channel *	data = D;

  return More_ChannelDriverGetOption(optionTable, data->channel,
				     D, interp, optionName, optionValue);
}

/* GetOptionLogVariable --

	Retrieves the name of the log variable.

   Arguments:

	D -		Pointer to the instance structure.
	interp -	If not NULL, the interpreter in which
			report errors.
	optionName -	Pointer to the option string.
	optionValue -	Pointer to a dynamic string that will
			hold the value.

   Results:

        Returns TCL_OK or TCL_ERROR.

*/

int
GetOptionLogVariable(D, interp, optionName, optionValue)
     ClientData		D;
     Tcl_Interp *	interp;
     CONST char *	optionName;
     Tcl_DString *	optionValue;
{
  Channel *	data = D;


  Tcl_DStringAppend(optionValue, (('i' == optionName[1])? \
		     ((data->input.name)?  data->input.name : "{}") :
		     ((data->output.name)? data->output.name : "{}")), -1);
  return TCL_OK;
}

/* SetupLogBuffer --

	Sets up the link between a variable and a channel.

	  If a log buffer is registered: the writer is detached from it;
	this allows a  log variable to be detached  from the channel and
	replaced with a new one.

	  If the variable name is  the empty string: no new log variable
	is associated to the channel.

   Arguments:

	interp-		The interpreter to which the variable
			belongs.
	name -		The variable name as string.
	logBufPtr -	Pointer to the log buffer structure.

   Results:

        Returns TCL_OK or TCL_ERROR. This  function can fail if an error
        occurs when testing the  accessibility of the selected variable;
        in this case "errno" is set to EINVAL.

*/

int
SetupLogBuffer (Tcl_Interp *interp, CONST char *name, LogBuffer *logBufPtr)
{
  Tcl_Obj *	nameObj;
  More_Variable	variable;
  int		e;


  if (LogBufferActive(*logBufPtr))
    {
      LogBufferFinal(logBufPtr);
    }

  if ((NULL == name) || (strlen(name) == 0))
    {
      debug("no variable");
      return TCL_OK;
    }

  nameObj = Tcl_NewStringObj(name, -1);
  More_VariableInit(variable, interp, nameObj);
  if (More_VariableExists(variable))
    {
      e = More_VariableTest(variable);
      if (TCL_OK != e)
	{
	  Tcl_SetErrno(EINVAL);
	  More_VariableFinal(variable);
	  return More_LogicError(interp);
	}
    }
  else
    {
      More_VariableAppend(variable, Tcl_NewStringObj(NULL, 0));
    }

  debug("linking tee channel to buffer");
  LogBufferInit(logBufPtr, variable);

  More_VariableFinal(variable);
  debug("done");
  return TCL_OK;
}

/* LogBufferInit --

	Initialises a log buffer in an already allocated base structure.

   Arguments:

	logBufPtr -		Pointer to the base structure.
	variable -		The variable.

   Results:

        Allocates resources.

*/

static void
LogBufferInit (LogBuffer *logBufPtr, More_Variable variable)
{
  logBufPtr->buffer = More_BufferAlloc();
  logBufPtr->name   = More_DupAllocString(More_VariableName(variable));
  More_CreateBufferVariable(variable, logBufPtr->buffer, NULL);
}

/* LogBufferFinal --

	Finalises a log buffer.

   Arguments:

	logBufPtr -	Pointer to the log buffer.

   Results:

        Releases all the resources, resets the base structure fields.

*/

static void
LogBufferFinal (LogBuffer *logBufPtr)
{
  assert(logBufPtr->buffer);
  assert(logBufPtr->name);

  debug("detaching buffer for %s", logBufPtr->name);
  More_BufferDetachWriter(logBufPtr->buffer);
  ckfree(logBufPtr->name);
  
  logBufPtr->buffer = NULL;
  logBufPtr->name   = NULL;
}


/* end of file */
/*
Local Variables:
mode: c
page-delimiter: "^$"
End:
*/
