/*
  This file is meant to be included (with #include) in a source
  file provided by you that implements the command functions. Your
  file must include the TCL header ("tcl.h") first and define the
  functions later.

    The reason of this inclusion is that here are declared the struct
  typedefs of the command invocation states. These types must be
  visible by the command functions. The stuff declared in this file
  uses the "static" keyword as much as possible.

    The command functions should be small: they should do the job by
  invoking other functions, the ones that the extension should export
  in its stub table.
*/


/** ------------------------------------------------------------
 ** Header files.
 ** ----------------------------------------------------------*/

#include "tclmore.h"



/** ------------------------------------------------------------
 ** Type declarations.
 ** ----------------------------------------------------------*/

/* OneState --

  This structure is used to hold the invocation state of the
  following commands:

                    [two   integer number]
*/
typedef struct OneState {
  int                     integer;
  double                  number;
  double                  unused1;
  double                  unused2;
  double                  unused3;
  double                  unused4;
  double                  unused5;
  double                  unused6;
} OneState;

static CONST OneState OneStateDefaults = {
  123, 1.2, 0, 1, 2, 3, 4, 5
};

/* TwoState --

  This structure is used to hold the invocation state of the
  following commands:

                    [three   ?options?]
                    [four   integer number ?options?]
                    [five pull more   integer number ?options?]
                    [five pull less   integer number ?options?]
*/
typedef struct TwoState {
  int                     integer;
  double                  number;
  int                     mode;
} TwoState;

static CONST TwoState TwoStateDefaults = {
  123, 1.2, 0
};

/* SixState --

  This structure is used to hold the invocation state of the
  following commands:

                    [six   mode]
*/
typedef struct SixState {
  int                     index;
} SixState;

static CONST SixState SixStateDefaults = {
  0
};

/* ObjState --

  This structure is used to hold the invocation state of the
  following commands:

                    [new_unsigned   unsigned]
                    [get_unsigned   unsigned]
                    [get_int_range   integer min max]
                    [get_wint_range   integer min max]
                    [get_unsigned_range   integer min max]
                    [new_sizet   size]
                    [get_sizet   size]
                    [new_float   float]
                    [get_float   float]
                    [get_float_range   float min max]
                    [makename]
*/
typedef struct ObjState {
  Tcl_Obj *               objPtr;
  Tcl_Obj *               objPtr1;
  Tcl_Obj *               objPtr2;
  Tcl_Obj *               objPtr3;
} ObjState;

static CONST ObjState ObjStateDefaults = {
  NULL, NULL, NULL, NULL
};

/* BufferState --

  This structure is used to hold the invocation state of the
  following commands:

                    [buffer]
*/
typedef struct BufferState {
  size_t                  buffer_size;
  int                     block_size;
  unsigned                times;
} BufferState;

static CONST BufferState BufferStateDefaults = {
  0, 0, 0
};

/* AsprintfState --

  This structure is used to hold the invocation state of the
  following commands:

                    [asprintf   format arg]
                    [objprintf   format arg]
                    [dsprintf   format arg]
*/
typedef struct AsprintfState {
  More_String             formatString;
  More_String             argumentString;
} AsprintfState;

static CONST AsprintfState AsprintfStateDefaults = {
  MORE_STRING_NULL_VALUE, MORE_STRING_NULL_VALUE
};

/* DScriptState --

  This structure is used to hold the invocation state of the
  following commands:

                    [dscript create   script]
                    [dscript queue   id]
                    [dscript abort   id value]
*/
typedef struct DScriptState {
  Tcl_Obj *               script;
  More_String             id;
  Tcl_Obj *               value;
} DScriptState;

static CONST DScriptState DScriptStateDefaults = {
  NULL, MORE_STRING_NULL_VALUE, NULL
};



/** ------------------------------------------------------------
 ** Implementation functions prototypes.
 ** ----------------------------------------------------------*/

/*
  You must provide the implementation of each of these functions in
  this file. When the function is invoked, the state invocation
  structure has been already filled with values from the command line
  arguments and options, with no errors.

    The function should look like the following:

  int
  MyCommandFunction (More_CommandFrame frame)
  {
     MyStateStruct *  state = (MyStateStruct *) frame->stateStructPtr;
     Tcl_Obj *        result;

     *
       Make use of the fields of "*state" to build a result by
       invoking functions from your library.
     *
     result = Mylib_DoSomething(state->value, state->another_value);

     Tcl_SetObjResult(frame->interp, result);
     return TCL_OK;
  }

    Being the implementation of a TCL command, this function is
  declared "static": it is useless to share this function among
  different modules. Remember that the role of a "More_CmdFunc"
  is to be an interface between the state invocation structure and
  the functions of your library that do the "real work".
*/

static More_CmdFunc         OneGreenFunction;
static More_CmdFunc         OneWhiteFunction;
static More_CmdFunc          OneRedDarkFunction;
static More_CmdFunc          OneRedBrightFunction;
static More_CmdFunc        TwoFunction;
static More_CmdFunc        ThreeFunction;
static More_CmdFunc        FourFunction;
static More_CmdFunc         FivePushFunction;
static More_CmdFunc          FivePullMoreFunction;
static More_CmdFunc          FivePullLessFunction;
static More_CmdFunc        SixFunction;
static More_CmdFunc        NewUnsignedFunction;
static More_CmdFunc        GetUnsignedFunction;
static More_CmdFunc        GetIntInRangeFunction;
static More_CmdFunc        GetWideIntInRangeFunction;
static More_CmdFunc        GetUnsignedInRangeFunction;
static More_CmdFunc        NewSizeTFunction;
static More_CmdFunc        GetSizeTFunction;
static More_CmdFunc        NewFloatFunction;
static More_CmdFunc        GetFloatFunction;
static More_CmdFunc        GetFloatInRangeFunction;
static More_CmdFunc        AsprintfFunction;
static More_CmdFunc        ObjprintfFunction;
static More_CmdFunc        DsprintfFunction;
static More_CmdFunc        MakenameFunction;
static More_CmdFunc        BadFunction;
static More_CmdFunc        BufferFunction;
static More_CmdFunc         DscriptCreateFunction;
static More_CmdFunc         DscriptQueueFunction;
static More_CmdFunc         DscriptAbortFunction;
static More_CmdFunc         DscriptFinalFunction;


/** ------------------------------------------------------------
 ** Extractor functions prototypes.
 ** ----------------------------------------------------------*/

/*
  You must provide the implementation of each of the static extractors
  in this file. Functions declared EXTERN can be implemented in other
  modules of your extension or in other libraries that are dependencies
  of this one (for example the TCLMORE library implements some common
  extractors).

    The functions must be available as pointers: you cannot directly use
  functions exported in a stub table; this is because the pointer must
  be included in a statically allocated data structure. Remember that
  stub function names are available as macros, each macro references
  an element in a vector of function pointers in another library: the
  value of the function pointer is available only at execution time.

    The only way to use an extractor from a stub table is to declare
  another extractor (with a similar name) that wraps the real function.
  Example:

     More_Extractor MyExtractor;
     ...
     int
     MyExtractor (More_ExtractorFrame frame)
     {
       return RealExtractor(frame);
     }

  "RealExtractor" can be the macro of a stub function.

    If a prototype is omitted: it is your responsibility to declare it
  before the line that includes this file in your source module. Obviously
  the declaration can be in an external header file; this is particularly
  useful if your extension is using a wrapper extractor for a real extractor
  from another extension, and the extractor is used in more than one
  set of commands.

    A common extractor has the following structure, lets assume that
  "Type" is the type of value we want:

     EXTERN int Mylib_GetTypeFromObj _ANSI_ARGS_((Tcl_Interp * interp,
                                                  Tcl_Obj *    srcObj,
                                                  Type *       valueVar));
     ...
     int
     TypeExtractor (More_ExtractorFrame frame)
     {
        return Mylib_GetTypeFromObj(frame->commandFrame->interp,
                                    frame->srcObj,
                                    (Type *) frame->fieldPtr);       
     }

  or:

     int
     TypeExtractor (More_ExtractorFrame frame)
     {
        int   e;

        e = Mylib_GetTypeFromObj(frame->commandFrame->interp,
                                 frame->srcObj,
                                 (Type *) frame->fieldPtr);       
        return ((TCL_OK == e)? e : More_LogicError(interp));
     }

  the implementation of a "GetTypeFromObj" function to do the real
  work is recommended because such functions can be used by other
  modules, and even exported in the stub table.

    Remember that the field pointer can reference a whole structure
  embedded in the state invocation one.

    At present an extractor can validate arguments only with complete
  independence from the current process state, that is it can test that
  a value is in the correct space and range. Coherence tests between
  argument values and process state must be performed by the command
  implementation functions. This is not exactly true, but it is the
  recommended way to do things.
*/

static More_Extractor      GetIntFromArg;
static More_Extractor      GetDoubleFromArg;
static More_Extractor      GetModeFromArg;
static More_Extractor      GetSixModeFromArg;
static More_Extractor      GetObjFromArg;
static More_Extractor      GetAStringFromArg;
static More_Extractor      GetSizeTFromArg;
static More_Extractor      GetUnsignedFromArg;


/** ------------------------------------------------------------
 ** Command arguments tables.
 ** ----------------------------------------------------------*/

/* No arguments for [one green]. */
#define one_green_ARGUMENT_NUMBER	0

/* No arguments for [one white]. */
#define one_white_ARGUMENT_NUMBER	0

/* No arguments for [one red dark]. */
#define one_red_dark_ARGUMENT_NUMBER	0

/* No arguments for [one red bright]. */
#define one_red_bright_ARGUMENT_NUMBER	0

#define two_ARGUMENT_NUMBER	2
static CONST More_CommandArgument TwoArguments[2] = {
  { GetIntFromArg,       offsetof(struct OneState, integer) },
  { GetDoubleFromArg,    offsetof(struct OneState, number) }
};

/* No arguments for [three]. */
#define three_ARGUMENT_NUMBER	0

#define four_ARGUMENT_NUMBER	2
static CONST More_CommandArgument FourArguments[2] = {
  { GetIntFromArg,       offsetof(struct TwoState, integer) },
  { GetDoubleFromArg,    offsetof(struct TwoState, number) }
};

/* No arguments for [five push]. */
#define five_push_ARGUMENT_NUMBER	0

#define five_pull_more_ARGUMENT_NUMBER	2
static CONST More_CommandArgument FivePullMoreArguments[2] = {
  { GetIntFromArg,       offsetof(struct TwoState, integer) },
  { GetDoubleFromArg,    offsetof(struct TwoState, number) }
};

#define five_pull_less_ARGUMENT_NUMBER	2
static CONST More_CommandArgument FivePullLessArguments[2] = {
  { GetIntFromArg,       offsetof(struct TwoState, integer) },
  { GetDoubleFromArg,    offsetof(struct TwoState, number) }
};

#define six_ARGUMENT_NUMBER	1
static CONST More_CommandArgument SixArguments[1] = {
  { GetSixModeFromArg,   offsetof(struct SixState, index) }
};

#define new_unsigned_ARGUMENT_NUMBER	1
static CONST More_CommandArgument New_unsignedArguments[1] = {
  { GetObjFromArg,       offsetof(struct ObjState, objPtr) }
};

#define get_unsigned_ARGUMENT_NUMBER	1
static CONST More_CommandArgument Get_unsignedArguments[1] = {
  { GetObjFromArg,       offsetof(struct ObjState, objPtr) }
};

#define get_int_range_ARGUMENT_NUMBER	3
static CONST More_CommandArgument Get_int_rangeArguments[3] = {
  { GetObjFromArg,       offsetof(struct ObjState, objPtr) },
  { GetObjFromArg,       offsetof(struct ObjState, objPtr1) },
  { GetObjFromArg,       offsetof(struct ObjState, objPtr2) }
};

#define get_wint_range_ARGUMENT_NUMBER	3
static CONST More_CommandArgument Get_wint_rangeArguments[3] = {
  { GetObjFromArg,       offsetof(struct ObjState, objPtr) },
  { GetObjFromArg,       offsetof(struct ObjState, objPtr1) },
  { GetObjFromArg,       offsetof(struct ObjState, objPtr2) }
};

#define get_unsigned_range_ARGUMENT_NUMBER	3
static CONST More_CommandArgument Get_unsigned_rangeArguments[3] = {
  { GetObjFromArg,       offsetof(struct ObjState, objPtr) },
  { GetObjFromArg,       offsetof(struct ObjState, objPtr1) },
  { GetObjFromArg,       offsetof(struct ObjState, objPtr2) }
};

#define new_sizet_ARGUMENT_NUMBER	1
static CONST More_CommandArgument New_sizetArguments[1] = {
  { GetObjFromArg,       offsetof(struct ObjState, objPtr) }
};

#define get_sizet_ARGUMENT_NUMBER	1
static CONST More_CommandArgument Get_sizetArguments[1] = {
  { GetObjFromArg,       offsetof(struct ObjState, objPtr) }
};

#define new_float_ARGUMENT_NUMBER	1
static CONST More_CommandArgument New_floatArguments[1] = {
  { GetObjFromArg,       offsetof(struct ObjState, objPtr) }
};

#define get_float_ARGUMENT_NUMBER	1
static CONST More_CommandArgument Get_floatArguments[1] = {
  { GetObjFromArg,       offsetof(struct ObjState, objPtr) }
};

#define get_float_range_ARGUMENT_NUMBER	3
static CONST More_CommandArgument Get_float_rangeArguments[3] = {
  { GetObjFromArg,       offsetof(struct ObjState, objPtr) },
  { GetObjFromArg,       offsetof(struct ObjState, objPtr1) },
  { GetObjFromArg,       offsetof(struct ObjState, objPtr2) }
};

#define asprintf_ARGUMENT_NUMBER	2
static CONST More_CommandArgument AsprintfArguments[2] = {
  { GetAStringFromArg,   offsetof(struct AsprintfState, formatString) },
  { GetAStringFromArg,   offsetof(struct AsprintfState, argumentString) }
};

#define objprintf_ARGUMENT_NUMBER	2
static CONST More_CommandArgument ObjprintfArguments[2] = {
  { GetAStringFromArg,   offsetof(struct AsprintfState, formatString) },
  { GetAStringFromArg,   offsetof(struct AsprintfState, argumentString) }
};

#define dsprintf_ARGUMENT_NUMBER	2
static CONST More_CommandArgument DsprintfArguments[2] = {
  { GetAStringFromArg,   offsetof(struct AsprintfState, formatString) },
  { GetAStringFromArg,   offsetof(struct AsprintfState, argumentString) }
};

#define makename_ARGUMENT_NUMBER	1
static CONST More_CommandArgument MakenameArguments[1] = {
  { GetObjFromArg,       offsetof(struct ObjState, objPtr) }
};

/* No arguments for [bad]. */
#define bad_ARGUMENT_NUMBER	0

/* No arguments for [buffer]. */
#define buffer_ARGUMENT_NUMBER	0

#define dscript_create_ARGUMENT_NUMBER	1
static CONST More_CommandArgument DscriptCreateArguments[1] = {
  { GetObjFromArg,       offsetof(struct DScriptState, script) }
};

#define dscript_queue_ARGUMENT_NUMBER	1
static CONST More_CommandArgument DscriptQueueArguments[1] = {
  { GetAStringFromArg,   offsetof(struct DScriptState, id) }
};

#define dscript_abort_ARGUMENT_NUMBER	1
static CONST More_CommandArgument DscriptAbortArguments[1] = {
  { GetAStringFromArg,   offsetof(struct DScriptState, id) }
};

/* No arguments for [dscript final]. */
#define dscript_final_ARGUMENT_NUMBER	0



/** ------------------------------------------------------------
 ** Command options tables.
 ** ----------------------------------------------------------*/

/* No options for [one green]. */

/* No options for [one white]. */

/* No options for [one red dark]. */

/* No options for [one red bright]. */

/* No options for [two]. */

static CONST More_CommandOption ThreeOptions[5] = {
  { "-integer", 1, GetIntFromArg, offsetof(struct TwoState, integer) },
  { "-number", 1, GetDoubleFromArg, offsetof(struct TwoState, number) },
  { "-on", 0, GetModeFromArg, offsetof(struct TwoState, mode) },
  { "-off", 0, GetModeFromArg, offsetof(struct TwoState, mode) },
  { NULL, 0, NULL, 0 }
};

static CONST More_CommandOption FourOptions[5] = {
  { "-integer", 1, GetIntFromArg, offsetof(struct TwoState, integer) },
  { "-number", 1, GetDoubleFromArg, offsetof(struct TwoState, number) },
  { "-on", 0, GetModeFromArg, offsetof(struct TwoState, mode) },
  { "-off", 0, GetModeFromArg, offsetof(struct TwoState, mode) },
  { NULL, 0, NULL, 0 }
};

/* No options for [five push]. */

static CONST More_CommandOption FivePullMoreOptions[5] = {
  { "-integer", 1, GetIntFromArg, offsetof(struct TwoState, integer) },
  { "-number", 1, GetDoubleFromArg, offsetof(struct TwoState, number) },
  { "-on", 0, GetModeFromArg, -1 },
  { "-off", 0, GetModeFromArg, -1 },
  { NULL, 0, NULL, 0 }
};

static CONST More_CommandOption FivePullLessOptions[5] = {
  { "-integer", 1, GetIntFromArg, offsetof(struct TwoState, integer) },
  { "-number", 1, GetDoubleFromArg, offsetof(struct TwoState, number) },
  { "-on", 0, GetModeFromArg, -1 },
  { "-off", 0, GetModeFromArg, -1 },
  { NULL, 0, NULL, 0 }
};

/* No options for [six]. */

/* No options for [new_unsigned]. */

/* No options for [get_unsigned]. */

/* No options for [get_int_range]. */

/* No options for [get_wint_range]. */

/* No options for [get_unsigned_range]. */

/* No options for [new_sizet]. */

/* No options for [get_sizet]. */

/* No options for [new_float]. */

/* No options for [get_float]. */

/* No options for [get_float_range]. */

/* No options for [asprintf]. */

/* No options for [objprintf]. */

/* No options for [dsprintf]. */

/* No options for [makename]. */

/* No options for [bad]. */

static CONST More_CommandOption BufferOptions[4] = {
  { "-bufferSize", 1, GetSizeTFromArg, offsetof(struct BufferState, buffer_size) },
  { "-blockSize", 1, GetIntFromArg, offsetof(struct BufferState, block_size) },
  { "-repeatTimes", 1, GetUnsignedFromArg, offsetof(struct BufferState, times) },
  { NULL, 0, NULL, 0 }
};

/* No options for [dscript create]. */

static CONST More_CommandOption DscriptQueueOptions[2] = {
  { "-value", 1, GetObjFromArg, offsetof(struct DScriptState, value) },
  { NULL, 0, NULL, 0 }
};

/* No options for [dscript abort]. */

/* No options for [dscript final]. */


/** ------------------------------------------------------------
 ** Leaf commands array.
 ** ----------------------------------------------------------*/

/*
  There is a leaf structure for each "More_CmdFunc". Each structure
  has references to the argument and option tables. It is not your
  business to mess with this structure.
*/

static More_CommandTreeLeaf commandTreeLeaves[] = {
  /* (0) synopsis: one green  */
  { 2, OneGreenFunction, 0, (CONST ClientData)NULL,
    one_green_ARGUMENT_NUMBER, NULL, NULL,
    NULL },

  /* (1) synopsis: one white  */
  { 2, OneWhiteFunction, 0, (CONST ClientData)NULL,
    one_white_ARGUMENT_NUMBER, NULL, NULL,
    NULL },

  /* (2) synopsis: one red dark  */
  { 3, OneRedDarkFunction, 0, (CONST ClientData)NULL,
    one_red_dark_ARGUMENT_NUMBER, NULL, NULL,
    NULL },

  /* (3) synopsis: one red bright  */
  { 3, OneRedBrightFunction, 0, (CONST ClientData)NULL,
    one_red_bright_ARGUMENT_NUMBER, NULL, NULL,
    NULL },

  /* (4) synopsis: two integer number */
  { 1, TwoFunction, sizeof(OneState), (CONST ClientData)&OneStateDefaults,
    two_ARGUMENT_NUMBER, TwoArguments, NULL,
    "integer number" },

  /* (5) synopsis: three ?options? */
  { 1, ThreeFunction, sizeof(TwoState), (CONST ClientData)&TwoStateDefaults,
    three_ARGUMENT_NUMBER, NULL, ThreeOptions,
    "?options?" },

  /* (6) synopsis: four integer number ?options? */
  { 1, FourFunction, sizeof(TwoState), (CONST ClientData)&TwoStateDefaults,
    four_ARGUMENT_NUMBER, FourArguments, FourOptions,
    "integer number ?options?" },

  /* (7) synopsis: five push  */
  { 2, FivePushFunction, 0, (CONST ClientData)NULL,
    five_push_ARGUMENT_NUMBER, NULL, NULL,
    NULL },

  /* (8) synopsis: five pull more integer number ?options? */
  { 3, FivePullMoreFunction, sizeof(TwoState), (CONST ClientData)&TwoStateDefaults,
    five_pull_more_ARGUMENT_NUMBER, FivePullMoreArguments, FivePullMoreOptions,
    "integer number ?options?" },

  /* (9) synopsis: five pull less integer number ?options? */
  { 3, FivePullLessFunction, sizeof(TwoState), (CONST ClientData)&TwoStateDefaults,
    five_pull_less_ARGUMENT_NUMBER, FivePullLessArguments, FivePullLessOptions,
    "integer number ?options?" },

  /* (10) synopsis: six mode */
  { 1, SixFunction, sizeof(SixState), (CONST ClientData)&SixStateDefaults,
    six_ARGUMENT_NUMBER, SixArguments, NULL,
    "mode" },

  /* (11) synopsis: new_unsigned unsigned */
  { 1, NewUnsignedFunction, sizeof(ObjState), (CONST ClientData)&ObjStateDefaults,
    new_unsigned_ARGUMENT_NUMBER, New_unsignedArguments, NULL,
    "unsigned" },

  /* (12) synopsis: get_unsigned unsigned */
  { 1, GetUnsignedFunction, sizeof(ObjState), (CONST ClientData)&ObjStateDefaults,
    get_unsigned_ARGUMENT_NUMBER, Get_unsignedArguments, NULL,
    "unsigned" },

  /* (13) synopsis: get_int_range integer min max */
  { 1, GetIntInRangeFunction, sizeof(ObjState), (CONST ClientData)&ObjStateDefaults,
    get_int_range_ARGUMENT_NUMBER, Get_int_rangeArguments, NULL,
    "integer min max" },

  /* (14) synopsis: get_wint_range integer min max */
  { 1, GetWideIntInRangeFunction, sizeof(ObjState), (CONST ClientData)&ObjStateDefaults,
    get_wint_range_ARGUMENT_NUMBER, Get_wint_rangeArguments, NULL,
    "integer min max" },

  /* (15) synopsis: get_unsigned_range integer min max */
  { 1, GetUnsignedInRangeFunction, sizeof(ObjState), (CONST ClientData)&ObjStateDefaults,
    get_unsigned_range_ARGUMENT_NUMBER, Get_unsigned_rangeArguments, NULL,
    "integer min max" },

  /* (16) synopsis: new_sizet size */
  { 1, NewSizeTFunction, sizeof(ObjState), (CONST ClientData)&ObjStateDefaults,
    new_sizet_ARGUMENT_NUMBER, New_sizetArguments, NULL,
    "size" },

  /* (17) synopsis: get_sizet size */
  { 1, GetSizeTFunction, sizeof(ObjState), (CONST ClientData)&ObjStateDefaults,
    get_sizet_ARGUMENT_NUMBER, Get_sizetArguments, NULL,
    "size" },

  /* (18) synopsis: new_float float */
  { 1, NewFloatFunction, sizeof(ObjState), (CONST ClientData)&ObjStateDefaults,
    new_float_ARGUMENT_NUMBER, New_floatArguments, NULL,
    "float" },

  /* (19) synopsis: get_float float */
  { 1, GetFloatFunction, sizeof(ObjState), (CONST ClientData)&ObjStateDefaults,
    get_float_ARGUMENT_NUMBER, Get_floatArguments, NULL,
    "float" },

  /* (20) synopsis: get_float_range float min max */
  { 1, GetFloatInRangeFunction, sizeof(ObjState), (CONST ClientData)&ObjStateDefaults,
    get_float_range_ARGUMENT_NUMBER, Get_float_rangeArguments, NULL,
    "float min max" },

  /* (21) synopsis: asprintf format arg */
  { 1, AsprintfFunction, sizeof(AsprintfState), (CONST ClientData)&AsprintfStateDefaults,
    asprintf_ARGUMENT_NUMBER, AsprintfArguments, NULL,
    "format arg" },

  /* (22) synopsis: objprintf format arg */
  { 1, ObjprintfFunction, sizeof(AsprintfState), (CONST ClientData)&AsprintfStateDefaults,
    objprintf_ARGUMENT_NUMBER, ObjprintfArguments, NULL,
    "format arg" },

  /* (23) synopsis: dsprintf format arg */
  { 1, DsprintfFunction, sizeof(AsprintfState), (CONST ClientData)&AsprintfStateDefaults,
    dsprintf_ARGUMENT_NUMBER, DsprintfArguments, NULL,
    "format arg" },

  /* (24) synopsis: makename  */
  { 1, MakenameFunction, sizeof(ObjState), (CONST ClientData)&ObjStateDefaults,
    makename_ARGUMENT_NUMBER, MakenameArguments, NULL,
    NULL },

  /* (25) synopsis: bad  */
  { 1, BadFunction, 0, (CONST ClientData)NULL,
    bad_ARGUMENT_NUMBER, NULL, NULL,
    NULL },

  /* (26) synopsis: buffer  */
  { 1, BufferFunction, sizeof(BufferState), (CONST ClientData)&BufferStateDefaults,
    buffer_ARGUMENT_NUMBER, NULL, BufferOptions,
    NULL },

  /* (27) synopsis: dscript create script */
  { 2, DscriptCreateFunction, sizeof(DScriptState), (CONST ClientData)&DScriptStateDefaults,
    dscript_create_ARGUMENT_NUMBER, DscriptCreateArguments, NULL,
    "script" },

  /* (28) synopsis: dscript queue id */
  { 2, DscriptQueueFunction, sizeof(DScriptState), (CONST ClientData)&DScriptStateDefaults,
    dscript_queue_ARGUMENT_NUMBER, DscriptQueueArguments, DscriptQueueOptions,
    "id" },

  /* (29) synopsis: dscript abort id value */
  { 2, DscriptAbortFunction, sizeof(DScriptState), (CONST ClientData)&DScriptStateDefaults,
    dscript_abort_ARGUMENT_NUMBER, DscriptAbortArguments, NULL,
    "id value" },

  /* (30) synopsis: dscript final  */
  { 2, DscriptFinalFunction, 0, (CONST ClientData)NULL,
    dscript_final_ARGUMENT_NUMBER, NULL, NULL,
    "" }
};


/** ------------------------------------------------------------
 ** Node commands array.
 ** ----------------------------------------------------------*/

/*
  These are node structures that declare subcommands with subcommands.
  It is not your business to mess with this structure.
*/

static More_CommandTreeNode OneRedCommandNode[] = {
  { "dark", TCLMORE_SAFE_COMMAND,  NULL, &commandTreeLeaves[2] },
  { "bright", TCLMORE_SAFE_COMMAND,  NULL, &commandTreeLeaves[3] },
  { NULL, TCLMORE_UNSAFE_COMMAND, NULL, NULL }
};

static More_CommandTreeNode OneCommandNode[] = {
  { "green", TCLMORE_SAFE_COMMAND,  NULL, &commandTreeLeaves[0] },
  { "white", TCLMORE_SAFE_COMMAND,  NULL, &commandTreeLeaves[1] },
  { "red", TCLMORE_SAFE_COMMAND,  OneRedCommandNode, NULL },
  { NULL, TCLMORE_UNSAFE_COMMAND, NULL, NULL }
};

static More_CommandTreeNode FivePullCommandNode[] = {
  { "more", TCLMORE_SAFE_COMMAND,  NULL, &commandTreeLeaves[8] },
  { "less", TCLMORE_UNSAFE_COMMAND,  NULL, &commandTreeLeaves[9] },
  { NULL, TCLMORE_UNSAFE_COMMAND, NULL, NULL }
};

static More_CommandTreeNode FiveCommandNode[] = {
  { "push", TCLMORE_SAFE_COMMAND,  NULL, &commandTreeLeaves[7] },
  { "pull", TCLMORE_SAFE_COMMAND,  FivePullCommandNode, NULL },
  { NULL, TCLMORE_UNSAFE_COMMAND, NULL, NULL }
};

static More_CommandTreeNode DscriptCommandNode[] = {
  { "create", TCLMORE_SAFE_COMMAND,  NULL, &commandTreeLeaves[27] },
  { "queue", TCLMORE_SAFE_COMMAND,  NULL, &commandTreeLeaves[28] },
  { "abort", TCLMORE_SAFE_COMMAND,  NULL, &commandTreeLeaves[29] },
  { "final", TCLMORE_SAFE_COMMAND,  NULL, &commandTreeLeaves[30] },
  { NULL, TCLMORE_UNSAFE_COMMAND, NULL, NULL }
};



/** ------------------------------------------------------------
 ** Root commands table.
 ** ----------------------------------------------------------*/

/*
  This array of structures declares the root commands. There is one of
  them for each <maincommand> and <command> markups with no ancestors
  in the original XML document. It is not your business to mess with
  the content of this array.

    You may create commands declared in this array with the following
  chunk of code:

    error = More_CreateCommands(interp, "namesp", More_TestCommands);
    if (error)
      {
	return More_ErrorResult(interp, error);
      }

  where "namesp" is the name of the namespace in which the command
  must be created. If the namespace is the root one: you can use NULL.
*/

More_CommandTreeNode More_TestCommands[] = {
  { "one", TCLMORE_SAFE_COMMAND, OneCommandNode, NULL },
  { "two", TCLMORE_SAFE_COMMAND, NULL, &commandTreeLeaves[4] },
  { "three", TCLMORE_SAFE_COMMAND, NULL, &commandTreeLeaves[5] },
  { "four", TCLMORE_SAFE_COMMAND, NULL, &commandTreeLeaves[6] },
  { "five", TCLMORE_SAFE_COMMAND, FiveCommandNode, NULL },
  { "six", TCLMORE_SAFE_COMMAND, NULL, &commandTreeLeaves[10] },
  { "new_unsigned", TCLMORE_SAFE_COMMAND, NULL, &commandTreeLeaves[11] },
  { "get_unsigned", TCLMORE_SAFE_COMMAND, NULL, &commandTreeLeaves[12] },
  { "get_int_range", TCLMORE_SAFE_COMMAND, NULL, &commandTreeLeaves[13] },
  { "get_wint_range", TCLMORE_SAFE_COMMAND, NULL, &commandTreeLeaves[14] },
  { "get_unsigned_range", TCLMORE_SAFE_COMMAND, NULL, &commandTreeLeaves[15] },
  { "new_sizet", TCLMORE_SAFE_COMMAND, NULL, &commandTreeLeaves[16] },
  { "get_sizet", TCLMORE_SAFE_COMMAND, NULL, &commandTreeLeaves[17] },
  { "new_float", TCLMORE_SAFE_COMMAND, NULL, &commandTreeLeaves[18] },
  { "get_float", TCLMORE_SAFE_COMMAND, NULL, &commandTreeLeaves[19] },
  { "get_float_range", TCLMORE_SAFE_COMMAND, NULL, &commandTreeLeaves[20] },
  { "asprintf", TCLMORE_SAFE_COMMAND, NULL, &commandTreeLeaves[21] },
  { "objprintf", TCLMORE_SAFE_COMMAND, NULL, &commandTreeLeaves[22] },
  { "dsprintf", TCLMORE_SAFE_COMMAND, NULL, &commandTreeLeaves[23] },
  { "makename", TCLMORE_SAFE_COMMAND, NULL, &commandTreeLeaves[24] },
  { "bad", TCLMORE_UNSAFE_COMMAND, NULL, &commandTreeLeaves[25] },
  { "buffer", TCLMORE_SAFE_COMMAND, NULL, &commandTreeLeaves[26] },
  { "dscript", TCLMORE_SAFE_COMMAND, DscriptCommandNode, NULL },
  { NULL, 0, NULL, NULL }
};


/* end of file */
/*
Local Variables:
mode: c
page-delimiter: "^$"
End:
*/

