/* variable.c --
   
   Part of: tclmore
   Contents: TCL variables functions
   Date: Tue Feb  3, 2004
   
   Abstract
   
	This  module   defines  an  interface  to  variable   in  a  TCL
	interpreter. The  main characteristic of this  interface is that
	error in acces to variables are ignored.
   
   Copyright (c) 2004 Marco Maggi
   
   The author  hereby grant permission to use,  copy, modify, distribute,
   and  license this  software  and its  documentation  for any  purpose,
   provided that  existing copyright notices  are retained in  all copies
   and that  this notice  is included verbatim  in any  distributions. No
   written agreement, license, or royalty  fee is required for any of the
   authorized uses.  Modifications to this software may be copyrighted by
   their authors and need not  follow the licensing terms described here,
   provided that the new terms are clearly indicated on the first page of
   each file where they apply.
   
   IN NO  EVENT SHALL THE AUTHOR  OR DISTRIBUTORS BE LIABLE  TO ANY PARTY
   FOR  DIRECT, INDIRECT, SPECIAL,  INCIDENTAL, OR  CONSEQUENTIAL DAMAGES
   ARISING OUT  OF THE  USE OF THIS  SOFTWARE, ITS DOCUMENTATION,  OR ANY
   DERIVATIVES  THEREOF, EVEN  IF THE  AUTHOR  HAVE BEEN  ADVISED OF  THE
   POSSIBILITY OF SUCH DAMAGE.
   
   THE  AUTHOR  AND DISTRIBUTORS  SPECIFICALLY  DISCLAIM ANY  WARRANTIES,
   INCLUDING,   BUT   NOT  LIMITED   TO,   THE   IMPLIED  WARRANTIES   OF
   MERCHANTABILITY,    FITNESS   FOR    A    PARTICULAR   PURPOSE,    AND
   NON-INFRINGEMENT.  THIS  SOFTWARE IS PROVIDED  ON AN "AS  IS" BASIS,
   AND  THE  AUTHOR  AND  DISTRIBUTORS  HAVE  NO  OBLIGATION  TO  PROVIDE
   MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
   
   $Id: variable.c,v 1.1.1.2 2004/02/03 15:44:37 marco Exp marco $
*/


/** ------------------------------------------------------------
 ** Header files.
 ** ----------------------------------------------------------*/

#include "tclmoreInt.h"

#define VARIABLE	variable[0]


/* More_VariableInit --

	Initialises a new instance.

   Arguments:

	variable -	Pointer to the instance.
	interp -	The interpreter to which the
			variable belongs.
	variableName -	Pointer to an object holding
			the variable name.

   Results:

        Fills the instance fields. Acquires a reference to
        the name object.

*/

void
More_VariableInit (variable, interp, variableName)
     More_Variable 	variable;
     Tcl_Interp *	interp;
     Tcl_Obj *		variableName;
{
  VARIABLE.interp = interp;
  VARIABLE.name   = variableName;
  Tcl_IncrRefCount(variableName);
}

/* More_VariableCopy --

	Duplicates an instance.

   Arguments:

	dest -		Pointer to the destination structure.
	source -	Pointer to the source structure.

   Results:

        Duplicates the fields.

*/

void
More_VariableCopy (More_Variable dest, More_Variable source)
{
  dest->interp = source->interp;
  dest->name   = source->name;
  Tcl_IncrRefCount(dest->name);
}

/* More_VariableFinal --

	Finalise an instance.

   Arguments:

	variable -	Pointer to the instance.

   Results:

        Releases the resources.

*/

void
More_VariableFinal (More_Variable variable)
{
  Tcl_DecrRefCount(VARIABLE.name);
}

/* More_VariableTest --

	Tests access to a variable.

   Arguments:

	variable -	Pointer to the instance.

   Results:

        Returns   TCL_OK  if  no   error  occurred,   otherwise  returns
        TCL_ERROR. This test will fail if the variable does not exist.

*/

int
More_VariableTest (More_Variable variable)
{
  Tcl_Obj *	objPtr;


  objPtr = Tcl_ObjGetVar2(VARIABLE.interp, VARIABLE.name, NULL,
			  TCL_LEAVE_ERR_MSG);
  return ((objPtr)? TCL_OK : TCL_ERROR);
}

/* More_VariableExists --

	Tests  if  a variable  exists  in  the  current context  of  the
	interpreter.

   Arguments:

	variable -	Pointer to the instance.

   Results:

        Returns true if the variable exists.

*/

int
More_VariableExists (More_Variable variable)
{
  Tcl_Obj *	objPtr;
  int		e, bool;


  objPtr = More_ObjPrintf("info exists %s", Tcl_GetString(VARIABLE.name));
  Tcl_IncrRefCount(objPtr);
  e = Tcl_EvalObjEx(VARIABLE.interp, objPtr, 0);
  Tcl_DecrRefCount(objPtr);
  if (TCL_OK != e)
    {
      return 0;
    }
  objPtr = Tcl_GetObjResult(VARIABLE.interp);
  e = Tcl_GetBooleanFromObj(VARIABLE.interp, objPtr, &bool);
  if (TCL_OK != e)
    {
      return 0;
    }
  return bool;
}

/* More_VariableSet --

	Stores a new value in a variable, overwriting the old one.

   Arguments:

	variable -	Pointer to the variable.
	objPtr -	Pointer to the new object.

   Results:

        Stores the value. Errors are ignored.

*/

void
More_VariableSet (More_Variable variable, Tcl_Obj *objPtr)
{
  Tcl_ObjSetVar2(VARIABLE.interp, VARIABLE.name, NULL, objPtr, 0);
}

/* More_VariableSetBlock --

	Stores a new byte array in a variable, overwriting the old one.

   Arguments:

	variable -	Pointer to the variable.
	block -		The data block.

   Results:

        Stores the value. Errors are ignored.

*/

void
More_VariableSetBlock (More_Variable variable, More_Block block)
{
  Tcl_Obj *	objPtr = Tcl_NewByteArrayObj(block.ptr, block.len);
  Tcl_IncrRefCount(objPtr);
  Tcl_ObjSetVar2(VARIABLE.interp, VARIABLE.name, NULL, objPtr, 0);
  Tcl_DecrRefCount(objPtr);
}

/* More_VariableAppend --

	Appends a value to a variable.

   Arguments:

	variable -	Pointer to the variable.
	objPtr -	Pointer to the new object.

   Results:

        Appends the value. Errors are ignored.
	
*/

void
More_VariableAppend (More_Variable variable, Tcl_Obj *objPtr)
{
  Tcl_ObjSetVar2(VARIABLE.interp, VARIABLE.name, NULL, objPtr,
		 TCL_APPEND_VALUE);
}

/* More_VariableAppendBlock --

	Appends a byte array to a variable.

   Arguments:

	variable -	Pointer to the variable.
	block -		The data block.

   Results:

        Appends the value. Errors are ignored.
	
*/

void
More_VariableAppendBlock (More_Variable variable, More_Block block)
{
  Tcl_Obj *	objPtr = Tcl_NewByteArrayObj(block.ptr, block.len);

  Tcl_IncrRefCount(objPtr);
  Tcl_ObjSetVar2(VARIABLE.interp, VARIABLE.name, NULL, objPtr,
		 TCL_APPEND_VALUE);
  Tcl_DecrRefCount(objPtr);
}

/* More_VariableClear --

	Resets a variable to the empty string.

   Arguments:

	variable -	Pointer to the instance.

   Results:

        Clears the variable, errors are ignored.

*/

void
More_VariableClear (More_Variable variable)
{
  Tcl_Obj *	objPtr;


  objPtr = Tcl_NewStringObj(NULL, -1);
  Tcl_IncrRefCount(objPtr);
  Tcl_ObjSetVar2(VARIABLE.interp, VARIABLE.name, NULL, objPtr, 0);
  Tcl_DecrRefCount(objPtr);  
}

/* More_VariableGet --

	Extracts the current value from a variable.

   Arguments:

	variable -	Pointer to the instance.

   Results:

        Returns a pointer to the object. Errors are ignored.

*/

Tcl_Obj *
More_VariableGet (More_Variable variable)
{
  return Tcl_ObjGetVar2(VARIABLE.interp, VARIABLE.name, NULL, 0);
}

/* More_VariableGetBlock --

	Extracts  the current  value from  a  variable, seen  as a  byte
	array.

   Arguments:

	variable -	Pointer to the instance.

   Results:

        Returns a pointer to the object. Errors are ignored.

*/

More_Block
More_VariableGetBlock (More_Variable variable)
{
  More_Block	block;
  Tcl_Obj *	objPtr;

  objPtr = Tcl_ObjGetVar2(VARIABLE.interp, VARIABLE.name, NULL, 0);
  block.ptr = Tcl_GetByteArrayFromObj(objPtr, &block.len);
  return block;
}

/* More_VariableName --

	Accesses the name of the variable.

   Arguments:

	variable -	The variable.

   Results:

        Returns a pointer to the variable name.

*/

CONST char *
More_VariableName (More_Variable variable)
{
  return Tcl_GetString(VARIABLE.name);
}


/* end of file */
/*
Local Variables:
mode: c
page-delimiter: "^$"
End:
*/
