/*
  This file is meant to be included (with #include) in a source
  file provided by you that implements the command functions. Your
  file must include the TCL header ("tcl.h") first and define the
  functions later.

    The reason of this inclusion is that here are declared the struct
  typedefs of the command invocation states. These types must be
  visible by the command functions. The stuff declared in this file
  uses the "static" keyword as much as possible.

    The command functions should be small: they should do the job by
  invoking other functions, the ones that the extension should export
  in its stub table.
*/


/** ------------------------------------------------------------
 ** Header files.
 ** ----------------------------------------------------------*/

#include "tclmore.h"



/** ------------------------------------------------------------
 ** Type declarations.
 ** ----------------------------------------------------------*/

/* OneState --

  This structure is used to hold the invocation state of the
  following commands:

                    [one   integer number]
*/
typedef struct OneState {
  int                     integer;
  double                  number;
} OneState;

static CONST OneState OneStateDefaults = {
  123, 1.2
};

/* TwoState --

  This structure is used to hold the invocation state of the
  following commands:

                    [two   ?options?]
*/
typedef struct TwoState {
  int                     foreground;
  double                  background;
} TwoState;

static CONST TwoState TwoStateDefaults = {
  DEFAULT_FOREGROUND_MACRO, DEFAULT_BACKGROUND_MACRO
};

/* ThreeState --

  This structure is used to hold the invocation state of the
  following commands:

                    [three   integer number ?options?]
*/
typedef struct ThreeState {
  int                     integer;
  double                  number;
  int                     mode;
} ThreeState;

static CONST ThreeState ThreeStateDefaults = {
  123, 1.2, 0
};



/** ------------------------------------------------------------
 ** Implementation functions prototypes.
 ** ----------------------------------------------------------*/

/*
  You must provide the implementation of each of these functions in
  this file. When the function is invoked, the state invocation
  structure has been already filled with values from the command line
  arguments and options, with no errors.

    The function should look like the following:

  int
  MyCommandFunction (More_CommandFrame frame)
  {
     MyStateStruct *  state = (MyStateStruct *) frame->stateStructPtr;
     Tcl_Obj *        result;

     *
       Make use of the fields of "*state" to build a result by
       invoking functions from your library.
     *
     result = Mylib_DoSomething(state->value, state->another_value);

     Tcl_SetObjResult(frame->interp, result);
     return TCL_OK;
  }

    Being the implementation of a TCL command, this function is
  declared "static": it is useless to share this function among
  different modules. Remember that the role of a "More_CmdFunc"
  is to be an interface between the state invocation structure and
  the functions of your library that do the "real work".
*/

static More_CmdFunc         OneGreenFunction;
static More_CmdFunc         OneWhiteFunction;
static More_CmdFunc          OneRedDarkFunction;
static More_CmdFunc          OneRedBrightFunction;
static More_CmdFunc        FirstFunction;
static More_CmdFunc        TwoFunction;
static More_CmdFunc        ThreeFunction;


/** ------------------------------------------------------------
 ** Extractor functions prototypes.
 ** ----------------------------------------------------------*/

/*
  You must provide the implementation of each of the static extractors
  in this file. Functions declared EXTERN can be implemented in other
  modules of your extension or in other libraries that are dependencies
  of this one (for example the TCLMORE library implements some common
  extractors).

    The functions must be available as pointers: you cannot directly use
  functions exported in a stub table; this is because the pointer must
  be included in a statically allocated data structure. Remember that
  stub function names are available as macros, each macro references
  an element in a vector of function pointers in another library: the
  value of the function pointer is available only at execution time.

    The only way to use an extractor from a stub table is to declare
  another extractor (with a similar name) that wraps the real function.
  Example:

     More_Extractor MyExtractor;
     ...
     int
     MyExtractor (More_ExtractorFrame frame)
     {
       return RealExtractor(frame);
     }

  "RealExtractor" can be the macro of a stub function.

    If a prototype is omitted: it is your responsibility to declare it
  before the line that includes this file in your source module. Obviously
  the declaration can be in an external header file; this is particularly
  useful if your extension is using a wrapper extractor for a real extractor
  from another extension, and the extractor is used in more than one
  set of commands.

    A common extractor has the following structure, lets assume that
  "Type" is the type of value we want:

     EXTERN int Mylib_GetTypeFromObj _ANSI_ARGS_((Tcl_Interp * interp,
                                                  Tcl_Obj *    srcObj,
                                                  Type *       valueVar));
     ...
     int
     TypeExtractor (More_ExtractorFrame frame)
     {
        return Mylib_GetTypeFromObj(frame->commandFrame->interp,
                                    frame->srcObj,
                                    (Type *) frame->fieldPtr);       
     }

  or:

     int
     TypeExtractor (More_ExtractorFrame frame)
     {
        int   e;

        e = Mylib_GetTypeFromObj(frame->commandFrame->interp,
                                 frame->srcObj,
                                 (Type *) frame->fieldPtr);       
        return ((TCL_OK == e)? e : More_LogicError(interp));
     }

  the implementation of a "GetTypeFromObj" function to do the real
  work is recommended because such functions can be used by other
  modules, and even exported in the stub table.

    Remember that the field pointer can reference a whole structure
  embedded in the state invocation one.

    At present an extractor can validate arguments only with complete
  independence from the current process state, that is it can test that
  a value is in the correct space and range. Coherence tests between
  argument values and process state must be performed by the command
  implementation functions. This is not exactly true, but it is the
  recommended way to do things.
*/

EXTERN More_Extractor      GetIntFromArg;
/* omitted More_Extractor GetDoubleFromArg */
static More_Extractor      GetColorFromArg;
static More_Extractor      GetModeFromArg;


/** ------------------------------------------------------------
 ** Command arguments tables.
 ** ----------------------------------------------------------*/

/* No arguments for [one green]. */
#define one_green_ARGUMENT_NUMBER	0

/* No arguments for [one white]. */
#define one_white_ARGUMENT_NUMBER	0

/* No arguments for [one red dark]. */
#define one_red_dark_ARGUMENT_NUMBER	0

/* No arguments for [one red bright]. */
#define one_red_bright_ARGUMENT_NUMBER	0

#define one_ARGUMENT_NUMBER	2
static CONST More_CommandArgument OneArguments[2] = {
  { GetIntFromArg,       offsetof(struct OneState, integer) },
  { GetDoubleFromArg,    offsetof(struct OneState, number) }
};

/* No arguments for [two]. */
#define two_ARGUMENT_NUMBER	0

#define three_ARGUMENT_NUMBER	2
static CONST More_CommandArgument ThreeArguments[2] = {
  { GetIntFromArg,       offsetof(struct ThreeState, integer) },
  { GetDoubleFromArg,    offsetof(struct ThreeState, number) }
};



/** ------------------------------------------------------------
 ** Command options tables.
 ** ----------------------------------------------------------*/

/* No options for [one green]. */

/* No options for [one white]. */

/* No options for [one red dark]. */

/* No options for [one red bright]. */

/* No options for [one]. */

static CONST More_CommandOption TwoOptions[3] = {
  { "-foreground", 1, GetColorFromArg, offsetof(struct TwoState, foreground) },
  { "-background", 1, GetColorFromArg, offsetof(struct TwoState, background) },
  { NULL, 0, NULL, 0 }
};

static CONST More_CommandOption ThreeOptions[3] = {
  { "-on", 0, GetModeFromArg, offsetof(struct ThreeState, mode) },
  { "-off", 0, GetModeFromArg, offsetof(struct ThreeState, mode) },
  { NULL, 0, NULL, 0 }
};


/** ------------------------------------------------------------
 ** Leaf commands array.
 ** ----------------------------------------------------------*/

/*
  There is a leaf structure for each "More_CmdFunc". Each structure
  has references to the argument and option tables. It is not your
  business to mess with this structure.
*/

static More_CommandTreeLeaf commandTreeLeaves[] = {
  /* (0) synopsis: one green  */
  { 2, OneGreenFunction, 0, (CONST ClientData)NULL,
    one_green_ARGUMENT_NUMBER, NULL, NULL,
    NULL },

  /* (1) synopsis: one white  */
  { 2, OneWhiteFunction, 0, (CONST ClientData)NULL,
    one_white_ARGUMENT_NUMBER, NULL, NULL,
    NULL },

  /* (2) synopsis: one red dark  */
  { 3, OneRedDarkFunction, 0, (CONST ClientData)NULL,
    one_red_dark_ARGUMENT_NUMBER, NULL, NULL,
    NULL },

  /* (3) synopsis: one red bright  */
  { 3, OneRedBrightFunction, 0, (CONST ClientData)NULL,
    one_red_bright_ARGUMENT_NUMBER, NULL, NULL,
    NULL },

  /* (4) synopsis: one integer number */
  { 1, FirstFunction, sizeof(OneState), (CONST ClientData)&OneStateDefaults,
    one_ARGUMENT_NUMBER, OneArguments, NULL,
    "integer number" },

  /* (5) synopsis: two ?options? */
  { 1, TwoFunction, sizeof(TwoState), (CONST ClientData)&TwoStateDefaults,
    two_ARGUMENT_NUMBER, NULL, TwoOptions,
    "?options?" },

  /* (6) synopsis: three integer number ?options? */
  { 1, ThreeFunction, sizeof(ThreeState), (CONST ClientData)&ThreeStateDefaults,
    three_ARGUMENT_NUMBER, ThreeArguments, ThreeOptions,
    "integer number ?options?" }
};


/** ------------------------------------------------------------
 ** Node commands array.
 ** ----------------------------------------------------------*/

/*
  These are node structures that declare subcommands with subcommands.
  It is not your business to mess with this structure.
*/

static More_CommandTreeNode OneRedCommandNode[] = {
  { "dark", TCLMORE_SAFE_COMMAND,  NULL, &commandTreeLeaves[2] },
  { "bright", TCLMORE_SAFE_COMMAND,  NULL, &commandTreeLeaves[3] },
  { NULL, TCLMORE_UNSAFE_COMMAND, NULL, NULL }
};

static More_CommandTreeNode OneCommandNode[] = {
  { "green", TCLMORE_SAFE_COMMAND,  NULL, &commandTreeLeaves[0] },
  { "white", TCLMORE_SAFE_COMMAND,  NULL, &commandTreeLeaves[1] },
  { "red", TCLMORE_SAFE_COMMAND,  OneRedCommandNode, NULL },
  { NULL, TCLMORE_UNSAFE_COMMAND, NULL, NULL }
};



/** ------------------------------------------------------------
 ** Root commands table.
 ** ----------------------------------------------------------*/

/*
  This array of structures declares the root commands. There is one of
  them for each <maincommand> and <command> markups with no ancestors
  in the original XML document. It is not your business to mess with
  the content of this array.

    You may create commands declared in this array with the following
  chunk of code:

    error = More_CreateCommands(interp, "namesp", ExampleCommands);
    if (error)
      {
	return More_ErrorResult(interp, error);
      }

  where "namesp" is the name of the namespace in which the command
  must be created. If the namespace is the root one: you can use NULL.
*/

More_CommandTreeNode ExampleCommands[] = {
  { "one", TCLMORE_SAFE_COMMAND, OneCommandNode, NULL },
  { "one", TCLMORE_SAFE_COMMAND, OneCommandNode, NULL },
  { "two", TCLMORE_SAFE_COMMAND, NULL, &commandTreeLeaves[5] },
  { "three", TCLMORE_SAFE_COMMAND, NULL, &commandTreeLeaves[6] },
  { NULL, 0, NULL, NULL }
};


/* end of file */
/*
Local Variables:
mode: c
page-delimiter: "^$"
End:
*/

